//***********************************************************************
//*           See Accompanying licence file QSAS_licence                *
//***********************************************************************

// Header for QIE File Object using QDOS 2 
// Allows read of file with single scan
// CAA compliant

// ensure cdf.h is included before declaractions dependent on CDF types 

#if !defined(___cdf_h___)
#include "cdf.h"
#endif
 
#ifndef _QIE_CAA_H_  
#define _QIE_CAA_H_

#define QIE_VERSION  "QIE V_3.6.3 [Sept 2012]"

#include "DvObj.h"
#include "Xrefs.h"
#include <fstream>

// DEFINES


#define DATA_DBL 0
#define DATA_INT 1
#define DATA_STR 2
#define DATA_TIME 3
#define DATA_EVENT 4
#define DATA_FLT 5 // gives smaller CDF files

#ifndef _QIE_H_ 
// avoid clash with qie.h
#define ISO_TIME 3030
#define ISO_TIME_RANGE 3060
#define CDF_OPEN_ERR 102
#endif

#define Q_TIME_WIDTH 30
#define Q_EVENT_WIDTH 50
#define Q_DOUBLE_WIDTH 22
#define Q_INT_WIDTH 13
#define MAX_LINE_LEN_CEF2 256

#define QIE_VERSION_NUM  "QIE V_4.0.18 [July 2016]"

#define TO_UPPER_LEN 500
#define QIE_RESET -1

class QieFT
{

public:
	QieFT();
	~QieFT(){;}
	void parseFreeTime(); // convert FreeTime string into pointers and lengths
	void assembleTime(size_t n, const char * record); // convert FreeTime fields to time elements
	
		// free time format elements
	string ft_format; // string containing free time format description
	string ft_all_time; // string with time to add to all free times
	string ft_all_time_fmt; // format of string to add to all free times
	DvTime ft_offset; // ISO time string to add to all free times
	bool use_ft_offset; // flag to know if to add ft_offset (DvTime() is never zero)
	bool use_DoY; // flag set if doy field found
	int yy_offset;  // start year offset
	int mm_offset;  // start month offset
	int dd_offset;  // start day offset
	double ss_offset;  // seconds offset
	int ft_y; // location of year in record
	int ft_y_len;  // length of year
	int ft_m; // location of month in record
	int ft_m_len; // length of month
	int ft_m_str; // location of text month in record (eg 'Apr')
	int ft_m_str_len; // length of month text
	int ft_d; // location of day in record
	int ft_d_len; // length of day
	int ft_d_frac; // location of fractional part of day in record
	int ft_d_frac_len; // length of fractional part of day field
	int ft_doy; // location of day of year in record
	int ft_doy_len; // length of day of year
	int ft_hr; // location of hour in record
	int ft_hr_len; // length of hour field
	int ft_hr_frac; // location of fractional part of hour in record
	int ft_hr_frac_len; // length of fractional part of hour field
	int ft_min; // location of minute in record
	int ft_min_len; // length of minute field
	int ft_min_frac; // location of fractional part of hour in record
	int ft_min_frac_len; // length of fractional part of hour field
	int ft_sec; // location of seconds in record
	int ft_sec_len; // length of seconds field
	int ft_sec_frac; // location of fractional part of seconds in record
	int ft_sec_frac_len; // length of fractional part of seconds field
	int ft_msec; // location of milliseconds in record
	int ft_msec_len; // length of milliseconds field
	int ft_msec_frac; // location of fractional part of milliseconds in record
	int ft_msec_frac_len; // length of fractional part of milliseconds field
	int ft_musec; // location of microseconds in record
	int ft_musec_len; // length of microseconds field
	int ft_musec_frac; // location of fractional part of microseconds in record
	int ft_musec_frac_len; // length of fractional part of microseconds field
	int ft_nsec; // location of nanoseconds in record
	int ft_nsec_len; // length of nanoseconds field
	int ft_nsec_frac; // location of fractional part of nanoseconds in record
	int ft_nsec_frac_len; // length of fractional part of nanoseconds field
	vector <int> year;
	vector <int> month;
	vector <int> day;
	vector <int> dayOfYear;
	vector <double> hour;
	vector <double> minute;
	vector <double> second;
	int NdecSec;

};


//-------------------------------------------

class QieFile 
{
   
public:
   QieFile();
   ~QieFile();
   void openFile(DvString fullName);
   void closeFile();
   char readLine(size_t &posStart, size_t &posEnd);
   void setEOL(char eol);
   void setCMark(char cmark);
   void setStartAfter(DvString &safter);
   void setDataUntil(DvString &safter);
   void setAttrDelim(char delim);
   void setDataDelim(char c);
   void setRecNumbering(bool recNumOn);
   bool skipNextRecord();
   int  skipRestOfLine();
   char get();
   size_t countRecords();
   bool dataStartIn(size_t posStart, size_t posEnd);
   DvString safePath(const char * path);

   std::ifstream ifp; 
   std::ofstream ofp; 
   DvString name;
   DvString header;
   DvString extn;
   char  end_of_line;
   char  commentMark;
   char attr_delim;
   char data_delim;
   string start_after;
   string data_until;
   char * content;
   size_t posn;
   vector <char *> recPosn;
   bool hasTitle;
   bool CDFbackward;
   bool force_eol;

protected:

public: 

private:

};


class QieVariable
{

public:
	QieVariable(const char *variableName="");
	~QieVariable();
	void ensureDeltas();
	void processTime(); // convert FreeTime elements to time

	DvString name;     // title string for var name 
	int datatype; // data type for variable
	int  n_posn;          // start position in vector of tokens in record
	int  col_start;       // posn for var in record, start at 0 
	int  col_width;       // column width of var in record - length between entries
	int  entry_len;       // data width of var in record - length of data entry
	bool recVary;		    // flag for variable true if record varying
	vector <size_t> objDim;	// list of data object dimensions
	size_t nEntries;		// Total number of entries (product of array dims)
	DvObject_var Vobj;     // The DVOS object itself for transferring to/from WL
	DvList Vxref;    	   // linked list to keep xrefs until Vobj ready
	bool getThis;         // flag for reading
	int nRecs;            // number of records (not always same for sparse cdf files)
	
	bool useFreeFormat; // flag that time field(s) not standard ISO
	QieFT ft;  // container for free time info
	bool skip; // flag to ignore var on import/write used to remove Date/Time component
	DvString NaNstr; // is set to "nan" if fillval is NaN
	
} ;



//-------------------------------------------
/// \class QieData
/// \brief This is the QieData class
///
/// \remarks { This class implements a CAA/CDF file object using DVOS classes
/// It provides native QSAS2 internals replicating the QIE QiSCDFContents object }

class QieData
{
   
public:
   QieData();
   ~QieData();

   int  queryFile(DvString fullName);
   bool readHeader();
   bool readGlobal(const char *globalAttrName);
   QieVariable * initVariable(const char *variableName, QieFile *vFile=NULL);
   DvObject_var readTT(size_t vnum, size_t recNum);
   DvObject_var readTT_TAB(size_t vnum);
   bool readVarData(int vnum);
   void readData(size_t recStart, size_t recEnd);
   DvObject * QieText_to_Attr(DvString  valueStr, QieFile *vFile=NULL);
   DvObject * QieText_to_Obj(QieVariable *V, std::string  valueStr, QieFile *vFile=NULL);
   bool parseLine(char *nextLine, size_t posStart, size_t posEnd);
   size_t  countEntryElements(string &inputStr, const char delim);
   string getNextEntry(string &inputStr, int &startAt, const char delim);
   void initSequence(QieVariable * V, int nRecs);
   void getVarRecordEntry(int nToGet, int recNum);
   void getVarRecordEntrySpSep(int nToGet, int recNum);
   void getVarRecordTableEntry(int nToGet, int recNum);
   void attachXrefs(int varN, DvString varName="");
   void ensureDependZero();

   // Public interface to completed QieData object
   int  findVarNumByName(const char * name);
   void clearGetThis();
   void setGetThis(int vnum);
   DvObject_var getD0_obj(int vnum); 
   void countEntries(int &gattr_datatype, int &num_elems, vector <string> &valueList);

   // Rosetta (LBL & TAB files)
   int  queryTABfile(DvString fullName);
   int read_LBL();
   bool initColumn(int delimLen=0);
   int getLBLDataType(DvString dataType);
   void postRunRosetta();

   // ASCI Writing Methods
   bool ensureFileName(bool replace, bool detached=false);
   bool writeFlat(bool replace, bool detached=false);
   void writeHeader(bool detached);
   void writeGlobals();
   void writeGEntry(DvObject_var &dobj, int r);
   void writeVarBlock(QieVariable *V, int &startPosn);
   void writeVattrs(QieVariable *V);
   void putDelim(char delim, bool quoted=false);
   void writeVarData(QieVariable *V, size_t r, char delim);
   void writeVarDataCAA(QieVariable *V, size_t r);
   void writeVarDataQFT(QieVariable *V, size_t r);
   void writeData();
   void writeDataCAA();
   void writeDataQFT();
   
   // write utilities
   void setSecDecimals(int value);
   DvString addObj(DvString name, DvObject_var obj, int accuracy);
    bool notHomogeneous();
   
   // CDF methods
   int queryCDFfile(DvString fullName);
   bool getCDFvariables(int n_vars);
   void getCDFVar(int nvar);
   void addVarAttrs(int nvar);
   void getCDFglobals();
   void getGlobalAttr();
   void getCDFvarDataList(vector <size_t> getList, int recStart, int recEnd);
   void getCDFvarData(size_t nvar, int recStart, int recEnd);
   void convertCDFbufferToObj(size_t vnum, size_t seq_size, long data_type, size_t strLength, void *data);
   void fixThemisEpoch(vector <size_t> getList);
   void getAttributeHelper(const char * fullName);
   void readAttrHelper();
   char * getNextRecEntry();
   string getNextRecEntryStr();
   char * getNextRecEntrySpSep();
   string getNextRecEntryStrSpSep();
   void setRecEntry(int n);
   void setRecEntrySpSep(int n);
   int getDataType(DvString dataType);
   DvString DataTypeToStr(int dataType);
   bool mergeTimeVar(int vnum);
   void convertColToRow(size_t vnum);
   
   // CDF writing
   bool writeCDF(bool replace=false, bool compress=true);
   void removeStrArrays();   
   bool createCDF(bool backCompat=false, bool replace=true);  
   bool writeContentsCDF();
   bool writeGlobalsCDF();
   bool writeVariableCDF(int n);
   size_t sizeOfCDF( DvObject_var &obj, size_t &len);
   long getCDFDataType(DvObject_var &obj);
   void ensureVariable(int n);
   void writeVarAttrs(int n);

   // flags and values
   vector <QieVariable*>  QieVars;   
   DvList GlobalList;
   size_t nRecords;
   int d0_var_num;   
   QieFile inFile;
   QieFile outFile;
   QieFile helperFile;
   long yp_last;
   DvString parameter;
   DvString value;
   int recordNumEntries;
   int maxCDFnumRecs; // used for cdf as we leave data in file until needed. Vars may have fewer recs.
   int nEntry;
   char *recVarEntry;
   int dataSecRes;
   bool precIsSet;
   
protected:

private:

	   
};

// Utility Functions
// -------------------------

void  QieMsg(const char *status, const char *line1, const char *line2 = NULL,  const char *line3 = NULL);
vector <DvString> QieFindHelperFiles(const char * dataFile, const char * dotExtn);
bool  QieStrIsNULL(const char *txt);
char QieGetDelim (std::string value);
bool QieIsISOtime(std::string TestStrISO);
bool QieIsISOtimeRange(std::string TestStrISO);
bool QieIsNumber(std::string ptr);
void setDebug(bool d_on, const char *fpath = NULL);
void *GAttrEntryMalloc( long num_elems, long data_type );
DvObject *CDFattrBufferToObj(void *data, int data_type, int num_elems);
bool cefWrite2 (int set);
bool cefRead2 (int set);
DvObject *createObj(int data_type, size_t seqLen, vector <size_t> &dim, DvString &typeStr);

int getSecResolution(string &value);

#endif
