/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
#ifndef _QWUI_H
#define _QWUI_H

#include <string.h>
#include <vector>
#include <qfiledialog.h>
#include <qstatusbar.h>

#include <qlayout.h>
#include <qfont.h>
#include "qt_version.h"

#include "qsw_helpwindow.h"
#include "qbu.h"

#include "qtpl.h"
#include "qdutil.h"
#include "qprofile.h"
#include "qxml.h"

#include "WList.h"
#include "WLItem.h"

#include <QTextStream>          
#include <QFile>          
#include <QProcess>         
#include <QToolBar>         
#include <QStatusBar>         
#include <QTextStream>      
#include <QDialogButtonBox> 
#include <QTextEdit>        
#include <QGridLayout>      
#include <QClipboard>


extern QFont defaultFontQSAS; // Initialised with default size at startup

class qwuiWindow;

   
class qwuiWindow : public QMainWindow
{
   Q_OBJECT

public:
    qwuiWindow(QWidget *parent=0, const char *name=0);
    void makeWindows();
    QCheckBox *StripBadData;
    QCheckBox *StripFills;
    QCheckBox * overWrite;
    QString _filename;        ///< Filename used to save the current session or working list content
    QString SAVE_EXTENSION;   ///< File extension used to save the current session or working list content
    bool fromSession;   

    static QWidget *toggleBar;
    void Qwui_msgBox(const char *  );
    void addFolderToMenu(const char *folder, qtplMenu *menu, bool first = true);
    void status(const QString& str) { statusBar()->showMessage(str);}
    void status(const QString& str,int ms) {statusBar()->showMessage(str,ms);}

    WList* wlview;
    qtplMenu * timeOps;
    qtplMenu * mathOps;
    qtplMenu * filterOps;
    qtplMenu * vectorOps;
    qtplMenu * arrayOps;
    qtplMenu * statOps;
    qtplMenu * unitOps;
    qtplMenu * eventTableOps;
    vector <qtplMenu *> pluginFolders;
    QString qsas_home;
    QswHelpWindow *help_window;
    QString _sessionFolder;
    QToolBar *mainToolBar;
    
    QAction *restoreAction;
    QAction *saveAction;
    QAction *openAction;
    QAction *selectorAction;
    QAction *deleteAction;
    QAction *dellistAction;
    QAction *calcAction;
    QAction *joinAction;
    QAction *plotAction;
    QAction *timeAction;
    
	signals:
    void saveSessionRequest(QString );
    void restoreSessionRequest(QString);
    void clearListRequest();
    void openQDA_Window(const char *);
    void runStaticPlugin(const char *);
    void updatePlotRequest();
    void directPrintRequest(const char *);
    void calcPageRequest(int);
    void restoreCalcRequest(QString fn);
    void restorePlotRequest(QString fn);
    void readDataRequest(DvString&, DvString&, DvEvent&);

  
public slots:
    void setPreferences();
    void launchManual();
    void launchManMainW();
    void launchManSampleW();
    void launchManDataH();
    void launchManSaveRestore();
    void launchVersion();
    void launchContact();
    void saveSession();
    void saveSessionAs();
    void restoreSession();
    void restoreNamedSession(QString fn);
    void restoreCalc(QString fn);
    void restorePlot(QString fn);
    void includeInWL();
    bool restoreWL(bool replace = true);
    void saveWL();
    void saveWLAs();
	void readData(DvString&, DvString&, DvEvent&);
	void calcPage(int p);
	void directPrint(const char *);
    void updatePlot();
    void bringToFront(){activateWindow();raise();}

protected:
	void closeEvent( QCloseEvent* );
	void dragMoveEvent( QDragMoveEvent *event){;}
	void dragEnterEvent( QDragEnterEvent *event);


private:
   QWidget* centralW;
   QVBoxLayout * centralBox;
   
};
 

void  setStandardConstants();



class SessionFolderDialog : public QFileDialog {

Q_OBJECT

  
  QStatusBar* _status_bar;
  QMenuBar *_menu_bar;
  
public slots:
    QString getSessionFolder();
  
signals:
  
public:
  SessionFolderDialog( QWidget *parent=0, const char *name=0 );
  ~SessionFolderDialog(){;}
  void status(const QString& str)  {_status_bar->showMessage(str);}
  QMenuBar* menu_bar(){ return _menu_bar;}
  QLabel * label;
  
protected:

private slots:
  
};

// Qwui Utilities

QString QwuiPutOnWL(DvObject_var obj, const QString& path, bool replace=false);
size_t Qwui4VecRcomp(DvObject_var obj);
DvObject_var QwuiGet4VecR(DvObject_var obj, size_t rComp);
DvObject_var QwuiGet3Vec(DvObject_var obj, size_t rComp);

QSize qsasGUISize();
void QwuiClearTemp();
QString QwuiTextInputDialog(const char *caption, const char *label, const char * defText, bool *OK, QWidget * parent);


#endif

/**********************************************************************/

