#include "qcalcEvaluator.h"


void QEvaluatorExp::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
	ok=true;
	DvObject_var result;
	
	QEvaluator::evaluate(in, out, ok, message);
	if(!ok)
		return;
	
	DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[0]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	
	if(val1->isThreeVector())
	{
		message=QString("Operation is not valid for vector objects");
		ok=false;
		return;
	}
	
	if (val1->hasUnits())
	{
		message=QString("WARNING!  Operation Exp() or Log() -> Subject has units");
		ok=false;
		return; 
	}
	
	// do arithmetic...

	result = val1->exp();
	

	if ( result.is_nil() )
	{
		message=QString( "Arithmetic>> Exp is not yet supported for the input type" );
		ok=false;
		return;
	}

	out[0]->setValue(result, true);
}


void QEvaluatorTrace::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
    ok=true;
    DvObject_var result;
    
    QEvaluator::evaluate(in, out, ok, message);
    if(!ok)
        return;
    
    DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[0]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    
    if(val1->nDims() != 2) 
    {
        message=QString("Operation is only valid for 2D matrix");
        ok=false;
        return;
    }
    
    if( val1->dims[0] != val1->dims[1] )
    {
        message=QString("Operation is only valid for square matrix");
        ok=false;
        return;
    }
    
    result = val1->trace();
    
    if ( result.is_nil() )
    {
        message=QString( "Trace returned empty object" );
        ok=false;
        return;
    }
    
    out[0]->setValue(result, true);
}


void QEvaluatorMatInv::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
    ok=true;
    DvObject_var result;
    
    QEvaluator::evaluate(in, out, ok, message);
    if(!ok)
        return;
    
    DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[0]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    
    if(val1->nDims() != 2) 
    {
        message=QString("Operation is only valid for 2D matrix");
        ok=false;
        return;
    }
    
    if( val1->dims[0] != val1->dims[1] )
    {
        message=QString("Operation is only valid for square matrix");
        ok=false;
        return;
    }
    
    result = val1->matrixInverse();
    
    if ( result.is_nil() )
    {
        message=QString( "Matrix Inverse returned empty object" );
        ok=false;
        return;
    }
    
    out[0]->setValue(result, true);
}

void QEvaluatorTranspose::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
    ok=true;
    DvObject_var result;
    
    QEvaluator::evaluate(in, out, ok, message);
    if(!ok)
        return;
    
    DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[0]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    
    if(val1->nDims() != 2) 
    {
        message=QString("Operation is only valid for 2D matrix");
        ok=false;
        return;
    }
    
    result = val1->transpose();
    
    if ( result.is_nil() )
    {
        message=QString( "Transpose returned empty object" );
        ok=false;
        return;
    }
    
    out[0]->setValue(result, true);
}

void QEvaluatorDeterminant::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
    ok=true;
    DvObject_var result;
    
    QEvaluator::evaluate(in, out, ok, message);
    if(!ok)
        return;
    
    DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[0]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    
    if(val1->nDims() != 2) 
    {
        message=QString("Operation is only valid for 2D matrix");
        ok=false;
        return;
    }
    
    if( val1->dims[0] != val1->dims[1] )
    {
        message=QString("Operation is only valid for square matrix");
        ok=false;
        return;
    }
    
    result = val1->det();
    
    if ( result.is_nil() )
    {
        message=QString( "Matrix Determinant returned empty object" );
        ok=false;
        return;
    }
    
    out[0]->setValue(result, true);
}

void QEvaluatorAdjoint::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
    ok=true;
    DvObject_var result;
    
    QEvaluator::evaluate(in, out, ok, message);
    if(!ok)
        return;
    
    DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[0]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    
    if(val1->nDims() != 2) 
    {
        message=QString("Operation is only valid for 2D matrix");
        ok=false;
        return;
    }
    
    if( val1->dims[0] != val1->dims[1] )
    {
        message=QString("Operation is only valid for square matrix");
        ok=false;
        return;
    }
    
    result = val1->adjoint();
    
    if ( result.is_nil() )
    {
        message=QString( "Matrix Adjoint returned empty object" );
        ok=false;
        return;
    }
    
    out[0]->setValue(result, true);
}

void QEvaluatorElement_i::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
    ok=true;
    DvObject_var result;
    
    QEvaluator::evaluate(in, out, ok, message);
    if(!ok)
        return;
    
    DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[0]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    DvObject_var val2=in[1]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[1]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    int i = val2->asInt() -1 ;
    if(i < 0 || i >= val1->arraySize() ) {
        message=QString("Input ")+in[1]->getName()+QString(" out of range");
        ok=false;
        return;
    }
    
    result = val1->element((size_t)i);
    
    if ( result.is_nil() )
    {
        message=QString( "Element(i) returned empty object" );
        ok=false;
        return;
    }
    
    out[0]->setValue(result, true);
}


void QEvaluatorInner::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
    ok=true;
    DvObject_var result;
    
    QEvaluator::evaluate(in, out, ok, message);
    if(!ok)
        return;
    
    DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[0]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    DvObject_var val2=in[1]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[1]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    result = val1 * val2;
    
    if ( result.is_nil() )
    {
        message=QString( "Inner product returned empty object" );
        ok=false;
        return;
    }
    
    out[0]->setValue(result, true);
}

void QEvaluatorOuter::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
    ok=true;
    DvObject_var result;
    
    QEvaluator::evaluate(in, out, ok, message);
    if(!ok)
        return;
    
    DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[0]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    DvObject_var val2=in[1]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[1]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    result = val1->outerProduct(val2);
    
    if ( result.is_nil() )
    {
        message=QString( "Outer product returned empty object" );
        ok=false;
        return;
    }
    
    out[0]->setValue(result, true);
}

void QEvaluatorElement_ijk::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
    ok=true;
    DvObject_var result;
    
    QEvaluator::evaluate(in, out, ok, message);
    if(!ok)
        return;
    
    DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
    if(!ok )
    {
        message=QString("Input ")+in[0]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    DvObject_var val2=in[1]->connectedSlot()->getValue(ok);
    if(!ok )
    {
        message=QString("Input ")+in[1]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    // of the form "1,2,...,n"
    size_t idx=0;
    size_t n = val1->nDims();
    int j=0;
    
    DvString str_idx = val2->asStr();
    for(size_t i=0; i<n-1; i++) {
        j = str_idx.before(',').toInt()-1;
        if(j < 0) {
            message=QString("Array indices must be > 1 (not C style)");
            return;
        }
        idx += (size_t)j;
        str_idx = str_idx.after(',');
        idx *= val1->dims[i+1];
    }
    j = str_idx.toInt()-1;
    if(j < 0) {
        message=QString("Array indices must be > 1 (not C style)");
        return;
    }
    idx += j;
    
    result = val1->element((size_t)idx);
    
    if ( result.is_nil() )
    {
        message=QString( "Element(i,j,k...) returned empty object" );
        ok=false;
        return;
    }
    
    out[0]->setValue(result, true);
}

void QEvaluatorRemove_i::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
    ok=true;
    DvObject_var result;
    
    QEvaluator::evaluate(in, out, ok, message);
    if(!ok)
        return;
    
    DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
    if(!ok || val1->nDims() != 1 )
    {
        message=QString("Input ")+in[0]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    DvObject_var val2=in[1]->connectedSlot()->getValue(ok);
    if(!ok )
    {
        message=QString("Input ")+in[1]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    int i = val2->asInt() -1 ;
    if(i < 0 ||  i >= val1->Dims()[0]) {
        message=QString("Input ")+in[1]->getName()+QString(" out of range");
        ok=false;
        return;
    }
    
    result = val1->subDimension((size_t)i);
    
    if ( result.is_nil() )
    {
        message=QString( "Remove(i) returned empty object" );
        ok=false;
        return;
    }
    
    out[0]->setValue(result, true);
}

void QEvaluatorTTtoSec::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
	
	ok=true;
	DvObject_var result;
	
	QEvaluator::evaluate(in, out, ok, message);
	if(!ok)
		return;
	
	DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[0]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	DvObject_var tt = val1->getTimeTags();
	
	DvObject_var startTime=in[1]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[1]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	
	if(tt.is_nil() ||  tt->seqSize() <= 0){
		message=QString("Could not get valid time tags from input");
		QplugAppendTextDisplay("Timetags to Sec Operator: timetags not found");
		ok=false;
		return;
	}
	if( !startTime->is_time() ){
		message=QString("2nd Object did not resolve to a time");
		QplugAppendTextDisplay("Timetags to Sec Operator: input not Time object. \nTry adding Extractor.");
		ok=false;
		return;
	}
	
	DvObject_var ss = new DvObject(0.0, tt->seqSize());
	for(int i=0; i<tt->seqSize(); i++){
		ss->dbl(i) = tt->time(i) - startTime->time(0);
	}
	
	DvString unitStr = "sec";
	ss->change_xref(UNITS, unitStr);
	DvString SI_Str = "1>s";
	ss->change_xref(SI_CONVERSION, SI_Str);
	DvString NameStr = "Sec after ";
	NameStr += startTime->time(0).getISOstring();
	
	ss->change_xref(FIELDNAM, NameStr);

	result = new DvObject(val1); // constructor copies xrefs
	
	result->change_xref(DEPEND_0, ss);
			 
	out[0]->setValue(result, true);

}

void QEvaluatorSectoTT::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
    
    ok=true;
    DvObject_var result;
    
    QEvaluator::evaluate(in, out, ok, message);
    if(!ok)
        return;
    
    DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[0]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    DvObject_var startTime=in[1]->connectedSlot()->getValue(ok);
    if(!ok)
    {
        message=QString("Input ")+in[1]->getName()+QString(" invalid");
        ok=false;
        return;
    }
    
    if( !startTime->is_time() ){
        message=QString("2nd Object did not resolve to a time");
        QplugAppendTextDisplay("Timetags to Sec Operator: input not Time object. \nTry adding Extractor.");
        ok=false;
        return;
    }
  
    result = startTime + val1;
    			 
    out[0]->setValue(result, true);
    
}

void QEvaluatorTIVLtoSec::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
	ok=true;
	DvObject_var result;
	
	QEvaluator::evaluate(in, out, ok, message);
	if(!ok)
		return;
	
	DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[0]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	
	DvEvent ti = val1->getTimeRange();
	result = new DvObject(ti.duration());
			
	DvString SI_Str = "1>s";
	result->change_xref(SI_CONVERSION, SI_Str);
			
	DvString unit_Str = "sec";
	result->change_xref(UNITS, unit_Str);

	DvString name_Str = "Seconds duration";
	result->change_xref(FIELDNAM, name_Str);

	out[0]->setValue(result, true);

}

void QEvaluatorTT2_TT1toSec::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
	
	ok=true;
	DvObject_var result;
	
	QEvaluator::evaluate(in, out, ok, message);
	if(!ok)
		return;
	
	DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[0]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	DvObject_var tt1 = val1->getTimeTags();
	
	DvObject_var val2=in[1]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[1]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	DvObject_var tt2 = val2->getTimeTags();
	
	if(tt1.is_nil() ||  tt1->seqSize() <= 0){
		message=QString("Could not get valid time tags from input 1");
		QplugAppendTextDisplay("Timetags Diff Operator: timetags not found");
		ok=false;
		return;
	}
	if(tt2.is_nil() ||  tt2->seqSize() <= 0){
		message=QString("Could not get valid time tags from input 2");
		QplugAppendTextDisplay("Timetags Diff Operator: timetags not found");
		ok=false;
		return;
	}
		
	if( tt1->seqSize() != tt2->seqSize())
	{
		message=QString("Timetag sequences of different lengths");
		QplugAppendTextDisplay("Timetags Diff Operator: sequences not same length");
		ok=false;
		return;

	}
	
	result = new DvObject(0.0, tt2->seqSize());
	for(int i=0; i<tt2->seqSize(); i++){
		result->dbl(i) = tt2->dbl(i) - tt1->dbl(i);
	}

	// get xrefs from inputs if requested
	result->copy_xrefs_from(tt1);
			
	DvString SI_Str = "1>s";
	result->change_xref(SI_CONVERSION, SI_Str);
			
	DvString unit_Str = "sec";
	result->change_xref(UNITS, unit_Str);

	DvString name_Str = "Time Diffs";
	result->change_xref(FIELDNAM, name_Str);
	
			 
	out[0]->setValue(result, true);

}

void QEvaluatorT2_T1toSec::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
	
	ok=true;
	DvObject_var result;
	
	QEvaluator::evaluate(in, out, ok, message);
	if(!ok)
		return;
	
	DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[0]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	
	DvObject_var val2=in[1]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[1]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	
	if( !val1->is_time() ){
		message=QString("Could not get valid time from input 1");
		QplugAppendTextDisplay("Time Diff Operator: time input not found");
		ok=false;
		return;
	}
	if(!val2->is_time() ){
		message=QString("Could not get valid time from input 2");
		QplugAppendTextDisplay("Timetags Operator: time input not found");
		ok=false;
		return;
	}
		
	double val = val2->time(0) - val1->time(0);

	result = new DvObject(val);
	result->copy_xrefs_from(val1);
			
	DvString SI_Str = "1>s";
	result->change_xref(SI_CONVERSION, SI_Str);
			
	DvString unit_Str = "sec";
	result->change_xref(UNITS, unit_Str);

	DvString name_Str = "Time Diff";
	result->change_xref(FIELDNAM, name_Str);

	out[0]->setValue(result, true);

}

// ----------------------

void QEvaluatorLinReg::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
	ok=true;
	DvObject_var fitLine;
	
	QEvaluator::evaluate(in, out, ok, message);
	if(!ok)
		return;
	
	DvObject_var val1=in[0]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[0]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	
	fitLine = val1->regression();
		
	if(fitLine.is_nil() )
	{
		message=QString("Object did not resolve to a scalar sequence");
		QplugAppendTextDisplay("Linear Regression: input not Scalar Sequence. \nTry adding Extractor.");
		ok=false;
		return;
	}
		
	out[0]->setValue(fitLine, true);

}

