#ifndef QCALC_EXTRACTION_ITEM_H
#define QCALC_EXTRACTION_ITEM_H

#include "qcalcCommon.h"
#include "qcalcCalcItem.h"
#include "qcalcSlots.h"

// Class that slices/resamples an input
class VExtractionItem: public VCalcItem
{
	public:
		// Basic initialisation. Registers itself to VCalcItem::currentItems
		VExtractionItem();
		
		// deletes its slots. Unregisters itself from VCalcItem::currentItems
		virtual ~VExtractionItem();
		
		// Returns the associated GUI, if it exists, or NULL
		virtual ExtractionItem* getGUI();
		
		// if there is a GUI associated, removes it from the scene, and deletes it
		void removeGUI();
		
		void setComp();
		
		virtual void reevaluateOutSlots();
		
		void updateSuffix();
		
		virtual QDomElement toXMLElement(QDomDocument&) const;

        virtual void fromXMLElement(QDomElement& el, QGraphicsScene* scene);
		inline virtual bool hasGui(){return gui!=NULL;}
		
		static int counter; // counter to find the right id for each instance
		int extraction_id;
		
		QString valueText;
		QString objName;
		
		// Parameters for automatic extraction
		bool askUser;
		QString Suffix;
		QString InputObjName;
		
	protected:
		ExtractionItem* gui;

};


class ExtractionItem: public CalcItem
{
	Q_OBJECT
	
	public:
		ExtractionItem(VExtractionItem* parentItem, VInSlot* input, VOutSlot* output, QGraphicsItem* parent=NULL);
		
		virtual ~ExtractionItem();
		
		virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget=NULL);

		inline virtual int type() const { return Type;}
		enum { Type = TYPE_EXTRACTIONITEM };
		VExtractionItem* parentItem;

		void waitForValidation();
        virtual QDomElement toXMLElement(QDomDocument& doc) const;
        virtual void fromXMLElement(QDomElement& el);

		QguWLinput* edit;
		virtual int page(){return parentItem->page;}
		CalcToolTip* calctooltip;

		
	public slots:
		
		void buttonPressed();
		void setValue();

	protected:
		void keyPressEvent(QKeyEvent * e );
		virtual QVariant itemChange(GraphicsItemChange change, const QVariant& value);
		bool okPressed;
		TimeInSlot* timeSlot;
};

#endif
