#ifndef QCALC_INPUT_ITEM_H
#define QCALC_INPUT_ITEM_H

#include "qcalcCommon.h"
#include "qcalcCalcItem.h"

// virtual input item. Only one virtual output slot
class VInputItem: public VCalcItem
{
	public:
		VInputItem();
		
		// Also removes the possible GUI
		virtual ~VInputItem();
		
		inline void setValue(DvObject_var value, const QString& text="")
		{
			this->not_joined_value= new DvObject(value);
            valueText=text;
			flagOutSlotsForReevaluation();
		}

        inline DvObject_var getNotJoinedValue()
        {
            return not_joined_value;
        }

        DvObject_var getDep0Tags();
        DvObject_var getInterval();
        DvObject_var getIntervalFromData();
				
		virtual void reevaluateOutSlots();
		
		virtual InputItem* getGUI();
		
		void removeGUI();
		
		int getAlgorithm() const {return joinSettings->algorithm;}
		double getWidth() const {return joinSettings->width;}
        int getGap() const {return joinSettings->gap;}
        int getEnd() const {return joinSettings->end;}
        int getMinBoxSize() const {return joinSettings->minBoxSize;}
        double getGapSize() const {return joinSettings->gapSize;}
		
		void setNoJoin(bool b){joinSettings->noJoin=b; flagOutSlotsForReevaluation();}
		void setAlgorithm(int i){joinSettings->algorithm=i;flagOutSlotsForReevaluation(); }
		void setWidth(double d) {joinSettings->width=d;flagOutSlotsForReevaluation();}
        void setGap(int i) {joinSettings->gap=i;flagOutSlotsForReevaluation();}
        void setEnd(int i) {joinSettings->end=i;flagOutSlotsForReevaluation();}
        void setGapSize(double d) {joinSettings->gapSize=d;flagOutSlotsForReevaluation();}
        void setMinBoxSize(int i) {joinSettings->minBoxSize=i;flagOutSlotsForReevaluation();}
		void setTags(DvObject_var tt);
		
        virtual QDomElement toXMLElement(QDomDocument&) const;

        virtual void fromXMLElement(QDomElement& el, QGraphicsScene* scene);

        QString valueText;
		
		static int counter;
		int input_id;
        inline virtual bool hasGui(){return gui!=NULL;}

	protected:
        DvObject_var value;
		InputItem* gui;

		JoinSettings* joinSettings;
        DvObject_var tags;
};


class InputItem: public CalcItem
{
	Q_OBJECT
	public:
		InputItem(VInputItem* parentItem, VOutSlot* output, QGraphicsItem* parent=NULL);
		
		virtual ~InputItem();
		
		virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget=NULL);
		
		inline virtual int type() const { return Type;}
		enum { Type = TYPE_INPUTITEM };
		VInputItem* parentItem;

        virtual QDomElement toXMLElement(QDomDocument& doc) const;

        virtual void fromXMLElement(QDomElement& el);
		virtual int page(){return parentItem->page;}
		void blockSig(){edit->blockSignals(true);}
		void allowSig(){edit->blockSignals(false);}

	protected:
		QguWLinput* edit;
		
		void dragEnterEvent(QGraphicsSceneDragDropEvent *);
		void dragLeaveEvent(QGraphicsSceneDragDropEvent *);
		void dropEvent(QGraphicsSceneDragDropEvent* event);
		virtual QVariant itemChange(GraphicsItemChange change, const QVariant& value);
	public slots:
		void setValue(const QString& text);
        void setValue();
		
	signals:
		void newInput(const QString&);
		
};


#endif
