#include "qcalcScene.h"
#include "qcalcOutputItem.h"
#include "qcalcInputItem.h"
#include "qcalcJoinItem.h"
#include <QGraphicsProxyWidget>
#include "qcalcTimeLine.h"
#include "DvRecord.h"


VOutputItem::VOutputItem():
		VCalcItem()
{
	output_id=counter++;
	VInSlot* slot=new VInSlot(this);
	inputs.push_back(slot);
	gui=NULL;
	name="Output";
	outputName=QString("result_%1").arg(output_id);

}

VOutputItem::~VOutputItem()
{
	if(gui!=NULL)
	{
		QGraphicsScene * scene=gui->scene();
		if(scene!=NULL)
		{
			scene->removeItem(gui);
		}
		gui->parentItem=NULL; // thus gui knows it is being deleted
		
		delete gui;
	}

}

QDomElement VOutputItem::toXMLElement(QDomDocument& document) const
{
	QDomElement tag=document.createElement("Output");
    if(gui!=NULL)
    {
		tag.appendChild(gui->toXMLElement(document));
    }

	tag.appendChild(createTextXMLNode(document, "XPosnInBlackBox", positionInBlackBox.x()));
	tag.appendChild(createTextXMLNode(document, "YPosnInBlackBox", positionInBlackBox.y()));

	tag.appendChild(createTextXMLNode(document, "ID", calc_id));
	tag.appendChild(createTextXMLNode(document, "Output_Name", outputName));
	tag.appendChild(createTextXMLNode(document, "Page", page));
    
    return tag;
}

void VOutputItem::fromXMLElement(QDomElement& el, QGraphicsScene* scene)
{
	VCalcItem::fromXMLElement(el, scene);


	QString outString="";
	getXMLNodeText(el, "Output_Name", outString);
	QDomElement guiElement=el.firstChildElement("GUI");
	if(!guiElement.isNull())
	{
		OutputItem* item=this->getGUI();
        scene->addItem(item);
        item->fromXMLElement(guiElement);
		item->setValue(outString);
	}
	else
	{
		setOutputName(outString);
	}

	QString val;
	
	if(getXMLNodeText(el, "XPosnInBlackBox", val))
	{
		positionInBlackBox.setX(val.toDouble());
	}
	if(getXMLNodeText(el, "YPosnInBlackBox", val))
	{
		positionInBlackBox.setY(val.toDouble());
	}

	if(getXMLNodeText(el, "Page", val))
	{
		page=val.toInt();
	}
	
}


void VOutputItem::reevaluateOutSlots()
{

}

bool VOutputItem::computeValue()
{
	// return is true if error
	VOutSlot* slot=inputs[0]->connectedSlot();
	if(slot==NULL)
	{
		if(gui!=NULL) gui->setErrMessage(true, "No connection");
		return true;
	}    
  
	// compute
	bool validity;
	DvObject_var res;

	res = slot->getValue(validity);
	
	if(validity && !outputName.isEmpty())
	{
		QwuiPutOnWL(res, QString(outputName));
		if(gui!=NULL) gui->setErrMessage(false);
		return false;
	}
	else
	{
		if(gui!=NULL) gui->setErrMessage(true, "Error when evaluating");
		return true;
	}
}

OutputItem* VOutputItem::getGUI()
{
	if(gui==NULL)
	{
		gui=new OutputItem(this, inputs[0]);
	}
	return gui;
}

void VOutputItem::removeGUI()
{
	if(gui!=NULL)
	{
		gui->scene()->removeItem(gui);
		delete gui;
		gui=NULL;
	}
}




OutputItem::OutputItem(VOutputItem* parentItem, VInSlot* input, QGraphicsItem* parent):
		CalcItem(parent)
{
	this->parentItem=parentItem;
	
	outlineColor=QColor(50, 75, 100);
	color1=QColor(150, 200, 255);
	color2=QColor(120, 170, 220);
	
	title=parentItem->getName();
	
	width=BASE_WIDTH;
	height=BASE_HEIGHT;
	setRect(-width/2., -height/2., width, height);
	
	edit=new QLineEdit;
	edit->setMaximumWidth((int)(BASE_WIDTH-6));
	edit->setMinimumWidth((int)(BASE_WIDTH-6));
	edit->setMaximumHeight((int)(HEADER_HEIGHT-2));
	edit->setMinimumHeight((int)(HEADER_HEIGHT-2));
	
	QGraphicsProxyWidget * editProxy=new QGraphicsProxyWidget(this);
	editProxy->setWidget(edit);
	editProxy->moveBy(-edit->width()/2., 0.);
	
	InSlot* slot=new InSlot(this, input);
	slot->setPos(0., -height/2.);
	inSlots.push_back(slot);
	
	
	rzButton->setPos(-width/2.+1.5*HEADER_HEIGHT, -height/2.+HEADER_HEIGHT/2.);
	closeButton->setPos(width/2.-HEADER_HEIGHT/2., -height/2.+HEADER_HEIGHT/2.);
	errButton->setPos(width/2.-1.5*HEADER_HEIGHT, -height/2.+HEADER_HEIGHT/2.);
	
	connect(edit, SIGNAL(textChanged(const QString&)), this, SLOT(setValue()));
	
	setAcceptDrops(false);
	edit->setAcceptDrops(false);
	
	edit->setText(parentItem->outputName);

}

OutputItem::~OutputItem()
{
	foreach(InSlot* slot, inSlots)
	{
		delete slot;
	}
	foreach(OutSlot* slot, outSlots)
	{
		delete slot;
	}
}

QVariant OutputItem::itemChange(GraphicsItemChange change, const QVariant& value)
{
	if (change == ItemPositionChange && scene()) {
		QPointF newPos = value.toPointF();
		
		updateConnectorsPositions();
		
		return newPos;
	}
	else if(change==ItemSelectedChange && scene())
	{
		if(value.toBool())
		{
			setZValue(1.);
		}
		else setZValue(0.);
		parentItem->setSelected(value.toBool());
	}
	return QGraphicsItem::itemChange(change, value);
}

void OutputItem::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget)
{
	CalcItem::paint(painter, option, widget);
}

void OutputItem::setValue()
{
	parentItem->setOutputName(edit->text());
}

void OutputItem::setValue(const QString& text)
{
	edit->setText(text);
	// will emit text changed which calls setValue() above
}

void OutputItem::updateConnectorsPositions()
{
	CalcItem::updateConnectorsPositions();
}

QDomElement OutputItem::toXMLElement(QDomDocument& doc) const
{
	return CalcItem::toXMLElement(doc);
}

void OutputItem::fromXMLElement(QDomElement& el)
{
	CalcItem::fromXMLElement(el);
}




