#include "qcalcTimeLine.h"

TimeLine::TimeLine()
{
	fromData=true;
    regularTT=false;
    fromObjString="";
    TTString="";
}

TimeLineHandler::TimeLineHandler():
	QObject()
{
	TimeLine* defaultTimeLine=new TimeLine;
	names.clear();
	timeLines.clear();
	names.push_back("Target");
	timeLines.push_back(defaultTimeLine);
	
}

TimeLineHandler::~TimeLineHandler()
{
	for(int i=0; i<timeLines.size(); ++i)
	{
		delete timeLines[i];
	}
	timeLines.clear();
	names.clear();
}

TimeLine* TimeLineHandler::getTimeLine(const QString& name)
{
	int index=names.indexOf(name);
	if(index<0)
		return NULL;
	
	return timeLines[index];
}

bool TimeLineHandler::deleteTimeLine(const QString& name)
{
	int index=names.indexOf(name);
	if(index<=0)
		return false;

	names.remove(index);
	TimeLine* tl=timeLines[index];
	timeLines.remove(index);
	delete tl;
	emit deletedTimeLine(name);
	return true;
}

bool TimeLineHandler::addTimeLine(const QString& name, bool fromData, bool regularTT, const QString& fromObjString, const QString TTString)
{
	int index=names.indexOf(name);
	if(index>=0)
		return false;
	
	TimeLine* tl=new TimeLine;
	tl->fromData=fromData;
	tl->regularTT=regularTT;
	tl->fromObjString=fromObjString;
	tl->TTString=TTString;
	
	names.push_back(name);
	timeLines.push_back(tl);
	
	emit addedTimeLine(name);
	
	return true;
}

QList<QString> TimeLineHandler::getNames()
{
	QList<QString> res;
	res.clear();
	for(int i=0; i<names.size(); ++i)
	{
		res.push_back(names[i]);
	}
	return res;
}

QDomElement TimeLineHandler::toXMLElement(QDomDocument& document)
{
	QDomElement tag=document.createElement("TimeLines");
	for(int i=0; i<names.size(); i++){
		QString timeLineName = names[i];
		
		TimeLine* tl = this->getTimeLine(timeLineName);
		if(tl != NULL)
		{
			QDomElement tlElement=document.createElement("Time_Line");
			tag.appendChild(tlElement);
			tlElement.appendChild(createTextXMLNode(document, "Name", timeLineName));
			tlElement.appendChild(createTextXMLNode(document, "From_Data", tl->fromData));
			tlElement.appendChild(createTextXMLNode(document, "Regular_Time_Tags", tl->regularTT));
			tlElement.appendChild(createTextXMLNode(document, "From_Obj", tl->fromObjString));
			tlElement.appendChild(createTextXMLNode(document, "Time_Tags", tl->TTString));
		}
	}
    
    return tag;
}

void TimeLineHandler::fromXMLElement(QDomElement& el)
{
	names.clear();
	timeLines.clear();
	
	QDomElement tlines=el.firstChildElement("TimeLines");
	if(!tlines.isNull())
	{
		QString val;

		QString name="Target";
		QDomElement tlElement=tlines.firstChildElement("Time_Line");
		if(!tlElement.isNull())
		{
			bool fromData=true;
			bool regularTT=true;
			QString fromObjString="";
			QString TTString="";
			if(getXMLNodeText(tlElement, "Name", val))
			{
				if(val == "Default") name="Target"; // remove old name for 1st timeline
                else name=val;
			}
			if(getXMLNodeText(tlElement, "From_Data", val))
			{
				fromData=val.toInt();
			}
			if(getXMLNodeText(tlElement, "Regular_Time_Tags", val))
			{
				regularTT=val.toInt();
			}
			if(getXMLNodeText(tlElement, "From_Obj", val))
			{
				fromObjString=val;
			}
			if(getXMLNodeText(tlElement, "Time_Tags", val))
			{
				TTString=val;
			}
				
			this->addTimeLine(name, fromData, regularTT, fromObjString, TTString);
			
			tlElement=tlElement.nextSiblingElement("Time_Line");
		}
	}
	if(timeLines.size() == 0) {
		this->addTimeLine("Target");
	}

}

