// -------------Derivative Window class ---------------------------------


#include "qpsd.h"
#include "qrtn.h"

extern qwuiWindow* qsasWindow;

long long Facet3D::currentStatus=1;
long long Facet4D::currentStatus=1;
long long Point4D::next_p_ID=1; // used to identify each point

PSD_Window::PSD_Window() 
	: QMainWindow() 
{

	if(QuSlotDebug())
	{
		printf("PSD_Window::PSD_Window \n");
		fflush(stdout);
	}
  	this->setWindowTitle("PSD tool");
	
  	running = false;
	
  	QMenu * help = new QMenu( this );
	this->menuBar()->addMenu(help );
	help->addAction("Particle Tools", this, SLOT( launchHelp() ));
	help->setTitle(" Help ");
	
	this->setPalette(qApp->palette());

	QFontMetrics qgu_font_mt(qApp->font());
	int char_width = qgu_font_mt.width("n");
   
//	setAttribute(Qt::WA_DeleteOnClose);
     
	QGroupBox * mainWidget=new QGroupBox(this);
	QVBoxLayout * mainLayout=new QVBoxLayout;
	
	// Data slot
	
	QGroupBox * slotPanel=new QGroupBox(mainWidget);
	QGridLayout * slotLayout=new QGridLayout;
	slotPanel->setTitle("Phase Space Distribution");
	
	objField = new QguWLinput(slotPanel,"objField");
	objField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_ANY, QS_ANY, QS_ANY));
	objField->setEnabled(true);
	objField->setMinimumWidth(char_width*80);
	connect(objField, SIGNAL(edited_OK()), this, SLOT(setNewName()));
	slotLayout->addWidget(objField,  0, 0, 1, 4, Qt::AlignLeft);
	
	parityLabel = new QLabel("PSD coordinates in:");
	slotLayout->addWidget( parityLabel, 1, 0, Qt::AlignRight );
	 
	Vparity = new QComboBox(slotPanel);
	Vparity->addItem("Look Direction");
	Vparity->addItem("Flow Direction" );
	slotLayout->addWidget( Vparity, 1, 1, Qt::AlignLeft );
	
	QLabel *pad1 = new QLabel("   ");
	slotLayout->addWidget( pad1, 1, 2, Qt::AlignRight );
	QLabel *pad2 = new QLabel("   ");
	slotLayout->addWidget( pad2, 1, 3, Qt::AlignRight );

	massChoiceLabl = new QLabel("Species:");
	slotLayout->addWidget( massChoiceLabl, 2, 0, Qt::AlignRight );
	
	massChoice = new QComboBox(slotPanel);
	massChoice->addItem("Electron");
	massChoice->addItem("Proton");
	massChoice->addItem("He+");
	massChoice->addItem("He++");
	massChoice->addItem("O+");
	massChoice->addItem("Set Manually");
	
	slotLayout->addWidget( massChoice, 2, 1, Qt::AlignLeft );
	connect( massChoice,  SIGNAL( currentIndexChanged(const QString &) ), this, SLOT( setMassChoice(const QString &) ) );
	massChoice->setToolTip("Species or mass number)");
	
	massLabl = new QLabel("Atomic Mass Number");
	slotLayout->addWidget( massLabl, 2, 4, Qt::AlignLeft );
	massField = new FloatSpinBox(0.0, 50., 7, 1., NULL, "massField");
	massField->setToolTip("Atomic mass number ( p+n, 0 for electrons)");
	slotLayout->addWidget( massField, 2, 2, 1, 2, Qt::AlignRight );

	chargeLabl = new QLabel("Particle Charge");
	slotLayout->addWidget( chargeLabl, 1, 4, Qt::AlignLeft );
	chargeField = new QSpinBox();
	chargeField->setMinimum(-9);  //SJS allowed negative
	chargeField->setMaximum(9);
	chargeField->setValue(1);
	chargeField->setToolTip("Charge number of particle"); //SJS
	slotLayout->addWidget( chargeField, 1, 2, 1, 2, Qt::AlignRight );

	massChoice->setCurrentIndex(5);


	slotPanel->setLayout(slotLayout);
	mainLayout->addWidget(slotPanel);
	
	// Reference Frame transformation panel
	
	QGroupBox * gse_panel=new QGroupBox(mainWidget);
	QGridLayout *gse_layout = new QGridLayout;
	gse_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	gse_layout->setMargin(QSAS_profiles->current()->GUImargins);	
	gse_panel->setTitle("Conversion from Instrument to Science coordinates");
	
	transLab = new QLabel("Source: ");
	gse_layout->addWidget( transLab, 0, 0, Qt::AlignRight );

	transMethod = new QComboBox(gse_panel);
	transMethod->addItem("Cluster PEACE ( SR2 )");
	transMethod->addItem("Cluster CIS ( ISR2 )" );
	transMethod->addItem("Other (use rotation matrix)" );
	gse_layout->addWidget( transMethod, 0, 1, Qt::AlignLeft );

	QLabel * targetLab = new QLabel("To:");
	gse_layout->addWidget( targetLab, 0, 2, Qt::AlignRight );
 	transTarget = new QComboBox(gse_panel);
	transTarget->addItem("GSE");
	transTarget->addItem("GSM" );
	transTarget->addItem("From Rotn Matrix" );
	transTarget->addItem("Do Not Rotate" );
	gse_layout->addWidget( transTarget, 0, 3, Qt::AlignLeft );

	methodLabl = new QLabel("  Caution, using PSD in supplied frame");
	gse_layout->addWidget( methodLabl, 0, 0, Qt::AlignLeft );
	methodLabl->hide();

	
	transMethod->adjustSize();
	int comboWidth = transMethod->width()-4;
	
	connect( transMethod,  SIGNAL( currentIndexChanged(const QString &) ), this, SLOT( setMethod(const QString &) ) );
	connect( transTarget,  SIGNAL( currentIndexChanged(const QString &) ), this, SLOT( setTarget(const QString &) ) );
	
	latLabl = new QLabel("SC spin axis Latitude: ");
	latField = new QguWLinput(gse_panel,"latField");
	latField->setToolTip(" in either GSE or GSM ");

	latField->setMinimumWidth(comboWidth);
	gse_layout->addWidget( latField, 1, 1, Qt::AlignLeft );
	gse_layout->addWidget( latLabl, 1, 0, Qt::AlignRight );
	latField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_ANY, QS_ANY));

	longLabl = new QLabel("SC spin axis Longitude: ");
	longField = new QguWLinput(gse_panel,"longField");
	longField->setToolTip(" in either GSE or GSM ");
	gse_layout->addWidget( longField, 2, 1, Qt::AlignLeft );
	gse_layout->addWidget( longLabl, 2, 0, Qt::AlignRight );
	longField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_ANY, QS_ANY));
	longField->setMinimumWidth(comboWidth);

	rotnLabl = new QLabel("Rotation Matrix: ");
	rotnField = new QguWLinput(gse_panel,"rotnField");
	gse_layout->addWidget( rotnField, 3, 1, Qt::AlignLeft );
	gse_layout->addWidget( rotnLabl, 3, 0, Qt::AlignRight );
	rotnField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_2D, QS_ANY, QS_ANY));
	rotnField->setMinimumWidth(comboWidth);
	rotnLabl->hide();
	rotnField->hide();
	
	gse_panel->setLayout(gse_layout);
	mainLayout->addWidget(gse_panel);

	// process panel
	
	QGroupBox * process_panel=new QGroupBox(mainWidget);
	QGridLayout * process_layout=new QGridLayout;
	process_panel->setTitle("Process PSD");
	
	processMethod = new QComboBox(process_panel);
	processMethod->addItem("Moments of PSD");
	processMethod->addItem("Simple Slice PSD");
	processMethod->addItem("Advanced Slice PSD");
	processMethod->addItem("Pitch Angle Distribution");
	processMethod->addItem("Save Rotation Matrix");
	process_layout->addWidget( processMethod, 0, 0, 1, 2, Qt::AlignLeft );
	connect( processMethod,  SIGNAL( currentIndexChanged(const QString &) ), this, SLOT( setProcess(const QString &) ) );
	
	// Slice Inputs
	
	centreChoiceLabl = new QLabel("Slice Centre:");
	process_layout->addWidget( centreChoiceLabl, 1, 0, Qt::AlignRight );
	centreChoiceLabl->hide();
	
	centreChoice = new QComboBox(process_panel);
	centreChoice->addItem("Use PSD Centre");
	centreChoice->addItem("Centre Velocity");
	process_layout->addWidget( centreChoice, 1, 1, Qt::AlignLeft );
	connect( centreChoice,  SIGNAL( currentIndexChanged(const QString &) ), this, SLOT( setCentreChoice(const QString &) ) );
	centreChoice->hide();
	
	centreLabl = new QLabel("(Typ Bulk Flow Vel)");
	process_layout->addWidget( centreLabl, 1, 4, Qt::AlignLeft );
	centreLabl->hide();
	centreField = new QguWLinput(process_panel,"centreField");
	centreField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_NOT_SEQ, QS_ANY));
	process_layout->addWidget( centreField, 1, 2, 1, 2, Qt::AlignRight );
	centreField->setEnabled(false);
	centreField->hide();

	dirnChoiceLabl = new QLabel("Slice Normal:");
	process_layout->addWidget( dirnChoiceLabl, 2, 0, Qt::AlignRight );
	dirnChoiceLabl->hide();

	dirnChoice = new QComboBox(process_panel);
	dirnChoice->addItem("Z axis ");
	dirnChoice->addItem("Y axis ");
	dirnChoice->addItem("X axis ");
	dirnChoice->addItem("Along vector");
	process_layout->addWidget( dirnChoice, 2, 1, Qt::AlignLeft );
	connect( dirnChoice,  SIGNAL( currentIndexChanged(const QString &) ), this, SLOT( setDirnChoice(const QString &) ) );
	dirnChoice->hide();

	dirnLabl = new QLabel("Slice Normal vector");
	process_layout->addWidget( dirnLabl, 2, 4, Qt::AlignLeft );
	dirnLabl->hide();
	dirnField = new QguWLinput(process_panel,"dirnField");
	dirnField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_NOT_SEQ, QS_ANY));
	dirnField->setToolTip("Slice Normal Direction: (eg B field, normalisation not necessary) ");
	process_layout->addWidget( dirnField, 2, 2, 1, 2, Qt::AlignRight );
	dirnField->setEnabled(false);
	dirnField->hide();
	
	dirnAdvChoiceLabl = new QLabel("Vectors V1,V2 defining slice");
	process_layout->addWidget( dirnAdvChoiceLabl, 2, 0, Qt::AlignRight );
	dirnAdvChoiceLabl->hide();

	dirnAdvChoice = new QComboBox(process_panel);
	dirnAdvChoice->addItem("n||V1; V2 in slice with V2.X>0");
	dirnAdvChoice->addItem("n=V1xV2; X||V1");
	process_layout->addWidget( dirnAdvChoice, 2, 1, Qt::AlignLeft );
	dirnAdvChoice->hide();

	dirnAvecLabl = new QLabel("V1-vector");
	process_layout->addWidget( dirnAvecLabl, 2, 4, Qt::AlignLeft );
	dirnAvecLabl->hide();
	
	dirnAvecField = new QguWLinput(process_panel,"dirnAvecField");
	dirnAvecField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_NOT_SEQ, QS_ANY));
	dirnAvecField->setToolTip("Vector V1");
	process_layout->addWidget( dirnAvecField, 2, 2, 1, 2, Qt::AlignRight );
	dirnAvecField->setEnabled(false);
	dirnAvecField->hide();

	dirnBvecLabl = new QLabel("V2-vector");
	process_layout->addWidget( dirnBvecLabl, 3, 4, Qt::AlignLeft );
	dirnBvecLabl->hide();

	dirnBvecField = new QguWLinput(process_panel,"dirnBvecField");
	dirnBvecField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_NOT_SEQ, QS_ANY));
	dirnBvecField->setToolTip("Vector V2");
	process_layout->addWidget( dirnBvecField, 3, 2, 1, 2, Qt::AlignRight );
	dirnBvecField->setEnabled(false);
	dirnBvecField->hide();
	
//	***** end adv slice normal
	
	binsLabl = new QLabel("Velocity bins: ");
	process_layout->addWidget( binsLabl, 3, 0, Qt::AlignRight );
	binsLabl->hide();
	
	numberBins = new QSpinBox();
	numberBins->setMinimum(16);
	numberBins->setMaximum(200);
	numberBins->setValue(70);
	numberBins->setToolTip("Number of bins on velocity axes");
	process_layout->addWidget( numberBins, 3, 1, Qt::AlignLeft );
	numberBins->hide();
	
	
	// special PAD inputs - note re-use some from slice, relabelled in setProcess
	numberPADBins = new QSpinBox();
	numberPADBins->setMinimum(2);
	numberPADBins->setMaximum(100);
	numberPADBins->setValue(16);
	numberPADBins->setToolTip("Number of pitch angle bins");
	process_layout->addWidget( numberPADBins, 3, 3, Qt::AlignLeft );
	numberPADBins->hide();
	
	PADbinsLabl = new QLabel("Pitch Angle bins: ");
	process_layout->addWidget( PADbinsLabl, 3, 2, Qt::AlignRight );
	PADbinsLabl->hide();
	
	numberPADphiBins = new QSpinBox();
	numberPADphiBins->setMinimum(2);
	numberPADphiBins->setMaximum(100);
	numberPADphiBins->setValue(16);
	numberPADphiBins->setToolTip("Num. bins for phi average");
	process_layout->addWidget( numberPADphiBins, 3, 5, Qt::AlignLeft );
	numberPADphiBins->hide();
	
	PADphibinsLabl = new QLabel("Num bins phi avg");
	process_layout->addWidget( PADphibinsLabl, 3, 4, Qt::AlignRight );
	PADphibinsLabl->hide();	

	// Moments inputs
	
	integrationLabl = new QLabel("Integration Method:");
	process_layout->addWidget( integrationLabl, 1, 0, Qt::AlignRight );
	integrationMethod = new QComboBox(process_panel);
	integrationMethod->addItem("Sum over data bins");
	integrationMethod->addItem("Phase-space tetrahedra");
	process_layout->addWidget( integrationMethod, 1, 1, Qt::AlignLeft );
	
	dBtn = new QCheckBox("Density:", this);
	dBtn->setChecked(true);
	process_layout->addWidget( dBtn, 3, 0, Qt::AlignLeft );

	vBtn = new QCheckBox("Velocity:", this);
	process_layout->addWidget( vBtn, 3, 1, Qt::AlignLeft );

	pBtn = new QCheckBox("Pressure (&Temp):", this);
	process_layout->addWidget( pBtn, 3, 2, Qt::AlignLeft );

	hBtn = new QCheckBox("Heat Flux:", this);
	process_layout->addWidget( hBtn, 3, 3, Qt::AlignLeft );
	
	allBtn = new QCheckBox("All:", this);
	process_layout->addWidget( allBtn, 3, 4, Qt::AlignLeft );
	allBtn->setChecked(false);
	connect( allBtn,  SIGNAL( stateChanged(int) ), this, SLOT( setMomentsAll(int) ) );
	
	process_panel->setLayout(process_layout);
	mainLayout->addWidget(process_panel);

	// Spacecraft potential correction and energy threshold
	
	QGroupBox * potential_panel=new QGroupBox(mainWidget);
	QGridLayout * potential_layout=new QGridLayout;
	potential_panel->setTitle("Spacecraft Potential and Energy Thresholding");

	potentialChoiceLabl = new QLabel("Spacecraft Potential:");
	potential_layout->addWidget( potentialChoiceLabl, 1, 0, Qt::AlignRight );
	
	potentialChoice = new QComboBox(process_panel);
	potentialChoice->addItem("No Correction");
	potentialChoice->addItem("value in V");
	potentialChoice->addItem("object w/ SI_conversion");
	potential_layout->addWidget( potentialChoice, 1, 1, Qt::AlignLeft );
	connect( potentialChoice,  SIGNAL( currentIndexChanged(const QString &) ), this, SLOT( setPotentialChoice(const QString &) ) );
	potentialChoice->setToolTip("Spacecraft potential for energy correction");
	
	potentialLabl = new QLabel("S/C Potential ");
	potential_layout->addWidget( potentialLabl, 1, 4, Qt::AlignLeft );
	potentialField = new QguWLinput(process_panel,"potentialField");
	potentialField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));
	potential_layout->addWidget( potentialField, 1, 2, 1, 2, Qt::AlignLeft );
	potentialField->setEnabled(false);

	thresholdChoiceLabl = new QLabel("Energy Threshold:");
	potential_layout->addWidget( thresholdChoiceLabl, 2, 0, Qt::AlignRight );
	
	thresholdChoice = new QComboBox(process_panel);
	thresholdChoice->addItem("No Threshold");
	thresholdChoice->addItem("value in eV");
	thresholdChoice->addItem("object w/ SI_conversion");
	potential_layout->addWidget( thresholdChoice, 2, 1, Qt::AlignLeft );
	connect( thresholdChoice,  SIGNAL( currentIndexChanged(const QString &) ), this, SLOT( setThresholdChoice(const QString &) ) );
	thresholdChoice->setToolTip("Threshold under which energies are not taken into account");
	
	thresholdLabl = new QLabel("Alt: use data pulldown");
	potential_layout->addWidget( thresholdLabl, 2, 4, Qt::AlignLeft );
	thresholdField = new QguWLinput(process_panel,"thresholdField");
	thresholdField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));
	potential_layout->addWidget( thresholdField, 2, 2, 1, 2, Qt::AlignLeft );
	thresholdField->setToolTip("Alternatively, use pull-down->Time and array slice");
	thresholdField->setEnabled(false);

	potential_panel->setLayout(potential_layout);
	mainLayout->addWidget(potential_panel);
	
// "Save as" panel
	
	QGroupBox * name_panel=new QGroupBox(mainWidget);
	QHBoxLayout * name_layout=new QHBoxLayout;
	
	QLabel *name_label = new QLabel("Name stem for save ", name_panel);
	name_layout->addWidget(name_label);
	
	name_field = new QguWLinput(name_panel, "name_field");
	name_field->setAcceptable(Qcondition(QS_NARROW, QS_TEXT, QS_0D, QS_NOT_SEQ, QS_ANY));
	name_field->setEnabled(true);
	name_field->setMinimumWidth(char_width*45);
	userSet = false;
	connect( name_field,  SIGNAL( edited_OK() ), this, SLOT( setUserInput() ) );
	name_layout->addWidget(name_field);
	
	name_panel->setLayout(name_layout);
	mainLayout->addWidget(name_panel);
	
	// Buttons Layout
	
	QHBoxLayout * buttons_layout=new QHBoxLayout;
	
	saveBtn = new QPushButton( "Evaluate");
	saveBtn->setMinimumSize(saveBtn->sizeHint());
	connect( saveBtn, SIGNAL(clicked()), this, SLOT(evaluate()) );
	saveBtn->setFocusPolicy(Qt::StrongFocus);
	buttons_layout->addWidget(saveBtn);
	saveBtn->setDefault(true);

	percentage = new QLabel("    ");
	buttons_layout->addWidget(percentage);
	
	dismissBtn = new QPushButton( "Dismiss");
	dismissBtn->setMinimumSize(dismissBtn->sizeHint());
	connect( dismissBtn, SIGNAL(clicked()), this, SLOT(close()) );
	buttons_layout->addWidget(dismissBtn);
	
	mainLayout->addLayout(buttons_layout);	
	
	mainWidget->setLayout(mainLayout);
	setCentralWidget(mainWidget);
	
//	this->setMinimumSize(this->sizeHint());

}

PSD_Window::~PSD_Window() 
{
}
// -------------


void PSD_Window::setEvalBtn(QString txt) 
{
	saveBtn->setText(txt);
}
// -------------

void PSD_Window::launchHelp() 
{
	 if(QuSlotDebug())
 	{
		printf("PSD_Window::launchHelp \n");
		fflush(stdout);
	 }
 
  	qsasWindow->help_window->manEntry("QSAS_analysis.html","#particles");

}
// -------------

void PSD_Window::setUserInput(){
  if(name_field->lineBox->hasFocus()){
	
     if(name_field->lineBox->text().isEmpty() ) userSet = false; // allow refresh from drop to object slot
	 else userSet = true; // user typed input, disable update from data slot
  }
  // else ignore
}

// -------------

void PSD_Window::setMethod(const QString & text){
    
	if( text.contains("PEACE") ) {
		latField->show();
		longField->show();
		rotnField->hide();
		latLabl->show();
		longLabl->show();
		rotnLabl->hide();
		methodLabl->hide();
	}
	else if( text.contains("CIS") ) {
		latField->show();
		longField->show();
		rotnField->hide();
		latLabl->show();
		longLabl->show();
		rotnLabl->hide();
		methodLabl->hide();

	}
	else {
		latField->hide();
		longField->hide();
		rotnField->show();
		latLabl->hide();
		longLabl->hide();
		rotnLabl->show();		
		methodLabl->hide();
	}
}

// -------------

void PSD_Window::setTarget(const QString & text){
    
	if( text.contains("GSE") ) {
		transLab->show();
		transMethod->show();
		methodLabl->hide();
		setMethod(transMethod->currentText());
	}
	else if( text.contains("GSM") ) {
		transLab->show();
		transMethod->show();
		methodLabl->hide();
		setMethod(transMethod->currentText());
	}
	else if( text.contains("Rotn") ) {
		transLab->show();
		transMethod->show();
		methodLabl->hide();
		transMethod->setCurrentIndex(2);
		rotnField->show();
		rotnLabl->show();		
	}
	else  {
		transMethod->hide();
		transLab->hide();
		methodLabl->show();
		latField->hide();
		longField->hide();
		rotnField->hide();
		latLabl->hide();
		longLabl->hide();
		rotnLabl->hide();
	}
	
}

// -------------

void PSD_Window::setProcess(const QString & text){
    
	if( text.contains("Moments") ) {
		centreField->hide();
		centreChoice->hide();
		centreChoiceLabl->hide();
		centreLabl->hide();
		dirnField->hide();
		dirnChoice->hide();
		dirnChoiceLabl->hide();
		dirnLabl->hide();
		dirnAvecLabl->hide();
		dirnBvecLabl->hide();
		dirnAvecField->hide();
		dirnBvecField->hide();
		dirnAdvChoice->hide();
		dirnAdvChoiceLabl->hide();
		binsLabl->hide();
		numberBins->hide();
		numberPADBins->hide();
		PADbinsLabl->hide();
		numberPADphiBins->hide();
		PADphibinsLabl->hide();
		potentialField->show();
		potentialChoice->show();
		potentialChoiceLabl->show();
		potentialLabl->show();		
		thresholdField->show();
		thresholdChoice->show();
		thresholdChoiceLabl->show();
		thresholdLabl->show();	
		integrationMethod->show();
		integrationLabl->show();
		massField->show();
		chargeField->show();
		massChoice->show();		
		massChoiceLabl->show();
		massLabl->show();	
		chargeLabl->show();	
		objField->show();
		parityLabel->show();
		Vparity->show();
		allBtn->show();
		dBtn->show();
		vBtn->show();
		pBtn->show();
		hBtn->show();
		if(!userSet) setObjectName();

	}
	else if( text.contains("Simple") ) {
		centreField->show();
		centreChoice->show();
		centreChoiceLabl->setText("Slice Centre:");  // PAD re-uses and resets
		centreChoiceLabl->show();
		centreLabl->show();
		dirnField->show();
		dirnChoice->show();
		dirnChoiceLabl->setText("Slice Normal");
		dirnChoiceLabl->show();
		dirnLabl->setText("Slice Normal vector");
		dirnLabl->show();
		dirnAvecLabl->hide();
		dirnBvecLabl->hide();
		dirnAvecField->hide();
		dirnBvecField->hide();
		dirnAdvChoice->hide();
		dirnAdvChoiceLabl->hide();
		binsLabl->setText("Num Velocity bins");
		binsLabl->show();
		numberBins->setToolTip("Number of bins on velocity axes");
		numberBins->show();	
		numberPADBins->hide();
		PADbinsLabl->hide();
		numberPADphiBins->hide();
		PADphibinsLabl->hide();
		potentialField->show();
		potentialChoice->show();
		potentialChoiceLabl->show();
		potentialLabl->show();			
		thresholdField->show();
		thresholdChoice->show();
		thresholdChoiceLabl->show();
		thresholdLabl->show();		
		integrationMethod->hide();
		integrationLabl->hide();
		massField->show();
		chargeField->show();
		massChoice->show();		
		massChoiceLabl->show();
		massLabl->show();
		chargeLabl->show();
		objField->show();
		parityLabel->show();
		Vparity->show();
		allBtn->hide();
		dBtn->hide();
		vBtn->hide();
		pBtn->hide();
		hBtn->hide();
		if(!userSet) setObjectName();
	}
		else if( text.contains("Advanced") ) {
		centreField->show();
		centreChoice->show();
		centreChoiceLabl->setText("Slice Centre:");  // PAD re-uses and resets
		centreChoiceLabl->show();
		centreLabl->show();
		dirnField->hide();
		dirnChoice->hide();
		dirnChoiceLabl->hide();
		dirnLabl->hide();
		dirnAvecLabl->show();
		dirnBvecLabl->show();
		dirnAvecField->show();
		dirnBvecField->show();
		dirnAvecField->setEnabled(true);
		dirnBvecField->setEnabled(true);
		dirnAdvChoice->show();
		dirnAdvChoiceLabl->show();
		binsLabl->setText("Num V bins - log-spaced");
		binsLabl->show();
		numberBins->setToolTip("Number of bins on velocity axes");
		numberBins->show();	
		numberPADBins->hide();
		PADbinsLabl->hide();
		numberPADphiBins->hide();
		PADphibinsLabl->hide();
		potentialField->show();
		potentialChoice->show();
		potentialChoiceLabl->show();
		potentialLabl->show();			
		thresholdField->show();
		thresholdChoice->show();
		thresholdChoiceLabl->show();
		thresholdLabl->show();		
		integrationMethod->hide();
		integrationLabl->hide();
		massField->show();
		chargeField->show();
		massChoice->show();		
		massChoiceLabl->show();
		massLabl->show();
		chargeLabl->show();
		objField->show();
		parityLabel->show();
		Vparity->show();
		allBtn->hide();
		dBtn->hide();
		vBtn->hide();
		pBtn->hide();
		hBtn->hide();
		if(!userSet) setObjectName();
	}
	else if( text.contains("Pitch") ) {
		centreField->show();
		centreChoice->show();
		centreChoiceLabl->setText("Origin for PAD");
		centreChoiceLabl->show();
		centreLabl->show();
		dirnField->show();
		dirnChoice->show();
		dirnChoiceLabl->setText("Field Direction");   // re-use from slice
		dirnChoiceLabl->show();
		dirnLabl->setText("Field Direction in science frame - eg GSE");
		dirnLabl->show();
		dirnAvecLabl->hide();
		dirnBvecLabl->hide();
		dirnAvecField->hide();
		dirnBvecField->hide();
		dirnAdvChoice->hide();
		dirnAdvChoiceLabl->hide();
		binsLabl->setText("Num E-bins (log-spaced)");
		binsLabl->show();
		numberBins->setToolTip("Number of Energy bins");
		numberBins->show();
		numberPADBins->show();
		PADbinsLabl->show();
		numberPADphiBins->show();
		PADphibinsLabl->show();		
		potentialField->show();
		potentialChoice->show();
		potentialChoiceLabl->show();
		potentialLabl->show();			
		thresholdField->show();
		thresholdChoice->show();
		thresholdChoiceLabl->show();
		thresholdLabl->show();		
		integrationMethod->hide();
		integrationLabl->hide();
		massField->show();  // I think particle info may be needed?
		chargeField->show();
		massChoice->show();		
		massChoiceLabl->show();
		massLabl->show();
		chargeLabl->show();
		objField->show();
		parityLabel->show();
		Vparity->show();
		allBtn->hide();
		dBtn->hide();
		vBtn->hide();
		pBtn->hide();
		hBtn->hide();
		if(!userSet) setObjectName();
	}
	else if( text.contains("Rotation Matrix") ){
		centreField->hide();
		centreChoice->hide();
		centreChoiceLabl->hide();
		centreLabl->hide();
		dirnField->hide();
		dirnChoice->hide();
		dirnChoiceLabl->hide();
		dirnLabl->hide();
		dirnAvecLabl->hide();
		dirnBvecLabl->hide();
		dirnAvecField->hide();
		dirnBvecField->hide();
		dirnAdvChoice->hide();
		dirnAdvChoiceLabl->hide();
		binsLabl->hide();
		numberBins->hide();
		numberPADBins->hide();
		PADbinsLabl->hide();
		numberPADphiBins->hide();
		PADphibinsLabl->hide();		
		potentialField->hide();
		potentialChoice->hide();
		potentialChoiceLabl->hide();
		potentialLabl->hide();			
		thresholdField->hide();
		thresholdChoice->hide();
		thresholdChoiceLabl->hide();
		thresholdLabl->hide();			
		integrationMethod->hide();
		integrationLabl->hide();
		massField->hide();
		chargeField->hide();
		massChoice->hide();		
		massChoiceLabl->hide();
		massLabl->hide();
		chargeLabl->hide();
		objField->hide();
		parityLabel->hide();
		Vparity->hide();
		allBtn->hide();
		dBtn->hide();
		vBtn->hide();
		pBtn->hide();
		hBtn->hide();
		if(!userSet) name_field->setText("Rxyz");
	}
}

// -------------

void PSD_Window::setDirnChoice(const QString & text){
    
	if( text.contains("Z") ) {
		dirnField->setEnabled(false);
	}
	else if( text.contains("Y") ) {
		dirnField->setEnabled(false);
	}
	else if( text.contains("X") ) {
		dirnField->setEnabled(false);
	}
	else {
		dirnField->setEnabled(true);
	}
}

// -------------

void PSD_Window::setPotentialChoice(const QString & text){
    
	if( text.contains("No") ) {
		potentialField->setEnabled(false);
	}
	else if( text.contains("eV") ) {
		potentialField->setEnabled(true);
	}
	else {
		potentialField->setEnabled(true);
	}
}

// -------------

void PSD_Window::setMomentsAll(int on){

	dBtn->setChecked(on);
	vBtn->setChecked(on);
	pBtn->setChecked(on);
	hBtn->setChecked(on);
	
}

// -------------

void PSD_Window::setThresholdChoice(const QString & text){
    
	if( text.contains("No") ) {
		thresholdField->setEnabled(false);
	}
	else if( text.contains("eV") ) {
		thresholdField->setEnabled(true);
	}
	else {
		thresholdField->setEnabled(true);
	}
}
// -------------

void PSD_Window::setMassChoice(const QString & text){
    
	if( text.contains("Electron") ) {
		massField->setEnabled(false);
		chargeField->setEnabled(false);
		massField->setValue(0.0005447);
		chargeField->setValue(-1);   // SJS change
	}
	else if( text.contains("Proton") ) {
		massField->setEnabled(false);
		chargeField->setEnabled(false);
		massField->setValue(1);
		chargeField->setValue(1);
	}
	else if( text.contains("He++") ) {
		// must test He++ first as He+ also tests true
		massField->setEnabled(false);
		chargeField->setEnabled(false);
		massField->setValue(4);
		chargeField->setValue(2);
	}
	else if( text.contains("He+") ) {
		massField->setEnabled(false);
		chargeField->setEnabled(false);
		massField->setValue(4);
		chargeField->setValue(1);
	}
	else if( text.contains("O+") ) {
		massField->setEnabled(false);
		chargeField->setEnabled(false);
		massField->setValue(16);
		chargeField->setValue(1);
	}
	else {
		massField->setEnabled(true);
		chargeField->setEnabled(true);
		massField->setValue(1.);
		chargeField->setValue(1);
	}

	
}

// -------------

void PSD_Window::setCentreChoice(const QString & text){
    
	if( text.contains("PSD") ) {
		centreField->setEnabled(false);
	}
	else {
		centreField->setEnabled(true);
	}
}


// -------------

void PSD_Window::setNewName(){
     
	if(QuSlotDebug())
	{
		printf("PSD_Window::setNewName \n");
		fflush(stdout);
	}
	
	// Look for PEACE or CIS
	DvObject_var dobj = objField->getObject();

	bool isPEACE = false;
	bool isCIS = false;
	if(dobj->xref_exists(OBJECT_NAME) ){
		DvString name_str = dobj->getXrefText(OBJECT_NAME);
		if( name_str.contains("CIS")) isCIS = true;
		else if( name_str.contains("PEA")) isPEACE = true;
	}
	
	bool isGSE = false;
	bool isGSM = false;
	DvString attr_str = dobj->getFrame();
	if( attr_str.contains("GSE")) isGSE = true;
	else if( attr_str.contains("GSM") )isGSM = true;

	if(isGSE || isGSM) statusBar()->showMessage("PSD already in GSE or GSM frame",0);
	else statusBar()->showMessage(" ",0);
	

	if(dobj->xref_exists("ENTITY") ){
		DvString entity_str = dobj->getXrefText("ENTITY");
		if (entity_str=="electron") {
			massChoice->setCurrentIndex(0);
			setMassChoice("Electron");
		}
		else if (entity_str=="proton") {
			massChoice->setCurrentIndex(1);
			setMassChoice("Proton");
		}
		else if (entity_str=="helium+") {
			massChoice->setCurrentIndex(2);
			setMassChoice("He+");
		}
		else if (entity_str=="helium++") {
			massChoice->setCurrentIndex(3);
			setMassChoice("He++");
		}
		else if (entity_str=="alpha") {
			massChoice->setCurrentIndex(3);
			setMassChoice("He++");
		}
		else if (entity_str=="oxygen+") {
			massChoice->setCurrentIndex(4);
			setMassChoice("O+");
		}
		else {
			massChoice->setCurrentIndex(5);
			massField->setValue(1);
			chargeField->setValue(1);
		}
	}
// setNewName signaled by edit to dataobj, which incl pulldown-subset
// So subsetting undoes mass choice. For now, leave unchanged and let
// user set if entity_str isn't present in data.
//	else{
//		massChoice->setCurrentIndex(5);
//	}
	
	if(isPEACE){
		Vparity->setCurrentIndex(0);
		transMethod->setCurrentIndex(0);
		massChoice->setCurrentIndex(0);
	}
	else if(isCIS){
		Vparity->setCurrentIndex(1);
		transMethod->setCurrentIndex(1);
	}
			
	// set output name default
	if (!userSet) setObjectName();

	this->adjustSize();
}

// -----------------

void PSD_Window::setObjectName()
{  
	DvString objName = this->objField->textObject().toLatin1().data();
   
	if(!objName.empty()){
		objName = objName.back("/");
		objName = objName.front(" [");
		this->name_field->setText(objName);
	}
	else{
		this->name_field->setText("PSD");
	}
   
}
// -------------

void PSD_Window::keyPressEvent(QKeyEvent * e )
{  
     
 if(QuSlotDebug())
 {
	printf("PSD_Window::keyPressEvent \n");
	fflush(stdout);
 }
  
	  // use escape key to toggle default buttons
  if(e->key() == Qt::Key_Escape){
	  if( saveBtn->isDefault()  ) {
		   saveBtn->setDefault(false);
		   dismissBtn->setDefault(true);
	  }
	  else if(dismissBtn->isDefault()  )  {
		   dismissBtn->setDefault(false);
		   saveBtn->setDefault(true);
	  }
	  return;
  }
  // use return or enter key to perform default
  if(e->key() == Qt::Key_Return ||  e->key() == Qt::Key_Enter) {
     if(saveBtn->isDefault()) evaluate();
	 else if (dismissBtn->isDefault()) close();
     return;
  }

  // else pass on to list view 
  e->ignore(); // let selection change

}

// -------------

void PSD_Window::setObject(const char * WLctxt){
    
 if(QuSlotDebug())
 {
	printf("PSD_Window::setObject \n");
	fflush(stdout);
 }
  
   objField->setText(WLctxt);
   objField->reValidate(true);
   
}


// -------------

void PSD_Window::evaluate(){
	
	static PSD_Parameters *para_ptr=NULL;
	
	if(QuSlotDebug())
	{
		printf("PSD_Window::evaluate \n");
		fflush(stdout);
	}
	if(running){
		running = false;
		// reset GUI
		setEvalBtn("Evaluate");
		percentage->setText(" " );
		
		if(para_ptr != NULL) para_ptr->stop = true;
		this->statusBar()->showMessage("Interrupt: output data truncated", 20000);
		
		return;

	}
	
	running = true;
	
	setEvalBtn(" Interrupt ");

	qApp->processEvents();

	
	if (processMethod->currentText().contains("Rotation Matrix")){
		// don't need PSD
		DvObject_var RotnMtx = this->makeRotnMatrix();
		
		if( RotnMtx.is_ok() ) this->saveRotnMatrix(RotnMtx);
		else this->statusBar()->showMessage("Need Lat and Long", 10000);
		
		// reset GUI
		setEvalBtn("Evaluate");
		percentage->setText(" " );
		running = false;
		return;
	}
	
	DvObject_var data = objField->getObject();
	int direction;
	if( Vparity->currentText().contains("Look") ) direction = -1;
	else  direction = 1;
	
    PSD_Parameters para(data, direction);
	para_ptr = &para;

    if(!para.init_ok)
    {
		this->statusBar()->showMessage(para.message, 20000);
		// reset GUI
		setEvalBtn("Evaluate");
		percentage->setText(" " );
		running = false;
        return;
    }
	
	// set the mass and charge
	
	para.init_species(massField->value(), chargeField->value());
		
	// set rotation matrix: transforms PSD coords into target science (e.g. GSE)

	DvObject_var rtnMtx = this->makeRotnMatrix();

   	if(rtnMtx.is_nil()) {
		this->statusBar()->showMessage("Failed to create rotation matrix", 20000);
		// reset GUI
		setEvalBtn("Evaluate");
		percentage->setText(" " );
		running = false;
		return; 
	}
	
	if( para.init_rotn(rtnMtx) == false) {
		this->statusBar()->showMessage(para.message, 20000);
		// reset GUI
		setEvalBtn("Evaluate");
		percentage->setText(" " );
		running = false;
		return;
	}

	// Save Slice, PAD, or Moments
	
	if (processMethod->currentText().contains("Simple Slice")){
	
		// set up parameters
		
		// centre of velocity to subtract off
		DvObject_var centreOfPSD; // will be is_nil() if PSD centre required
		
		if(centreChoice->currentText().contains("Velocity") ){
			// user provided velocity centre
			centreOfPSD = centreField->getObject();
			
		} 
		if( !para.init_PSDcentre(centreOfPSD) ) {
			this->statusBar()->showMessage(para.message, 20000);
			// reset GUI
			setEvalBtn("Evaluate");
			percentage->setText(" " );
			running = false;
			return;
		}
		
		// direction of normal
		DvObject_var normal; 
		DvString toframe = rtnMtx->getXrefText(TOFRAME);
		
		if(dirnChoice->currentText().contains("Z") ){
			normal = getWLObject("Constants/Iz");
			if( !toframe.empty() ) normal->change_xref(FRAME, toframe);
		}
		else if(dirnChoice->currentText().contains("Y") ){
			normal = getWLObject("Constants/Iy");
			if( !toframe.empty() ) normal->change_xref(FRAME, toframe);
		}
		else if(dirnChoice->currentText().contains("X") ){
			normal = getWLObject("Constants/Ix");
			if( !toframe.empty() ) normal->change_xref(FRAME, toframe);
		}
		else if(dirnChoice->currentText().contains("Along") ){

			// user provided direction
			normal = dirnField->getObject();

		} 
		if(para.init_sliceNormal(normal) == false){
			this->statusBar()->showMessage(para.message, 20000);
		// reset GUI
			setEvalBtn("Evaluate"); percentage->setText(" " ); running = false;
			return;
		}

		// set SC potential 		
		DvObject_var pot_var;
		if(potentialChoice->currentText().contains("No Correction")){
			pot_var = new DvObject(0.0);
		}
		else {
			pot_var = potentialField->getObject();
		}		
		if( !para.init_potential(pot_var) ) {
			this->statusBar()->showMessage(para.message, 20000);
			// reset GUI
			setEvalBtn("Evaluate");
			percentage->setText(" " );
			running = false;
			return;
		}

		// set threshold energy 		 
		DvObject_var threshold_var;
		if(thresholdChoice->currentText().contains("No Threshold")){
			threshold_var = new DvObject(0.0);
		}
		else {
			threshold_var = thresholdField->getObject();
		}		
		if( !para.init_threshold(threshold_var) ) {
			this->statusBar()->showMessage(para.message, 20000);
			// reset GUI
			setEvalBtn("Evaluate");
			percentage->setText(" " );
			running = false;
			return;
		}
			
		// set up memory for quantities used often
		para.init_cartesian();

		para.resolution = numberBins->value();

		evaluateSlice(para);
			
	}
	
	// Advanced Slice
	if (processMethod->currentText().contains("Advanced Slice")){
	
		// set up parameters
		
		// centre of velocity to subtract off
		DvObject_var centreOfPSD; // will be is_nil() if PSD centre required
		
		if(centreChoice->currentText().contains("Velocity") ){
			// user provided velocity centre
			centreOfPSD = centreField->getObject();			
		} 
		if( !para.init_PSDcentre(centreOfPSD) ) {
			this->statusBar()->showMessage(para.message, 20000);
			// reset GUI
			setEvalBtn("Evaluate");
			percentage->setText(" " );
			running = false;
			return;
		}
		
		// direction of Avec,Bvec for normal & coord sys
		// check all in same frame. Note - sameFrame returns true if frame missing
		DvObject_var Avec, Bvec;
		Avec = dirnAvecField->getObject();
		Bvec = dirnBvecField->getObject();
		
		DvString toframe = rtnMtx->getXrefText(TOFRAME);
		if (!Avec->sameFrame(toframe) || !Bvec->sameFrame(toframe)){
			this->statusBar()->showMessage("V1,V2, PSD->Sci frames don't match");
		// reset GUI
			setEvalBtn("Evaluate"); percentage->setText(" " ); running = false;
			return;
		}
		
		int choice;
		if(dirnAdvChoice->currentText().contains("n||V1") ) 
		{ choice=0;
		} else if(dirnAdvChoice->currentText().contains("V1xV2") )
		{ choice=1;
		} else {
			this->statusBar()->showMessage("unknown normal choice");
		// reset GUI
			setEvalBtn("Evaluate"); percentage->setText(" " ); running = false;
			return;
		}
		
		// go set up sliceX, sliceY, and sliceNormal vectors
		if(para.init_AdvsliceNormal(Avec,Bvec,choice) == false){
			this->statusBar()->showMessage(para.message, 20000);
		// reset GUI
			setEvalBtn("Evaluate"); percentage->setText(" " ); running = false;
			return;
		}

		// set SC potential 		
		DvObject_var pot_var;
		if(potentialChoice->currentText().contains("No Correction")){
			pot_var = new DvObject(0.0);
		}
		else {
			pot_var = potentialField->getObject();
		}		
		if( !para.init_potential(pot_var) ) {
			this->statusBar()->showMessage(para.message, 20000);
			// reset GUI
			setEvalBtn("Evaluate");
			percentage->setText(" " );
			running = false;
			return;
		}

		// set threshold energy 		 
		DvObject_var threshold_var;
		if(thresholdChoice->currentText().contains("No Threshold")){
			threshold_var = new DvObject(0.0);
		}
		else {
			threshold_var = thresholdField->getObject();
		}		
		if( !para.init_threshold(threshold_var) ) {
			this->statusBar()->showMessage(para.message, 20000);
			// reset GUI
			setEvalBtn("Evaluate");
			percentage->setText(" " );
			running = false;
			return;
		}
			
		// set up memory for quantities used often
		para.init_cartesian();  // to hold cartes form of PSD array

		para.resolution = numberBins->value();

		evaluateAdvSlice(para, choice);
			
	}

	
	// Pitch angle distribution
	
	if (processMethod->currentText().contains("Pitch")){
	
		// set up parameters		
		// centre of velocity to subtract off
		DvObject_var centreOfPSD; // will be is_nil() if PSD centre required
		
		if(centreChoice->currentText().contains("Velocity") ){
			// user provided velocity centre
			centreOfPSD = centreField->getObject();			
		} 
		if( !para.init_PSDcentre(centreOfPSD) ) {
			this->statusBar()->showMessage(para.message, 20000);
			// reset GUI
			setEvalBtn("Evaluate");
			percentage->setText(" " );
			running = false;
			return;
		}
		
		// direction of magnetic field
		DvObject_var fielddir; 
		DvString toframe = rtnMtx->getXrefText(TOFRAME);
		
		if(dirnChoice->currentText().contains("Z") ){
			fielddir = getWLObject("Constants/Iz");
			if( !toframe.empty() ) fielddir->change_xref(FRAME, toframe);
		}
		else if(dirnChoice->currentText().contains("Y") ){
			fielddir = getWLObject("Constants/Iy");
			if( !toframe.empty() ) fielddir->change_xref(FRAME, toframe);
		}
		else if(dirnChoice->currentText().contains("X") ){
			fielddir = getWLObject("Constants/Ix");
			if( !toframe.empty() ) fielddir->change_xref(FRAME, toframe);
		}
		else if(dirnChoice->currentText().contains("Along") ){

			// user provided field direction
			fielddir = dirnField->getObject();
		} 
		if(para.init_PADfielddir(fielddir) == false){
			this->statusBar()->showMessage(para.message, 20000);
		// reset GUI
			setEvalBtn("Evaluate"); percentage->setText(" " ); running = false;
			return;
		}
		
		// set SC potential 		
		DvObject_var pot_var;
		if(potentialChoice->currentText().contains("No Correction")){
			pot_var = new DvObject(0.0);
		}
		else {
			pot_var = potentialField->getObject();
		}		
		if( !para.init_potential(pot_var) ) {
			this->statusBar()->showMessage(para.message, 20000);
			// reset GUI
			setEvalBtn("Evaluate");
			percentage->setText(" " );
			running = false;
			return;
		}

		// set threshold energy 		 
		DvObject_var threshold_var;
		if(thresholdChoice->currentText().contains("No Threshold")){
			threshold_var = new DvObject(0.0);
		}
		else {
			threshold_var = thresholdField->getObject();
		}		
		if( !para.init_threshold(threshold_var) ) {
			this->statusBar()->showMessage(para.message, 20000);
			// reset GUI
			setEvalBtn("Evaluate");
			percentage->setText(" " );
			running = false;
			return;
		}
	
		// set up memory for quantities used often
		para.init_cartesian();

		para.resolution = numberBins->value();
		int npa = numberPADBins->value();
		int nphi = numberPADphiBins->value();

		evaluatePAD(para,npa,nphi);	
	}
	
	else if (processMethod->currentText().contains("Moments of PSD")){
	
		// set SC potential 		
		DvObject_var pot_var;
		if(potentialChoice->currentText().contains("No Correction")){
			pot_var = new DvObject(0.0);
		}
		else {
			pot_var = potentialField->getObject();
		}		
		if( !para.init_potential(pot_var) ) {
			this->statusBar()->showMessage(para.message, 20000);
			// reset GUI
			setEvalBtn("Evaluate");
			percentage->setText(" " );
			running = false;
			return;
		}

		// set threshold energy 		 
		DvObject_var threshold_var;
		if(thresholdChoice->currentText().contains("No Threshold")){
			threshold_var = new DvObject(0.0);
		}
		else {
			threshold_var = thresholdField->getObject();
		}		
		if( !para.init_threshold(threshold_var) ) {
			this->statusBar()->showMessage(para.message, 20000);
			// reset GUI
			setEvalBtn("Evaluate");
			percentage->setText(" " );
			running = false;
			return;
		}

		para.init_cartesian();
		

		evaluateMoments(para);

	}

	// reset GUI
	setEvalBtn("Evaluate");
	percentage->setText(" " );
	running = false;


}
//-----------------

bool PSD_Window::evaluateSlice(PSD_Parameters &para){

	DvObject_var PSD_slice;

	para.getSlice(PSD_slice, percentage);
		
	QString newName = name_field->text();
	newName += "_slice";
	QwuiPutOnWL(PSD_slice, newName);
	
	return true;
}

bool PSD_Window::evaluateAdvSlice(PSD_Parameters &para, int choice){

	DvObject_var PSD_slice;
	DvObject_var dep1;
	DvObject_var dep2;
	DvString temp_str;

	para.getAdvSlice(PSD_slice, choice, percentage);
	dep1 = PSD_slice->get_xref("DEPEND_1");
	dep2 = PSD_slice->get_xref("DEPEND_2");
			
	QString newName = name_field->text();
	newName += "_slice";
	
	DvObject_var Avec, Bvec;
	Avec = dirnAvecField->getObject();
	Bvec = dirnBvecField->getObject();

	DvString Avec_str;
	if (Avec->xref_exists("LABLAXIS"))
	{	Avec_str = Avec->getXrefText("LABLAXIS");
	} else if (Avec->xref_exists("FIELDNAM"))
	{	Avec_str = Avec->getXrefText("FIELDNAM");
	} else {Avec_str = dirnBvecField->text().toStdString();
	}
	
	DvString Bvec_str;
	if (Bvec->xref_exists("LABLAXIS"))
	{	Bvec_str = Bvec->getXrefText("LABLAXIS");
	} else if (Bvec->xref_exists("FIELDNAM"))
	{	Bvec_str = Bvec->getXrefText("FIELDNAM");
	} else {Bvec_str = dirnBvecField->text().toStdString();
	}
	
	DvString orientation_str = "Slice geometry: V1 = ";
	orientation_str += Avec_str;
	orientation_str =+ "   V2 =";
	orientation_str += Bvec_str;
	if (choice==0){
		orientation_str += " n along V1 direction";
		temp_str = "V along perp part of ";
		temp_str += Bvec_str;
		dep1->change_xref("LABLAXIS",temp_str);
		temp_str = "V along ";
		temp_str += Avec_str; 
		temp_str +=" x ";
		temp_str +=Bvec_str;
		dep2->change_xref("LABLAXIS",temp_str);
		
	} else if (choice==1){
		orientation_str += " n along V1xV2 direction";
		temp_str = "V along  ";
		temp_str += Avec_str;
		dep1->change_xref("LABLAXIS",temp_str);
		temp_str = "V along perp part of ";
		temp_str += Bvec_str;
		dep2->change_xref("LABLAXIS",temp_str);
	}
	PSD_slice->change_xref("Slice Orientation",orientation_str);
	PSD_slice->change_xref("DEPEND_1",dep1);
	PSD_slice->change_xref("DEPEND_2",dep2);
	
	QwuiPutOnWL(PSD_slice, newName);
	
	return true;
}

bool PSD_Window::evaluatePAD(PSD_Parameters &para, int npa, int nphi){

	DvObject_var PSD_PAD;

	para.getPAD(PSD_PAD, npa, nphi, percentage);
		
	QString newName = name_field->text();
	newName += "_PAD";
	QwuiPutOnWL(PSD_PAD, newName);
	
	return true;
}


//-----------------

bool PSD_Window::evaluateMoments(PSD_Parameters &para){

	DvObject_var density;
	DvObject_var velocity;
	DvObject_var pressure;
	DvObject_var heat;
	DvObject_var temperature;
	
	int flags = MOM_DENSITY|MOM_VELOCITY; // p, T and h need n & v
	if(pBtn->isChecked()) flags = flags|MOM_PRESSURE;
	if(hBtn->isChecked()) flags = flags|MOM_HEATFLUX;
	
	if(integrationMethod->currentText().contains("Sum") ){
		// Sum over bins
		
		para.getMomentsSummed(flags, density, velocity, pressure, heat, temperature, percentage);
	}
	else{
		// Integrate using Delauney triangulation
		
		para.getMomentsIntegrated(flags, density, velocity, pressure, heat, temperature, percentage);
		
	}
	 
	if( dBtn->isChecked() && !density.is_nil() ){		

		QString newName = name_field->text();
		newName += "_den";
		QwuiPutOnWL(density, newName);
	}
	if( vBtn->isChecked() && !velocity.is_nil() ){		

		QString newName = name_field->text();
		newName += "_vel";
		QwuiPutOnWL(velocity, newName);
	}
	if( pBtn->isChecked() && !pressure.is_nil() ){		

		QString newName = name_field->text();
		newName += "_pres";
		QwuiPutOnWL(pressure, newName);
	}
	if( hBtn->isChecked() && !heat.is_nil() ){		

		QString newName = name_field->text();
		newName += "_heatFlux";
		QwuiPutOnWL(heat, newName);
	}
	if( pBtn->isChecked() && !temperature.is_nil() ){		

		QString newName = name_field->text();
		newName += "_T";
		QwuiPutOnWL(temperature, newName);
	}
	return true;
}


//-----------------

bool PSD_Window::saveRotnMatrix(DvObject_var &rtnMtx){

	if(rtnMtx.is_nil()){
		this->statusBar()->showMessage("Rotation Matrix is empty", 10000);
		return false;
	}
	
	QString name = name_field->text();
	if(transTarget->currentIndex() == GSE) name += "_gse";
	else if(transTarget->currentIndex() == GSM) name += "_gsm";
	
	QwuiPutOnWL(rtnMtx, name);
	return true;
}

//-----------------

DvObject_var PSD_Window::makeRotnMatrix(){
	
	DvObject_var rtnMtx;
	DvObject_var times;
	bool isToGSE = true;
	bool isToGSM = false;
	int n_elements;

    if(transTarget->currentText().contains("Do Not")){
        // need unit matrix to avoid failure trap on return
        rtnMtx = new DvObject(0.0, DIM33, 1);
        rtnMtx->dbl(0,0,0) = 1.0;
        rtnMtx->dbl(0,1,1) = 1.0;
        rtnMtx->dbl(0,2,2) = 1.0;
        return rtnMtx;
    }
	else if(transMethod->currentText().contains("use rotation")){
		rtnMtx = rotnField->getObject();
		if(rtnMtx.is_nil()){
			this->statusBar()->showMessage("No Rotation Matrix ", 10000);
			return rtnMtx;
		}
		times = rtnMtx->getTimeTags();
		n_elements = rtnMtx->seqSize();
		
		if(rtnMtx->xref_exists("ToFrame") ){
			DvString toframe = rtnMtx->getXrefText("ToFrame");		
			
			if( toframe.contains("gsm") ) { 
				isToGSM = true; 
				isToGSE = false;
			}
		}
	}
	else {
	
	
		DvObject_var lat_temp = latField->getObject();
    	DvObject_var lon_temp = longField->getObject();

    	double conv2rad_lat=M_PI/180.;
    	double conv2rad_lon=M_PI/180.;
    
		if(lat_temp.is_nil() || lon_temp.is_nil()) return rtnMtx;

		DvString units = lat_temp->getXrefText("UNITS");
		if(units.contains("rad")) conv2rad_lat=1.;
    	

		units = lon_temp->getXrefText("UNITS");
		if(units.contains("rad")) conv2rad_lon=1.;
   
    	
		DvString frame = lat_temp->getFrame();		
		if(frame.empty()) frame = lon_temp->getFrame();	
			
		if(frame.contains("gsm") ) { 
			isToGSM = true; 
			isToGSE = false;
		}
			
		if( !lon_temp->isJoined(*lat_temp) )
		{
			lon_temp = lon_temp->Join(lat_temp);
		}
        
		times = lat_temp->getTimeTags();
		n_elements=lat_temp->seqSize();
		
		rtnMtx = new DvObject(0.0, DIM33, n_elements);
		       
		double lat, lon;
		double SCx, SCz, SCy, alpha;
 
		for(int i=0; i<n_elements; i++)
		{
			lat = lat_temp->asDouble(i)*conv2rad_lat;
            
			lon = lon_temp->asDouble(i)*conv2rad_lon;

			SCx = cos(lat)*cos(lon);
			SCz = sin(lat);
			SCy = cos(lat)*sin(lon);
			alpha = sqrt(1.-SCx*SCx);

			if(transMethod->currentText().contains("PEA")){
				rtnMtx->dbl(i,0, 0)=alpha;
				rtnMtx->dbl(i,0, 1)=0.;
				rtnMtx->dbl(i,0, 2)=SCx;
				rtnMtx->dbl(i,1, 0)=-SCx*SCy/alpha;
				rtnMtx->dbl(i,1, 1)=SCz/alpha;
				rtnMtx->dbl(i,1, 2)=SCy;
				rtnMtx->dbl(i,2, 0)=-SCx*SCz/alpha;
				rtnMtx->dbl(i,2, 1)=-SCy/alpha;
				rtnMtx->dbl(i,2, 2)=SCz;
			}
			else if(transMethod->currentText().contains("CIS")){
				rtnMtx->dbl(i,0, 0)=alpha;
				rtnMtx->dbl(i,0, 1)=0.;
				rtnMtx->dbl(i,0, 2)=-SCx;
				rtnMtx->dbl(i,1, 0)=-SCx*SCy/alpha;
				rtnMtx->dbl(i,1, 1)=-SCz/alpha;
				rtnMtx->dbl(i,1, 2)=-SCy;
				rtnMtx->dbl(i,2, 0)=-SCx*SCz/alpha;
				rtnMtx->dbl(i,2, 1)=SCy/alpha;
				rtnMtx->dbl(i,2, 2)=-SCz;
			}
            
		}
		if(times.is_ok()) rtnMtx->change_xref(DEPEND_0, times);
   
    
	} 
	
	// rotate to GSM?
	
	if(transTarget->currentText().contains("GSM") && isToGSE ){
		if( times.is_nil() ) {
			this->statusBar()->showMessage("No timetags for GSE to GSM rotation", 10000);
			if(transMethod->currentIndex() == 0) rtnMtx->setFrameAttr( "rot>SR2PEA_xyz");
			else if(transMethod->currentIndex() == 1) rtnMtx->setFrameAttr( "rot>ISR2CIS_xyz");
		
			rtnMtx->setToFrameAttr("gse"); // ToFrame is GSE as rotn didn't happen

		}
		else{
	
			RtnVector vec;
			vec.n_recs = n_elements;
			vec.in_frame = GSE;
			vec.out_frame = GSM;
			vec.t = times;
			vec.IsPositionVector = false;

			DvObject_var gse2gsm = vec.QrGetRotnMatx();
		
			rtnMtx = gse2gsm * rtnMtx;
						
		}
	}
	
	// rotate to GSE?
	if(transTarget->currentText().contains("GSE") && isToGSM ){
		if( times.is_nil() ) {
			this->statusBar()->showMessage("No timetags for GSM to GSE rotation", 10000);
			if(transMethod->currentIndex() == 0) rtnMtx->setFrameAttr( "rot>SR2_xyz");
			else if(transMethod->currentIndex() == 1) rtnMtx->setFrameAttr( "rot>ISR2_xyz");
		
			rtnMtx->setToFrameAttr("gsm"); // ToFrame is GSM as rotn didn't happen
		}
		else{
		
			RtnVector vec;
			vec.n_recs = n_elements;
			vec.in_frame = GSM;
			vec.out_frame = GSE;
			vec.t = times;
			vec.IsPositionVector = false;

			DvObject_var gsm2gse = vec.QrGetRotnMatx();
			
			rtnMtx = gsm2gse * rtnMtx;

		}
	}
	
	// set ToFrame attribute
	
	if(transMethod->currentIndex() == 0) rtnMtx->setFrameAttr( "rot>SR2_xyz");
	else if(transMethod->currentIndex() == 1) rtnMtx->setFrameAttr( "rot>ISR2_xyz");
		
	if(transTarget->currentIndex() == GSM) rtnMtx->setToFrameAttr("gsm");
	else rtnMtx->setToFrameAttr("gse");


	return rtnMtx;

}



