
/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/

#include "qplug_if.h"
#include "DvObj.h"

extern "C" QplugReturnStatus DvToObjUnits(vector<DvObject_var>&, vector<DvObject_var>&);

QplugReturnStatus DvToObjUnits(vector<DvObject_var>& inputs, vector<DvObject_var>& outputs)
{

	DvObject_var in_ptr;
	DvObject_var out_ptr;
	DvObject_var SI_target;

	DvString out_name;

	if(inputs.size() < 2) {
    	QplugAppendTextDisplay ("2 inputs required\n");
    	return QPLUG_FAILURE; 
	} 
 
	//	Unpack the input data objects

	in_ptr = inputs[0];
	SI_target = inputs[1];
	
	if(in_ptr.is_nil() || SI_target.is_nil()){
    	QplugAppendTextDisplay ( "Invalid input object \n");
    	return QPLUG_FAILURE; 
	}
	
	if( !in_ptr->sameBaseSI(*SI_target) ){
    	QplugAppendTextDisplay ( "Not same base SI units \n");
    	return QPLUG_FAILURE; 
	}

	DvString SIcTarget =  SI_target->getXrefText(SI_CONVERSION);
	DvString UnitsTarget =  SI_target->getXrefText(UNITS);

 
	if( SIcTarget.empty()){
    	QplugAppendTextDisplay ( "Second object has no SI_conversion \n");
    	return QPLUG_FAILURE; 
	}
	if( UnitsTarget.empty()){
		QplugAppendTextDisplay ( "Second object has no UNITS attribute \n");
		 
		UnitsTarget = SIcTarget.after(">");
		UnitsTarget += " / ";
		UnitsTarget += SIcTarget.before(">").toDouble();
	}
 
	// Get name of output object

 	out_name = GetSlotText(0);
 	if (out_name.empty()) {
    	out_name = "InNewUnits";
    	SetSlotText(0, out_name.c_str());
	}


	// create output object
    // this creates new converted object, copies xrefs across and sets SI_conversion
  
	out_ptr = in_ptr->changeUnitsTo(SIcTarget, UnitsTarget);

	if (out_ptr->nError() > 0) QplugAppendTextDisplay( out_ptr->getError().c_str() );


	//	..add  the output series to the list

	outputs.push_back(out_ptr);

	QplugAppendTextDisplay ("Convert to new units - Done.\n");
  
	return QPLUG_SUCCESS;

}


