/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	ComplexProduct.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS ComplexProduct wrapper
*
* PURPOSE:      Wrapper for NR 'spctrm' power spectrum.
* 
* DESCRIPTION:  Plug In module checking input parameters.
*
* ARGUMENTS:
*    IN:      
*
*   OUT:
*         
* GLOBALS:
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
* FUNCTIONS_USED:	
*
* ALGORITHM:  
*
* HISTORY: created  15/3/2001 Markus Fraenz 
*
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"

extern "C" QplugReturnStatus	DvComplexProd(vector <DvObject_var>&, vector <DvObject_var>&);

using namespace QSAS;



QplugReturnStatus	DvComplexProd(vector <DvObject_var>& inputs, vector <DvObject_var>& outputs)
{

// define some C variables

int	        attribute_ok,list_length,vectorset=0;
int	Ndata;
int qar_status;

DvString str_out;

// define QSAS data objects


DvObject_var out_im_series;
DvObject_var out_re_series;


//
// ...first, unpack the input data objects
//

// ...first item is input Series

QplugAppendTextDisplay("ComplexProduct \n");

DvObject_var Re1 = inputs[0];

if(Re1.is_nil() || Re1->arraySize() > 1 || Re1->not_dbl() ){
      QplugAppendTextDisplay ("ComplexProduct>> ERROR - Re1 sequence type not handled\n");
      return QPLUG_FAILURE;  
}

Ndata = Re1->seqSize();

DvObject_var Im1 = inputs[1];

if(Im1.is_nil()  || Im1->arraySize() > 1 || Im1->not_dbl()){
      QplugAppendTextDisplay ("ComplexProduct>> ERROR - Im1 sequence type not handled\n");
      return QPLUG_FAILURE;  
}


DvObject_var Re2 = inputs[2];

if(Re2.is_nil() || Re2->arraySize() > 1 || Re2->not_dbl() ){
      QplugAppendTextDisplay ("ComplexProduct>> ERROR - Re2 sequence type not handled\n");
      return QPLUG_FAILURE;  
}


DvObject_var Im2 = inputs[3];

if(Im2.is_nil()  || Im2->arraySize() > 1 || Im2->not_dbl()){
      QplugAppendTextDisplay ("ComplexProduct>> ERROR - Im2 sequence type not handled\n");
      return QPLUG_FAILURE;  
}


// DVOS will join on the fly if needed
// It will also attache time line and modified xrefs

QplugAppendTextDisplay ("ComplexProduct>> Calculating.\n");


out_re_series = new DvObject(0.0, Ndata);
out_im_series = new DvObject(0.0, Ndata);

for (long i =0; i<Ndata; i++) { 
 out_re_series->dbl(i)  = Re1->dbl(i) * Re2->dbl(i);
 out_re_series->dbl(i)  += Im1->dbl(i) * Im2->dbl(i);
 out_im_series->dbl(i) = Re1->dbl(i) * Im2->dbl(i);
 out_im_series->dbl(i) -= Im1->dbl(i) * Re2->dbl(i);
}


  out_re_series->change_xref(LABLAXIS, "ReCProd");
  out_re_series->change_xref(FIELDNAM, "ReCProd");

  out_im_series->change_xref(LABLAXIS, "ImCProd");
  out_im_series->change_xref(FIELDNAM, "ImCProd");

//	..add them to the list

  outputs.push_back(out_re_series);    
  outputs.push_back(out_im_series);     


//	..some messages to the analysis interface window..

str_out = "ComplexProduct>> ";
str_out += Ndata;
str_out += " data points\n";
QplugAppendTextDisplay (str_out.c_str());
QplugAppendTextDisplay ("ComplexProduct>> finished...\n");


return QPLUG_SUCCESS;


}



