
/**********************************************************************
* SCCS id: @(#)CorCoefPlug.cc	1.1 14:32:23 12/17/98
************************************************************************
* PROJECT:     Cluster UK CDHF
* COMPONENT:   Correlation Coefficient Plugin 
* MODULE:      CorCoefPlug.cc
* LANGUAGE:	   C++
* FUNCTION:
* PURPOSE:     Compute Correlation Coefficient between 2 scalar time series 
* DESCRIPTION:
* ARGUMENTS: [passed via call_list] - see qtpl and CorCoef.c for details
*
*   INPUT  ANY_TIME_SERIES xts_in, 
*   INPUT  ANY_TIME_SERIES yts_in, 
*   INPUT  TIME_INTERVAL start_end,
*   INPUT  FLOAT dt_lag_max,
*   OUTPUT SCALAR_SERIES dt_lag,
*   OUTPUT SCALAR_SERIES rho_out,
*   OUTPUT VECTOR_SERIES rho_vec,
*
* RETURN:
*    TYPE:     QplugReturnStatus
*    VALUES:   see qplug_if.h
* FUNCTIONS_USED: CorCoef, stats, CorCoefPlug_utils, IndexTSSubset
* ALGORITHM:   see qtpl and CorCoef.c
* HISTORY:	   Steve Schwartz  Dec 98
*               vector input/ouput added M.Fraenz Feb 2001
***********************************************************************/

#ifndef __cplusplus
#define __cplusplus
#endif

#include "CorCoefPlug.h"       // this then sucks in all other headers

QplugReturnStatus  DvCorCoefStart  ( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs)
{

CorCoefPlugReturnStatus CCP_return;     // used to return from sub-routines
CCutilsSuccessorFailure       ok;
CorCoefReturnStatus     CC_return;

CorCoefPlug_ptrs       ccp, *ccpp=&ccp;  // holds pointers to data arrays
CorCoefpacket          ccpack;

double half_spacing, tol;
int			ixstart,iystart;
                                        

QplugAppendTextDisplay("Correlation Coefficient. Feb 2001 & May 2015\n");

CCP_return=CorCoefPlug_ini(ccpp);   // initialise NULL ptrs

//	...first, check the types of the input arguments, and load the data

CCP_return =CorCoefPlug_checkload_input (inputs, ccpp);
if (CCP_return != CORCOEF_PLUG_SUCCESS){
    QplugAppendTextDisplay(CorCoefPlugReturnMessage[CCP_return]);
    return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);
}

#ifdef VERBOSE_CC
QplugAppendTextDisplay("CorCoefPlug input types ok and loaded successfully\n");
#endif

CCP_return = CorCoefPlug_SubsetXY (ccpp);
if (CCP_return != CORCOEF_PLUG_SUCCESS){
    QplugAppendTextDisplay(CorCoefPlugReturnMessage[CCP_return]);
    return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);
}
ixstart = ccpp->ixstart;
iystart = ccpp->iystart;
// ... test timetags for regular and joined. Set tag spacing in ms

ok = Test_regular (&(ccpp->tx[ixstart]), ccpp->nxsub, &(ccpp->dt_spacing));
if (ok != CCUTILSSUCCESS){
 QplugAppendTextDisplay("CorCoefPlug ERROR>> x timetags not regular\n");
 return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);
}
ccpp->dt_spacing *= 2.0;    // Test_regular returns half_spacing

// Use spacing to deterine tag tolerance. Minumum of 10 ms or 10% spacing
tol = ccpp->dt_spacing > 100 ? 10 : 0.1*ccpp->dt_spacing;

ok = Test_join(&(ccpp->tx[ixstart]), ccpp->nxsub, 
               &(ccpp->ty[iystart]), ccpp->nysub, tol);
if (ok!=CCUTILSSUCCESS){
 QplugAppendTextDisplay("CorCoefPlug ERROR>> Subsetted tx and ty not Joined\n");
 return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);
}

#ifdef VERBOSE_CC
  QplugAppendTextDisplay("CorCoefPlug: DO joined and regular tags\n");
#endif
 
QplugAppendTextDisplay(">>CorCoef >> Input checked\n");

// load packet for CorCoef c routine:
ccpack.npts = ccpp->nxsub;
ccpp->nlags =  (int) floor(ccpp->dt_lag_max / ccpp->dt_spacing);
ccpack.nlags = ccpp->nlags;
ccpack.x = ccpp->x + ixstart;
ccpack.y = ccpp->y + iystart;
ccpp->out_array_size = 2*(ccpp->nlags)+1;
ccpp->rho = new float[ccpp->out_array_size];
ccpack.corcoef_results =ccpp->rho;
ccpp->rvecset = ccpp->xvecset + ccpp->yvecset;



// ... do it
CC_return = CorCoef(&ccpack);
if (CC_return != CORCOEF_OK){
  QplugAppendTextDisplay(CorCoefErrorMessage[CC_return]);
  return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);  
}


// ... now do the vector components one by one
if (ccpp->rvecset != 0) {
 ccpp->rho_vx = new float[ccpp->out_array_size];
 ccpp->rho_vy = new float[ccpp->out_array_size];
 ccpp->rho_vz = new float[ccpp->out_array_size];
 
 ccpack.corcoef_results =ccpp->rho_vx;
 if (ccpp->xvecset != 0) {
  ccpack.x = ccpp->x_vx + ixstart;
  if (ccpp->yvecset != 0) ccpack.y = ccpp->y_vx + iystart;
 }
 else ccpack.y = ccpp->y_vx + iystart;
 
 CC_return = CorCoef(&ccpack);
 if (CC_return != CORCOEF_OK){
  QplugAppendTextDisplay(CorCoefErrorMessage[CC_return]);
  return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);  
 }
 
 ccpack.corcoef_results =ccpp->rho_vy;
 if (ccpp->xvecset != 0) {
  ccpack.x = ccpp->x_vy + ixstart;
  if (ccpp->yvecset != 0) ccpack.y = ccpp->y_vy + iystart;
 }
 else ccpack.y = ccpp->y_vy + iystart;

 CC_return = CorCoef(&ccpack);
 if (CC_return != CORCOEF_OK){
  QplugAppendTextDisplay(CorCoefErrorMessage[CC_return]);
  return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);  
 }
 
 ccpack.corcoef_results =ccpp->rho_vz;
 if (ccpp->xvecset != 0) {
  ccpack.x = ccpp->x_vz + ixstart;
  if (ccpp->yvecset != 0) ccpack.y = ccpp->y_vz + iystart;
 }
 else ccpack.y = ccpp->y_vz + iystart;

 CC_return = CorCoef(&ccpack);
 if (CC_return != CORCOEF_OK){
  QplugAppendTextDisplay(CorCoefErrorMessage[CC_return]);
  return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);  
 }

}


QplugAppendTextDisplay("CorCoefPlug>> Calculation of Cor Coef Finished OK\n");
//printf("CorCoefPlug>> Calculation of Cor Coef Finished OK\n");


// Create and fill output data object

CCP_return = CorCoefPlug_create_out_DOs(outputs, ccpp);
if (CCP_return != CORCOEF_PLUG_SUCCESS){
  QplugAppendTextDisplay(CorCoefPlugReturnMessage[CCP_return]);
  return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);  
}  


//	..clean up and exit

QplugAppendTextDisplay ("CorCoefPlug>> - cleaning up ...\n");
return CorCoefPlug_cleanup (QPLUG_SUCCESS, ccpp);  
 
}  // end CorCoefPlug

/***********************************************************************
**************************************** CorCoefPlug_ini
***********************************************************************/

CorCoefPlugReturnStatus  CorCoefPlug_ini(CorCoefPlug_ptrs *ccpp){

ccpp->x = NULL;
ccpp->x_vx = NULL;
ccpp->x_vy = NULL;
ccpp->x_vz = NULL;

ccpp->tx=NULL;
ccpp->y = NULL;
ccpp->y_vx = NULL;
ccpp->y_vy = NULL;
ccpp->y_vz = NULL;
ccpp->ty=NULL;
ccpp->rho = NULL;
ccpp->rho_vx = NULL;
ccpp->rho_vy = NULL;
ccpp->rho_vz = NULL;
ccpp->dt_lag=NULL;

return CORCOEF_PLUG_SUCCESS;
}


/***********************************************************************
**************************************** CorCoefPlug_checkload_input
***********************************************************************/
//
// This routine:
//  o	Checks there are enough arguments on the list
//  o	Checks each input argument type
//  o	Puts x, y, interval data objects in the ccpp structure
//  o	Extracts ( x(i),tx(i) ) and ( y(i),ty(i) ) arrays into new'ed arrays
//  o	Extracts tstart and tend from subsetting interval and puts in ccpp
//  o	Extracts dt_lag_max and puts it into ccpp

CorCoefPlugReturnStatus  CorCoefPlug_checkload_input(vector <DvObject_var>& inputs, CorCoefPlug_ptrs *ccpp){


if (inputs.size() < 4){    // are there enough args?
   QplugAppendTextDisplay( "Not enough entries in argument list from GUI\n");
   return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}  //endif

// 1111111111111111111111111111111111111111111111111111111111111111111111

ccpp->xdo = inputs[0];
if(ccpp->xdo.is_nil() || ccpp->xdo->not_dbl() ){
      QplugAppendTextDisplay ("CorCoefPlug>> ERROR - X sequence type not handled\n");
      return CORCOEF_PLUG_CHECK_INPUT_ERROR; 
}

DvObject_var xtt = ccpp->xdo->getDep0();
if(xtt.is_nil()){
   QplugAppendTextDisplay("CorCoefPlug ERROR>> X not a Time Series\n");	
   return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}

if(ccpp->xdo->arraySize() == 1){
     ccpp->xvecset = 0;
}
else if ( ccpp->xdo->isThreeVector() ) {
	ccpp->xdo_vec = ccpp->xdo;
	ccpp->xvecset = 1;
	ccpp->xdo     = ccpp->xdo_vec->getVectorComponent(DV_RMAG); 
}
else {
	QplugAppendTextDisplay ("CorCoefPlug>> ERROR - sequence type not handled\n");
	return CORCOEF_PLUG_CHECK_INPUT_ERROR; 
}

ccpp->nxdata_tot = ccpp->xdo->seqSize();
if (ccpp->nxdata_tot <=0){
    QplugAppendTextDisplay("CorCoefPlug ERROR>> nxdata_tot <=0\n");	
    return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}

ccpp->x = new float[ccpp->nxdata_tot];
ccpp->tx = new double[ccpp->nxdata_tot];

for (int i =0; i<ccpp->nxdata_tot; i++){
  ccpp->x[i] = (float) ccpp->xdo->dbl(i);
  ccpp->tx[i] = xtt->asDouble(i);
}

if (ccpp->xvecset != 0) {
  ccpp->x_vx = new float[ccpp->nxdata_tot];
  ccpp->x_vy = new float[ccpp->nxdata_tot];
  ccpp->x_vz = new float[ccpp->nxdata_tot];
  for (int i =0; i<ccpp->nxdata_tot; i++){
   ccpp->x_vx[i] = (float) ccpp->xdo_vec->dbl(i, 0);
   ccpp->x_vy[i] = (float) ccpp->xdo_vec->dbl(i, 1);
   ccpp->x_vz[i] = (float) ccpp->xdo_vec->dbl(i, 2);
  }
}

// 222222222222222222222222222222222222222222222222222222222222222222222

ccpp->ydo = inputs[1];
if(ccpp->ydo.is_nil() || ccpp->ydo->not_dbl() ){
      QplugAppendTextDisplay ("CorCoefPlug>> ERROR - Y sequence type not handled\n");
      return CORCOEF_PLUG_CHECK_INPUT_ERROR; 
}

DvObject_var ytt = ccpp->ydo->getDep0();
if(ytt.is_nil()){
   QplugAppendTextDisplay("CorCoefPlug ERROR>> Y not a Time Series\n");	
   return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}

if(ccpp->ydo->arraySize() == 1){
     ccpp->yvecset = 0;
}
else if ( ccpp->ydo->isThreeVector() ) {
	ccpp->ydo_vec = ccpp->ydo;
	ccpp->yvecset = 1;
	ccpp->ydo     = ccpp->ydo_vec->getVectorComponent(DV_RMAG); 
}
else {
	QplugAppendTextDisplay ("CorCoefPlug>> ERROR - sequence type not handled\n");
	return CORCOEF_PLUG_CHECK_INPUT_ERROR; 
}

ccpp->nydata_tot = (ccpp->ydo)->seqSize();
if (ccpp->nydata_tot <=0){
    QplugAppendTextDisplay("CorCoefPlug ERROR>> nydata_tot <=0\n");	
    return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}

ccpp->y = new float[ccpp->nydata_tot];
ccpp->ty = new double[ccpp->nydata_tot];

for (int i =0; i<ccpp->nydata_tot; i++){
  ccpp->y[i] = (float) ccpp->ydo->dbl(i);
  ccpp->ty[i] = ytt->asDouble(i);
}

if (ccpp->yvecset != 0) {
  ccpp->y_vx = new float[ccpp->nydata_tot];
  ccpp->y_vy = new float[ccpp->nydata_tot];
  ccpp->y_vz = new float[ccpp->nydata_tot];
  for (int i =0; i<ccpp->nydata_tot; i++){
	ccpp->y_vx[i] = (float) ccpp->ydo_vec->dbl(i,0);
	ccpp->y_vy[i] = (float) ccpp->ydo_vec->dbl(i,1);
	ccpp->y_vz[i] = (float) ccpp->ydo_vec->dbl(i,2);
  }
}


// 333333333333333333333333333333333333333333333333333333333333333333333

ccpp->interval_do = new DvObject(inputs[2]->getTimeRange());

if ( (ccpp->interval_do).is_nil() ) { 
 QplugAppendTextDisplay("CorCoefPlug ERROR>> Third arg has no Time Interval\n");
   return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}

ccpp->tstart = ccpp->interval_do->event(0).start().getEpoch2000() ;
ccpp->tend =   ccpp->interval_do->event(0).end().getEpoch2000();


// 444444444444444444444444444444444444444444444444444444444444444444444

DvObject_var lag_ptr = inputs[3];

if ( lag_ptr.is_nil() || (lag_ptr->not_dbl() && lag_ptr->not_int()) ) { 
   QplugAppendTextDisplay("CorCoefPlug ERROR>> Fourth arg not a float");
   return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}

// epoch2000 works in seconds
ccpp->dt_lag_max = (float) lag_ptr->asDouble(0);

if (ccpp->dt_lag_max > 0.5* fabs(ccpp->tend - ccpp->tstart)) {
  QplugAppendTextDisplay("CorCoefPlug: dt_lag_max greater than 1/2 subsetted interval\n");
  QplugAppendTextDisplay("dt_lag_max set to 1/4 subsetted interval\n");
  ccpp->dt_lag_max = 0.25*fabs(ccpp->tend - ccpp->tstart);
}

return CORCOEF_PLUG_SUCCESS;
}  //end CorCoefPlug_checkload_input


/***********************************************************************
**************************************** CorCoefPlug_SubsetXY
***********************************************************************/

CorCoefPlugReturnStatus  CorCoefPlug_SubsetXY (CorCoefPlug_ptrs *ccpp){

IndexTSSubsetErrorStatus index_ok;
char out_message[500];
int iend;

// subset x array:

index_ok = IndexTSSubset (ccpp->tstart, ccpp->tend, 
              ccpp->tx, ccpp->nxdata_tot, 
              &(ccpp->ixstart), &iend, out_message);
if (index_ok != INDEX_TSSUBSET_OK){
   QplugAppendTextDisplay ("CorCoefPlug Error Subsetting x. Message:\n");
   QplugAppendTextDisplay (out_message);
   return CORCOEF_PLUG_SUBSET_ERROR;
}
ccpp->nxsub = iend - (ccpp->ixstart) + 1;

index_ok = IndexTSSubset (ccpp->tstart, ccpp->tend, 
              ccpp->ty, ccpp->nydata_tot, 
              &(ccpp->iystart), &iend, out_message);
if (index_ok != INDEX_TSSUBSET_OK){
   QplugAppendTextDisplay ("CorCoefPlug Error Subsetting y. Message:\n");
   QplugAppendTextDisplay (out_message);
   return CORCOEF_PLUG_SUBSET_ERROR;
}
ccpp->nysub = iend - (ccpp->iystart) + 1;

return CORCOEF_PLUG_SUCCESS;
}


/***********************************************************************
**************************************** CorCoefPlug_create_out_DOs
***********************************************************************/
CorCoefPlugReturnStatus      CorCoefPlug_create_out_DOs(vector <DvObject_var>& outputs, CorCoefPlug_ptrs *ccpp)
{

DvObject_var rho_out;
DvObject_var rho_vecout;
DvObject_var dt_lags_out;
vector <size_t> dims;
dims.push_back(3);

float  *dtzerop, n2dtlag;
double rhomax=-1.0, rho_vecmax[3]={-1.0,-1.0,-1.0};
double rhomaxlag=0.0,rho_vecmaxlag[3]={0.0,0.0,0.0};
int i, nlags, jmax,jmaxx,jmaxy,jmaxz;

int n_min_attrs=5;    // attribute shufflers

const char  *min_attr_ptr[]={"FIELDNAM", "Frame",    "SI_conversion", "UNITS",    "LABLAXIS"  };         
const char *rho_atts[]={"Correlation Coefficient", "scalar>na", "1>(cor coef)", " ","Cor Coef"};
const char *rho_v_atts[]={"Correlation Coefficient Comp", "vector>na_xyz", "1>(cor coef)", " ", "Cor Coef"};
const char *dt_atts[]={"Lag time (Var 1 leads Var2)", "scalar>na", "1>s", "seconds",  "Lag (V1 leads V2)"};

string fieldnamex, fieldnamey, fields_string, dt_lags_out_fieldnam, tinterval_string;
char	str_out[150];


nlags = ccpp->nlags;
ccpp->dt_lag = new float[ccpp->out_array_size];

dtzerop = ccpp->dt_lag + nlags;
n2dtlag = ccpp->dt_spacing;
for (i=-nlags; i<nlags+1; i++){
   dtzerop[i]=i*n2dtlag;
}                                                                                                                                                                                                                                                                                                                                                                                                                                                      // Create the output data Series and fill them. Add attributes

rho_out     = new DvObject( (double)0.0, ccpp->out_array_size);
dt_lags_out = new DvObject( (double)0.0, ccpp->out_array_size );

for(int j =0; j<ccpp->out_array_size ; j++){
  rho_out->dbl(j) = (double) ccpp->rho[j];
  if (ccpp->rho[j] >= rhomax) {
    jmax=j;
	rhomax=ccpp->rho[j];
  }
  dt_lags_out->dbl(j) = (double) ccpp->dt_lag[j];
}

rhomaxlag=ccpp->dt_lag[jmax];

if (ccpp->rvecset !=0) {
  rho_vecout = new DvObject( (double) 0.0, dims, ccpp->out_array_size);
  for(int j =0; j<ccpp->out_array_size ; j++) {
	rho_vecout->dbl(j,0) = (double) ccpp->rho_vx[j];
	if (ccpp->rho_vx[j] >= rho_vecmax[0]){
		jmaxx=j;
		rho_vecmax[0]=ccpp->rho_vx[j];
	}
	rho_vecout->dbl(j,1) = (double) ccpp->rho_vy[j];
	if (ccpp->rho_vy[j] >= rho_vecmax[1]){
		jmaxy=j;
		rho_vecmax[1]=ccpp->rho_vy[j];
	}
	rho_vecout->dbl(j,2) = (double) ccpp->rho_vz[j];
	if (ccpp->rho_vz[j] >= rho_vecmax[2]){
		jmaxz=j;
		rho_vecmax[2]=ccpp->rho_vz[j];
	}
  }
  rho_vecmaxlag[0]=ccpp->dt_lag[jmaxx];
  rho_vecmaxlag[1]=ccpp->dt_lag[jmaxy];
  rho_vecmaxlag[2]=ccpp->dt_lag[jmaxz]; 
}

// print maximal values to plugin window and append as xref:
sprintf(str_out,"CorCoef for Magnitudes>> max. correlation:%f, at  timelag: %f \n", rhomax, rhomaxlag);
QplugAppendTextDisplay (str_out);
DvObject_var rhomax_var = new DvObject(rhomax);
rho_out->change_xref("MaxCor", rhomax_var);
DvObject_var rhomaxlag_var = new DvObject(rhomaxlag);
rho_out->change_xref("LagMax", rhomaxlag_var);

if (ccpp->rvecset !=0) {
 sprintf(str_out,"CorCoef for X-Component>> max. correlation:%f, at  timelag: %f \n", rho_vecmax[0],rho_vecmaxlag[0]);
 QplugAppendTextDisplay (str_out); 
 sprintf(str_out,"CorCoef for Y-Component>> max. correlation:%f, at  timelag: %f \n", rho_vecmax[1],rho_vecmaxlag[1]);
 QplugAppendTextDisplay (str_out); 
 sprintf(str_out,"CorCoef for Z-Component>> max. correlation:%f, at  timelag: %f \n", rho_vecmax[2],rho_vecmaxlag[2]);
 QplugAppendTextDisplay (str_out);
 DvObject_var rho_vecmax_var = new DvObject((double) 0.0, dims, 1);
 for (int i=0; i<3; i++) rho_vecmax_var->dbl(i) = rho_vecmax[i];
 rho_vecout->change_xref("MaxCor", rho_vecmax_var);
 DvObject_var rho_vecmaxlag_var = new DvObject((double) 0.0, dims, 1);
 for (int i=0; i<3; i++) rho_vecmaxlag_var->dbl(i) = rho_vecmaxlag[i];
 rho_vecout->change_xref("LagMax", rho_vecmaxlag_var);
}


for (i=0; i<n_min_attrs; i++){
   rho_out->change_xref(min_attr_ptr[i], rho_atts[i] );
   if (ccpp->rvecset != 0) rho_vecout->change_xref(min_attr_ptr[i], rho_v_atts[i]);
   dt_lags_out->change_xref(min_attr_ptr[i], dt_atts[i] );
}

DvTime ts(ccpp->tstart);
DvTime te(ccpp->tend);

tinterval_string = "Time Interval: ";
tinterval_string += ts.getISOstring(0).c_str();
tinterval_string += " to ";
tinterval_string += te.getISOstring(0).c_str();
rho_out->change_xref("Time_Interval", tinterval_string.c_str());
if (ccpp->rvecset != 0) rho_vecout->change_xref("Time_Interval", tinterval_string.c_str());


if (ccpp->xvecset != 0)  fieldnamex = ccpp->xdo_vec->getXrefText("FIELDNAM").c_str();
else fieldnamex = ccpp->xdo->getXrefText("FIELDNAM").c_str();
if (ccpp->yvecset != 0)   fieldnamey = ccpp->ydo_vec->getXrefText("FIELDNAM").c_str();
else fieldnamey = ccpp->ydo->getXrefText("FIELDNAM").c_str();

if ( fieldnamex.empty()  || fieldnamey.empty() ){
  rho_out->change_xref("Analysis_Notes", "1st data series leads 2nd");  
  if (ccpp->rvecset != 0) rho_vecout->change_xref("Analysis_Notes", "1st data series leads 2nd");
}
else{ 
  fields_string = fieldnamex;
  fields_string += " leads ";
  fields_string += fieldnamey;
  rho_out->change_xref("Analysis_Notes", fields_string.c_str());
  if (ccpp->rvecset != 0) rho_vecout->change_xref("Analysis_Notes", fields_string.c_str());  

  dt_lags_out_fieldnam = "Lag Time ";
  dt_lags_out_fieldnam += fields_string;
  dt_lags_out->change_xref("FIELDNAM", dt_lags_out_fieldnam.c_str());  
}

//  args to return   

  rho_out->change_xref(DEPEND_0,dt_lags_out);
  if (ccpp->xvecset != 0)   rho_vecout->change_xref(DEPEND_0,dt_lags_out);
  outputs.push_back(rho_out);
  outputs.push_back(rho_vecout);
  
return CORCOEF_PLUG_SUCCESS;
}


// *****************************************************************
// *****************************************************************
// ***************************************** CorCoefPlug_cleanup **
// *****************************************************************
// This routine cleans up after CorCoefPlug, deleting all new'ed memory
// See SJS Storage Management Notes
// *****************************************************************
//

QplugReturnStatus  CorCoefPlug_cleanup  (QplugReturnStatus returnstatus,
                           CorCoefPlug_ptrs *ccpp)
{

delete[]	ccpp->x;
delete[]	ccpp->x_vx;
delete[]	ccpp->x_vy;
delete[]	ccpp->x_vz;
delete[]	ccpp->tx;
delete[]	ccpp->y;
delete[]	ccpp->y_vx;
delete[]	ccpp->y_vy;
delete[]	ccpp->y_vz;
delete[]	ccpp->ty;
delete[]	ccpp->rho;
delete[]	ccpp->rho_vx;
delete[]	ccpp->rho_vy;
delete[]	ccpp->rho_vz;
delete[]	ccpp->dt_lag;

return returnstatus;

}  // end CorCoefPlug_cleanup


