#include <stdio.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"
#include <qdatetime.h>

#define MVAB  = 0     // minimum variance of B-field
#define MFR   = 1     // minimum faraday residue
#define MMR   = 2     // minimum massflux residue
#define MVAJ  = 3     // minimum variance of current density
#define MLMR  = 4     // minimum linear momentum flux residue
#define MTER  = 5     // minimum total energy residue
#define MER   = 6     // minimum entropy residue
#define MVAV  = 7     // minimum variance of plasma velocity

// --- prototypes ----

int SingleAnalysis(DvObject_var &V, DvObject_var &B, DvObject_var &dB, DvObject_var &P,
		   DvObject_var &E, DvObject_var &N, DvObject_var &H,
		   DvObject_var &ttJoined, DvObject_var &ExWeights,
		   int constrained, double alpha, DvString SCs_ID,
		   bool doType[10],int nestsize,
           vector <DvObject_var>& outputs);

int NestedAnalysis(DvObject_var &V_, DvObject_var &B_, DvObject_var &dB_, DvObject_var &P_,
		   DvObject_var &E_, DvObject_var &N_, DvObject_var &H_,
   		   DvObject_var &ttJoined, DvObject_var &ExWeights,
		   int constrained, double alpha, int min_nest, DvString SCs_ID,
		   bool doType[10],int nestsize);

DvObject_var Q_MVAB_calc(DvObject_var &B, DvObject_var &V, DvObject_var &U, DvObject_var &Valf,
			  DvObject_var &N, double *wl, double *cc, int nestsize, bool doHT);

DvObject_var Q_MLMR_calc(DvObject_var &V, DvObject_var &B, DvObject_var &P,
			  DvObject_var &N,  DvObject_var &U, int nestsize);

DvObject_var Q_MER_calc(DvObject_var &V, DvObject_var &B, DvObject_var &P,
			 DvObject_var &N, DvObject_var &U, int nestsize);

DvObject_var Q_MFR_calc(DvObject_var &B, DvObject_var &V, DvObject_var &U, int nestsize);

DvObject_var Q_MTER_calc(DvObject_var &V, DvObject_var &B, DvObject_var &P, 
			  DvObject_var &E, DvObject_var &N, DvObject_var &H, 
			  DvObject_var &U, int nestsize);

DvObject_var Q_MMR_calc(DvObject_var &V,  DvObject_var &N, DvObject_var &U, int nestsize);

DvObject_var Q_MVAV_calc(DvObject_var &V, DvObject_var &U, int nestsize);

DvObject_var Projection(DvObject_var &B, DvObject_var &V, DvObject_var &U,   
			 DvObject_var &V_Alfven, DvObject_var &Evec, double alpha, 
			 int nestsize, int constrain, const double dens = 0.0);

double aveTotalPressure(DvObject_var &B, DvObject_var &P, int nestsize);
void Report(DvObject_var &Evalues, DvObject_var &Evec, bool outstd, const char *msg, int constrain, double alpha,  const char *SCc_ID, const bool changeOrder=false);

void ReportSiscoe(DvObject_var &n, DvObject_var &U,
		 DvObject_var &Bdave, DvObject_var &Vdave, DvObject_var &Vadave,
		 int nestsize, const char *msg, int constrain, double alpha, const char *SCc_ID);

DvObject_var LeviCivita();

void ToSpreadsheet(const char *filename, DvObject_var &Normal, DvObject_var &U, int row,  bool nl);

DvObject_var V_HofTel(DvObject_var B, DvObject_var V, int nestsize);

double slope(DvObject_var X_in, DvObject_var Y_in,int n);

double correlationCoeff(DvObject_var X_in, DvObject_var Y_in,int n);

double weight_k_calc(DvObject_var &eval);

double sigma(DvObject_var &V,DvObject_var &normal,int n);

double err_824(DvObject_var &B_ave, DvObject_var &Eigenvectors, DvObject_var &Eigenvalues,
	       DvObject_var &dYij, DvObject_var &Q_MVA, double sigma, int nestsize, int modify);

double jumpvalue(DvObject_var &V, const int n=5);

DvObject_var Q_GEN_calc(DvObject_var &TS);

double vecsum(DvObject_var Vector);

double vecsum2(DvObject_var Vector);
