
// -----------------------------------------------------------------------------------------
//
// Calculate Q-matrix for MVA of any vector Quantity
//
// SEH, Mar 27, 2003
//      Sep 13, 2003  - cloned fromm massflow and modified 
//      Sep 26, 2005  - MVAV version
//
// Inputs :
//     V = any vector time series, e.g., magnetic field   (n x 3)     
//
// -----------------------------------------------------------------------------------------

#include "qdutil.h"
#include "qplug_if.h"

#include "GenericMinVar.h"
#include "GenericResidue.h"
#include "DvObj.h"

// -----------------------------------------------------------------------------------------
//
// Returns the Q_matrix for MVAV (Minimum Variance of Velcity)
//
// SEH, 26 Sep, 2005, initiated from Q_MVAB_calc
// Doxygen info :
/*! 
  \brief Calculates the covariance matrix of the velocity.
  \param V -  vector time series with plasma velocity in units of km/s
  \param U -  holds the calculated velocity on successful return
  \param nestsize - number of elements in B and V.

*/ 

DvObject_var Q_MVAV_calc(DvObject_var &V, DvObject_var &U, int nestsize)
{
  int i,j,k;
  double normfac = (double) nestsize;
  DvObject_var Q  = new DvObject(0.0, DIM33);
  DvObject_var dV  = new DvObject(0.0, DIM3, nestsize);
  DvObject_var aveV  = new DvObject(0.0, DIM3);
  bool PO = true;                           // PrintOut to screen ? 


  for (k=0; k<nestsize; k++) {              // find <B>
    for (i=0; i<3; i++) {
      aveV->dbl(i) += V->dbl(k,i) / normfac;;
    }
  }

  if (PO)printf("Q_MVAV : ave V : [%E  %E  %E]\n",aveV->dbl(0), aveV->dbl(1), aveV->dbl(2));
  for (k=0; k<nestsize; k++) {             // find delta B
    for (i=0; i<3; i++) {
      dV->dbl(k,i) = V->dbl(k,i) - aveV->dbl(i);
    }
  }

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) Q->dbl(0,i,j) += dV->dbl(k,i) * dV->dbl(k,j) / normfac;
    }
   }


  // ----- calculate U == ave V ------

  for (i=0; i<3; i++) U->dbl(i) = aveV->dbl(i);
  if (PO)printf("Q_MVAV : ave V : [%E  %E  %E]\n",U->dbl(0), U->dbl(1), U->dbl(2));

  return (Q);
}





