#include "QbEditors.h"

#include "qdutil.h"

#include "QbTimeWidgets.h"


QbEditor::QbEditor(DvObject_var &dobj, QWidget *parent ) 
	: QGroupBox(parent)
{
	
	this->setPalette(qApp->palette());
	QFontMetrics qbu_font_mt(qApp->font());
	int char_width = qbu_font_mt.width("n");

	QVBoxLayout * mainLayout=new QVBoxLayout;
	
	valueBox = new QWidget(this);
	QHBoxLayout * valueLayout=new QHBoxLayout;
	valueLayout->setSpacing(0);

	valueLayout->setContentsMargins(0,0,0,0);

	valueLabel = new QLabel("  Value ");
	valueLayout->addWidget(valueLabel);
	
	valueField = new QguWLinput(valueBox,"value");
	valueField->setMinimumWidth(char_width*23);
	valueField->setEnabled(true);
	valueLayout->addWidget(valueField);
	
	valueBox->setLayout(valueLayout);
	mainLayout->addWidget(valueBox);

	if(dobj->is_dbl())
	{	
		setTitle("Double Data");
		valueField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));
		
	}
	else if(dobj->is_str())
	{	
		setTitle("Text");
		valueLabel->hide();
		valueField->setMinimumWidth(450);
		valueField->setAcceptable(Qcondition(QS_NARROW, QS_TEXT, QS_0D, QS_NOT_SEQ, QS_ANY));
	}
	else if(dobj->is_int())
	{
		setTitle("Integer Data");
		valueField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));		
	}
	else if(dobj->is_time())
	{
		setTitle("Time Data");
  
		valueField->setAcceptable(Qcondition(QS_NARROW, QS_TIME, QS_0D, QS_NOT_SEQ, QS_ANY));
		valueField->setMinimumWidth(250);
		valueField->adjustSize();
		connect(valueField, SIGNAL(TimeChosen()), this, SLOT(setTimeValue())  );
	}

    
	dep0Box = new QWidget(this);
	QHBoxLayout * timeLayout=new QHBoxLayout;
	timeLayout->setSpacing(0);

	timeLayout->setContentsMargins(0,0,0,0);
	timeField = new QbDateTimeWidget(dep0Box);
	timeField->setMinimumWidth(char_width*30);

	SdepField = new QguWLinput(dep0Box,"dep0");
	SdepField->setMinimumWidth(char_width*23);
	SdepField->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));

	D0 = dobj->getDep0();
    
	makeCopy(dobj);

	if(D0->is_time()) {
		QLabel * timeLabel = new QLabel("at Time ");
		timeLayout->addWidget(timeLabel, Qt::AlignRight);
		timeLayout->addWidget(timeField);
		timeField->setEnabled(true);
		SdepField->setEnabled(false);
		SdepField->setVisible(false);
	}
	else if(D0->is_dbl()) {
		QLabel * atLabel = new QLabel("at ");
		timeLayout->addWidget(atLabel, Qt::AlignRight);
		timeLayout->addWidget(SdepField);
		SdepField->setEnabled(true);
		timeField->setEnabled(false);
		timeField->setVisible(false);
	}
	
	dep0Box->setLayout(timeLayout);
	mainLayout->addWidget(dep0Box);
  
	
	this->setLayout(mainLayout);
	
	maxRec = 0;
	record = 0;
	connect(valueField, SIGNAL(edited_OK()), this, SLOT(valueChanged()) );
	connect(timeField,  SIGNAL(valueChanged()), this, SLOT(timeValueChanged()) );
	connect(SdepField,  SIGNAL(edited_OK()), this, SLOT(SdepValueChanged()) );

	setData(dobj);

}


QbEditor::~QbEditor(){
}

void QbEditor::setTimeValue(){
	DvObject_var tobj = valueField->getObject();
	if( tobj.is_ok() && tobj->is_time()){
		valueField->setText(tobj->time(0).getISOstring(0).c_str());
	}
   
}



void QbEditor::showValueField(bool OKtoShow){
   
	if ( OKtoShow ){
		valueBox->show();
	}
	else{
		valueBox->hide();
	}
   
	return;
}

void QbEditor::showDep0Field(bool OKtoShow){
   
	if ( OKtoShow ){
		dep0Box->show();
	}
	else{
		dep0Box->hide();
	}
   
	return;
}


void QbEditor::setData(DvObject_var &dobj){
	QString txt;

	if( dobj.is_nil() ) return;
   
	makeCopy(dobj);
	
	if(dobj->is_event()) return; // event editor has its own setData()
	
	maxRec = dobj->seqSize() - 1;
	record = 0;

	dimSize = dobj->nDims();
	index.clear();

	// add array sliders if needed
	for (int i=0; i<dimSize; i++){
		int maxN = dobj->Dims()[i];
		if(maxN == 1) continue; // scalar
		int nextInd = 0;
		index.push_back(nextInd);
		argSlider * argS = new argSlider(tr("index %1").arg(i+1), i, maxN, this);
		connect(argS, SIGNAL(QbEchanged(int, int)), this, SLOT(changeElement(int, int)) );

		layout()->addWidget(argS);
	}
	if(index.size() == 0) index.push_back(0);
		
	if( dobj->is_dbl() ){
		double value = dobj->dbl(0);
		txt.setNum(value);
		valueField->setText(txt);
	}
	else if( dobj->is_int() ){
		int value = dobj->itg(0);
		txt.setNum(value);
		valueField->setText(txt);
	}
	else if( dobj->is_time() ){
		valueField->setText(dobj->time(0).getISOstring(0).c_str());
	}
	else if( dobj->is_str() ){
		valueField->setText(dobj->str(0).c_str());
		valueField->setMinimumWidth(450);
		valueField->setCursorPosition(0);
	}
		
	D0 = dobj->getDep0();
		
	if ( D0.is_nil() ){
		showDep0Field(false);
	}
	else{
		showDep0Field(true);
		if ( D0->seqSize() == dobj->seqSize() )
		{
			if(D0->is_time()) timeField->setTime(D0->time(0));
			else if(D0->is_dbl()) SdepField->setDouble(D0->dbl(0));
		}

		else
			QuMsgBar("Wrong number of timetags/Depend_0", true);
	}
   
	((QHBoxLayout*)layout())->addStretch(1);
}



void QbEditor::valueChanged(){
    
	if(QuSlotDebug())
	{
		printf("QbEditor::valueChanged %s\n", valueField->text().toLatin1().data());
		fflush(stdout);
	}
 	bool OK;
 
	if(edited_var->is_dbl())
	{
		double val = valueField->text().toDouble(&OK);
		if (OK){
			edited_var->dbl(record, index) = val;
		}
	}
	else if(edited_var->is_str())
	{
		edited_var->str(record, index) = valueField->text().toLatin1().data();
	}
	else if(edited_var->is_int())
	{
		int val = valueField->text().toInt(&OK);
		if (OK){
			edited_var->itg(record, index) = val;
		}

	}
	else if(edited_var->is_time())
	{
		DvTime tag;
		if( tag.setFromISOstring(valueField->text().toLatin1().data() ) ){
			edited_var->time(record, index) = tag;
		}
	}
	
}

void QbEditor::timeValueChanged(){
 
	DvTime tag=timeField->time();
	if(D0.is_ok() && D0->is_time() && record < D0->seqSize()) {
		D0->time(record) = tag;
	}

}

void QbEditor::SdepValueChanged(){
 
	double val = SdepField->value();
	if(D0.is_ok() && D0->is_dbl() && record < D0->seqSize()) {
		D0->dbl(record) = val;
	}

}


void QbEditor::makeCopy(DvObject_var &dobj){

	if( dobj.is_nil() ) return;
   
	edited_var = new DvObject(dobj);
          
	DvObject_var D0_orig = edited_var->getDep0();
	if ( D0_orig.is_ok() ){
			
		D0 = new DvObject(D0_orig); // ensure we don't edit tags shared with another object
		if( D0.is_ok() ) edited_var->change_xref(DEPEND_0, D0);
	}

}

void QbEditor::updateForRec(){
	
	if(edited_var->is_dbl())
	{	
		QString txt;
		double value = edited_var->dbl(record, index);
		txt.setNum(value);
		
		valueField->setText(txt);
	}
	else if(edited_var->is_str())
	{	
		valueField->setText(edited_var->str(record, index));
	}
	else if(edited_var->is_int())
	{
		QString txt;
		int value = edited_var->itg(record, index);
		txt.setNum(value);
		
		valueField->setText(txt);
	}
	else if(edited_var->is_time())
	{
		valueField->setText(edited_var->time(record, index).getISOstring(0).c_str());
	}

	if ( D0.is_nil() ){
		showDep0Field(false);
	}
	else{
		showDep0Field(true);
		if ( D0->seqSize() > record )
		{
			if(D0->is_time()) timeField->setTime(D0->time(record));
			else if(D0->is_dbl()) SdepField->setDouble(D0->dbl(record));
		}
	}
}


void QbEditor::changeElement(int id, int idx){
	index[id] = idx;
   
	if(edited_var->is_dbl())
	{	
		QString txt;
		double value = edited_var->dbl(record, index);
		txt.setNum(value);
		
		valueField->setText(txt);
	}
	else if(edited_var->is_str())
	{	
		valueField->setText(edited_var->str(record, index));

	}
	else if(edited_var->is_int())
	{
		QString txt;
		int value = edited_var->itg(record, index);
		txt.setNum(value);
		
		valueField->setText(txt);
	}
	else if(edited_var->is_time())
	{
		valueField->setText(edited_var->time(record, index).getISOstring(0).c_str());
	}

}



// class argSlider

 argSlider::argSlider( QString title, int indx, int maxN, QWidget *parent, char * name)
	: QWidget(parent)
{

	maxNum = maxN-1;
	num = 0;
	id = indx;
   
	QGridLayout * layout=new QGridLayout(this);
	layout->setSpacing(0);
	layout->setMargin(0);
   
	QToolButton * leftSlide = new QToolButton(this);
	leftSlide->setArrowType(Qt::LeftArrow);
	leftSlide->setFixedSize(18, 20);
	layout->addWidget(leftSlide, 0,0);
   
	numField = new QguWLinput(this, "numField");
	numField->setAcceptable(Qcondition(QS_NARROW, QS_INT, QS_0D, QS_NOT_SEQ, QS_ANY));
	numField->setEnabled(true);
	numField->setText("1");
	numField->setFixedSize(40, 20);
	numField->setAlignment(Qt::AlignHCenter);
	layout->addWidget(numField, 0,1);
   
	QToolButton * rightSlide = new QToolButton(this);
	rightSlide->setArrowType(Qt::RightArrow);
	rightSlide->setFixedSize(18, 20);
	layout->addWidget(rightSlide, 0,2);
   
	QLabel * Spacer=new QLabel(this);
	Spacer->setFixedSize(18, 20);
	layout->addWidget(Spacer, 0,3);
   
	QLabel *sliderLabel = new QLabel(tr("in range [1,%1]").arg(maxNum+1), this);
	layout->addWidget(sliderLabel, 0, 4);
   
	layout->setColumnStretch(4, 1);
   
	connect( rightSlide, SIGNAL(clicked()), this, SLOT(incRec()) );
	connect( leftSlide,  SIGNAL(clicked()), this, SLOT(decRec()) );
	connect( numField,   SIGNAL(edited_OK()), this, SLOT(adjustRec()) );
   
}

argSlider::~argSlider(){
	delete numField;
}

void argSlider::incRec(){
	if(QuSlotDebug())
	{
		printf("argSlider::incRec \n");
		fflush(stdout);
	}
   
	num += 1;
	numField->setText(tr("%1").arg(num+1));
	adjustRec();  // done explicitly as textChanged() is not emitted if slot does not have focus
}

void argSlider::decRec(){
	if(QuSlotDebug())
	{
		printf("argSlider::decRec \n");
		fflush(stdout);
	}
    
	num -= 1;
	numField->setText(tr("%1").arg(num+1));
	adjustRec(); // done explicitly as textChanged() is not emitted if slot does not have focus
}

void argSlider::adjustRec(){
	if(QuSlotDebug())
	{
		printf("argSlider::adjustRec \n");
		fflush(stdout);
	}

	bool OK;
	num = numField->text().toInt(&OK);
	num -=1; // as feedback goes 1-N while code goes 0-N-1
	if(OK){
		if(num < 0) num = 0;
		if(num > maxNum) num = maxNum;
		numField->setText(tr("%1").arg(num+1));
		emit QbEchanged(id, num);
	}
	else
		numField->setText("1");
	emit QbEchanged(id, num);
}
