#include "QbTimeEditors.h"


QbTimeIntervalEditor::QbTimeIntervalEditor( DvObject_var dobj, QWidget *parent, QString name)
	:QbEditor(dobj, parent)
{
	
	if(!name.isEmpty()) setTitle(name);
	
	interval_edit=new TimeIntervalWidget(this, 1);

	int ptr = name.indexOf("/");
	if(ptr >= 0) name.remove(0, ptr+1); // remove "xxx/" 
	interval_edit->nameFld->setText(name);
	
	
	setData(dobj);
	
	this->layout()->addWidget(interval_edit);
	connect(interval_edit, SIGNAL(intervalUpdated()), this, SLOT(updateTI()));
}


void QbTimeIntervalEditor::updateTI()
{
	_time_interval_var->event(0).set(interval_edit->getStart(), interval_edit->getEnd());
		
	edited_var = _time_interval_var;

}

void QbTimeIntervalEditor::setData(DvObject_var &dobj)
{
	showDep0Field(false);
	showValueField(false);

	interval_edit->setInterval(dobj);

	if( dobj.is_nil() )
	{ 
		DvTime end_time;
		end_time.set_today();
		DvTime start_time(end_time-86400.);
		DvEvent ivl(start_time, end_time);
		_time_interval_var = new DvObject(ivl);
		return;
	}
	
   
	if( dobj->is_event() ){
		_time_interval_var = new DvObject(dobj);
	}
	else{
		DvEvent range = dobj->getTimeRange();
		_time_interval_var = new DvObject(range);
	}
}


void QbTimeIntervalEditor::set_time_interval(DvEvent& tintv, bool setViewRangeToo)
{
	interval_edit->setInterval(tintv, setViewRangeToo);
}

DvEvent QbTimeIntervalEditor::get_time_interval( ) 
{
	return interval_edit->getInterval();
}

void QbTimeIntervalEditor::set_view_range(DvEvent& tintv)
{
    interval_edit->setViewRange(tintv);
}

double QbTimeIntervalEditor::get_duration()
{
	return (interval_edit->getDuration());
}



// ----------------------------------------------------------



QbEventTableEditor::QbEventTableEditor( DvObject_var dobj, QWidget *parent, QString name )
	:QbEditor(dobj, parent)
{
	setTitle(name);
	
	interval_edit = new TimeIntervalWidget(this, 2);
	
	setData(dobj);
	
	interval_edit->dobjPath = name;

	this->layout()->addWidget(interval_edit);
	
}


void QbEventTableEditor::setData(DvObject_var &dobj)
{
	showDep0Field(false);
	showValueField(false);
	if(dobj.is_nil()) return;

	
	if(dobj->is_event() )
	{
		interval_edit->setInterval(dobj);
		interval_edit->setEventTable(dobj);
		maxRec=dobj->seqSize()-1;

		_event_table_var = new DvObject(dobj);
		
		return;
	}
	else
		QuMsgBar("Data object not recognised", true);
}



void QbEventTableEditor::set_view_range(DvEvent& tintv)
{
    interval_edit->setViewRange(tintv);
}

DvEvent QbEventTableEditor::get_time_interval( ) 
{
    return interval_edit->getInterval();
}

void QbEventTableEditor::updateForRec()
{
	if(!_event_table_var.is_nil())
	{
		interval_edit->setInterval(_event_table_var->event(record), false);
	}
}


void QbuLineEdit::focusOutEvent(QFocusEvent *e)
{
	const QValidator *v=validator();
	int curpos=cursorPosition();
	QString str=text();
	bool changed=str!=_old_text;
	if(changed)
	{
		if(v)
		{
			if(v->validate(str,curpos)==QValidator::Acceptable)
			{
				if(curpos!=cursorPosition())
					setCursorPosition(curpos);
				if(str!=text())
					setText(str);
			}
			else 
			{
				if(curpos!=cursorPosition())
					setCursorPosition(curpos);
				v->fixup(str);
				if(str!=text())
					setText(str);
			}
		}
		else
		{
			if(curpos!=cursorPosition())
				setCursorPosition(curpos);
			if(str!=text())
				setText(str);
		}
	}
	QLineEdit::focusOutEvent(e);
	if(changed) 
		emit focusOut();
}


