#include "QbTimeWidgets.h"
#include "qprofile.h"

QbTimeValidator::QbTimeValidator(QObject* parent):
	QValidator(parent)
{}

QbTimeValidator::~QbTimeValidator()
{}

void QbTimeValidator::fixup(QString& input) const
{
	DvTime test;
	if(test.setFromISOstring(input.toLatin1().data()))
	{
		input=test.getISOstring().c_str();
	}
}

QValidator::State QbTimeValidator::validate(QString& input, int& pos) const
{
	DvTime test;
	if(!test.setFromISOstring(input.toLatin1().data()))
	{
		return QValidator::Intermediate;
	}
	return QValidator::Acceptable;
}

////////////////// QbDurationWidget ////////////////////

QbDurationWidget::QbDurationWidget(QWidget * parent)
	: QLineEdit(parent)
{
	setAcceptDrops(true);
	//QRegExp rx("^(\\s*\\-?\\d+\\.?\\d*[dhms]\\s*)(\\s\\-?\\d+\\.?\\d*[dhms]\\s*)*$");
	QRegExp rx("^(\\s*\\-?\\d*\\.?\\d*[dhms]\\s*)(\\s\\-?\\d*\\.?\\d*[dhms]\\s*)*$");
	validator=new QRegExpValidator(rx, this);
	setValidator(validator);
	
	active=true;
	altPalette=false;
	
	activeColor=qApp->palette().color(QPalette::Active, QPalette::Light);
	inactiveColor=qApp->palette().color(QPalette::Active, QPalette::Midlight);
	activeText=qApp->palette().color(QPalette::Active, QPalette::Text);
    inactiveText=qApp->palette().color(QPalette::Inactive, QPalette::Text);
	activeColorAlt=qApp->palette().color(QPalette::Active, QPalette::Midlight);
	inactiveColorAlt=qApp->palette().color(QPalette::Active, QPalette::Button);
	
	QFontMetrics font_mt(qApp->font());
	int font_width = font_mt.width("n");
	setMinimumWidth(font_width*21);
	
	setFocusPolicy(Qt::StrongFocus);
}

bool QbDurationWidget::getDuration(double& res)
{
	QString current_text=text();
	int length=current_text.length();
	
	if(validator->validate(current_text,length)!=QValidator::Acceptable)
		return false;
	
	double d(0.), m(0.), h(0.), s(0.);
	
	QStringList sl=current_text.split(" ", QString::SkipEmptyParts);
	
	for (QStringList::iterator iter = sl.begin(); iter != sl.end(); ++iter)
	{
		if((*iter).endsWith("d"))
		{
			(*iter).chop(1);
			d+=(*iter).toDouble();
		}
		else if((*iter).endsWith("h"))
		{
			(*iter).chop(1);
			h+=(*iter).toDouble();
		}
		else if((*iter).endsWith("m"))
		{
			(*iter).chop(1);
			m+=(*iter).toDouble();
		}
		else if((*iter).endsWith("s"))
		{
			(*iter).chop(1);
			s+=(*iter).toDouble();
		}
	}
	
	res=d*24.*60.*60.+h*60.*60.+m*60.+s;
	return true;
}

void QbDurationWidget::setDuration(double seconds)
{
	int d, h, m;
	double s;
	d=(int)floor(seconds/(24.*3600.));
	seconds-=d*24.*3600.;
	h=(int)floor(seconds/3600.);
	seconds-=h*3600.;
	m=(int)floor(seconds/60.);
	s=seconds-m*60.;

	QString res;
	if(d!=0)
		res+=QString("%1d ").arg(d);
	if(h!=0)
		res+=QString("%1h ").arg(h);
	if(m!=0)
		res+=QString("%1m ").arg(m);
	if(s!=0.)
		res+=QString("%1s ").arg(s);
	setText(res);

}

void QbDurationWidget::focusInEvent(QFocusEvent *e)
{
	QLineEdit::focusInEvent(e);
	emit gotFocus();
}

void QbDurationWidget::setActive(bool active)
{
	if(this->active!=active)
	{
		if(active)
		{
			setAcceptDrops(true);
			if(altPalette)
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), activeColorAlt);
 				this->setPalette(palette);
			}
			else
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), activeColor);
 				this->setPalette(palette);
			}
			QPalette palette;
			palette.setColor(this->foregroundRole(), activeText);
			this->setPalette(palette);
		}
		else
		{
			setAcceptDrops(false);
			if(altPalette)
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), inactiveColorAlt);
 				this->setPalette(palette);
			}
			else
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), inactiveColor);
 				this->setPalette(palette);
			}
			QPalette palette;
			palette.setColor(this->foregroundRole(), inactiveText);
			this->setPalette(palette);
		}
		this->active=active;
	}
}

void QbDurationWidget::setAltPalette(bool alt)
{
	if(this->altPalette!=alt)
	{
		this->altPalette=alt;
		if(active)
		{
			setAcceptDrops(true);
			if(altPalette)
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), activeColorAlt);
 				this->setPalette(palette);
			}
			else
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), activeColor);
 				this->setPalette(palette);
			}
			QPalette palette;
			palette.setColor(this->foregroundRole(), activeText);
			this->setPalette(palette);
		}
		else
		{
			setAcceptDrops(false);
			if(altPalette)
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), inactiveColorAlt);
 				this->setPalette(palette);
			}
			else
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), inactiveColor);
 				this->setPalette(palette);
			}
			QPalette palette;
			palette.setColor(this->foregroundRole(), inactiveText);
			this->setPalette(palette);
		}
	}
}

void QbDurationWidget::dropEvent(QDropEvent *event)
{
	DvObject_var droppedObject = getWLObject(event->mimeData()->text().toLatin1().data());
	
	if(droppedObject.is_nil()) return;
	
	DvEvent range = droppedObject->getTimeRange();
	
	double t = range.duration();
	setDuration(t);
	return;
	
}

void QbDurationWidget::dragEnterEvent(QDragEnterEvent *event)
{
	if(!active) return;
	this->activateWindow();
	this->raise();

	if (event->mimeData()->hasText()) event->acceptProposedAction();
}


////////////////// QbDateTimeWidget ////////////////////

QbDateTimeLineEdit::QbDateTimeLineEdit ( QWidget * parent) : QLineEdit(parent)
{
	QString baseInputMask("9999-99-99 99:99:99.");
	for(int i=0; i<QSAS_profiles->current()->secDecimal; ++i)
		baseInputMask.append("9");
	setInputMask(baseInputMask);
	validator=new QbTimeValidator(this);
	setValidator(validator);

	setText("2001-01-01 00:00:00.000000000");
	
	active=true;
	altPalette=false;
	hasBeenEdited=false;
	connect(this, SIGNAL(textChanged(const QString&)), this , SLOT(slotValueChanged()));
	
	activeColor=QApplication::palette().color(QPalette::Active, QPalette::Light);//QColorGroup(QApplication::palette().active()).light();
	inactiveColor=QApplication::palette().color(QPalette::Active, QPalette::Midlight);//QColorGroup(QApplication::palette().active()).midlight();
	activeText=QApplication::palette().color(QPalette::Active, QPalette::Text);//QColorGroup(QApplication::palette().active()).text();
	inactiveText=QApplication::palette().color(QPalette::Inactive, QPalette::Text);//QColorGroup(QApplication::palette().inactive()).text();
	activeColorAlt=QApplication::palette().color(QPalette::Active, QPalette::Midlight);
	inactiveColorAlt=QApplication::palette().color(QPalette::Active, QPalette::Button);
	
	QFontMetrics font_mt(qApp->font());
	int font_width = font_mt.width("n");
	setMinimumWidth(font_width*baseInputMask.length());
	
	setDragEnabled(true);

	setFocusPolicy(Qt::StrongFocus);
}

void QbDateTimeLineEdit::keyPressEvent(QKeyEvent * e)
{
	QLineEdit::keyPressEvent(e);
	if(e->key()==Qt::Key_Up)
	{
		addToCurrent(1);
	}
	else if(e->key()==Qt::Key_Down)
	{
		addToCurrent(-1);
	}
}

void QbDateTimeLineEdit::wheelEvent(QWheelEvent* e)
{
	if(e->delta()>0)
	{
		addToCurrent(+1);
	}
	else
	{
		addToCurrent(-1);
	}
}

void QbDateTimeLineEdit::addToCurrent(int delta)
{
	int pos=cursorPosition();
	QString t=text();
	QStringList l=t.split(QRegExp(":|-|\\s|\\."));
	int selected=0;
	if(l.size()<6) return;
	int boundaries[6];

	for(int i=0; i<6; ++i)
	{
		boundaries[i]=0;
		for(int j=0; j<=i; ++j)
		{
			boundaries[i]+=l[j].length()+1;
		}
	}
	
	if(pos<boundaries[0])
		selected=0;
	else if(pos>=boundaries[0] && pos<boundaries[1])
		selected=1;
	else if(pos>=boundaries[1] && pos<boundaries[2])
		selected=2;
	else if(pos>=boundaries[2] && pos<boundaries[3])
		selected=3;
	else if(pos>=boundaries[3] && pos<boundaries[4])
		selected=4;
	else if(pos>=boundaries[4] && pos<boundaries[5])
		selected=5;
	else selected=6;

	int y, M, d, h, m, s, ms;
	double s_temp;
	previousTime.breakDownMJD(&y, &M, &d, &h, &m, &s_temp);
	s=(int)floor(s_temp);
	ms=(int)((s_temp-s)*1000.);

	switch(selected)
	{
		case 0:
			y+=delta;
			break;
		case 1:
			M+=delta;
			break;
		case 2:
			d+=delta;
			break;
		case 3:
			h+=delta;
			break;
		case 4:
			m+=delta;
			break;
		case 5:
			s+=delta;
			break;
		default:
			ms+=delta;
			break;
	}
	
	char iso[25];
	sprintf(iso, "%04d-%02d-%02d %02d:%02d:%02d.%03d", y, M, d, h, m, s, ms);

	DvTime newTime;
	if(newTime.setFromISOstring(iso))
	{
		setText(newTime.getISOstring(0).c_str());
		setCursorPosition(pos);
	}
}

DvTime QbDateTimeLineEdit::time()
{
	DvTime res;
	if(res.setFromISOstring(text().toLatin1().data()))
	{
		previousTime=res;
	}
	else
	{
		res=previousTime;
		setTime(res);
	}
	return res;
}

void QbDateTimeLineEdit::setTime(const DvTime& t)
{
	this->blockSignals(true);
	previousTime=t;
	setText(t.getISOstring().c_str());
	this->blockSignals(false);
}

void QbDateTimeLineEdit::slotValueChanged()
{
	int foo=0;
	QString t=text();
	if(validator->validate(t, foo)==QValidator::Acceptable)
	{
		QToolTip::showText(QPoint(), "", this, QRect());
		DvTime test;
		if(test.setFromISOstring(text().toLatin1().data()))
		{
            hasBeenEdited=true;
			previousTime=test;
			emit valueChanged();
		}
		else
		{
			setTime(previousTime);
		}
	}
	else
	{
		QToolTip::showText(mapToGlobal(pos()), "Invalid", this, QRect());
	}
}

void QbDateTimeLineEdit::slotValueInvalid()
{
	QToolTip::showText(mapToGlobal(pos()), "Invalid", this, QRect());
}

void QbDateTimeLineEdit::setActive(bool active)
{
	if(this->active!=active)
	{
		if(active)
		{
			setAcceptDrops(true);
			if(altPalette)
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), activeColorAlt);
 				this->setPalette(palette);
			}
			else
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), activeColor);
 				this->setPalette(palette);
			}
			QPalette palette;
			palette.setColor(this->foregroundRole(), activeText);
			this->setPalette(palette);
		}
		else
		{
			setAcceptDrops(false);
			if(altPalette)
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), inactiveColorAlt);
 				this->setPalette(palette);
			}
			else
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), inactiveColor);
 				this->setPalette(palette);
			}
			QPalette palette;
			palette.setColor(this->foregroundRole(), inactiveText);
			this->setPalette(palette);
		}
		this->active=active;
	}
}

void QbDateTimeLineEdit::setAltPalette(bool alt)
{
	if(altPalette!=alt)
	{
		altPalette=alt;
		if(active)
		{
			setAcceptDrops(true);
			if(altPalette)
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), activeColorAlt);
 				this->setPalette(palette);
			}
			else
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), activeColor);
 				this->setPalette(palette);
			}
			QPalette palette;
			palette.setColor(this->foregroundRole(), activeText);
			this->setPalette(palette);
		}
		else
		{
			setAcceptDrops(false);
			if(altPalette)
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), inactiveColorAlt);
 				this->setPalette(palette);
			}
			else
			{
				QPalette palette;
 				palette.setColor(this->backgroundRole(), inactiveColor);
 				this->setPalette(palette);
			}
			QPalette palette;
			palette.setColor(this->foregroundRole(), inactiveText);
			this->setPalette(palette);
		}
	}
}

void QbDateTimeLineEdit::focusInEvent(QFocusEvent *e)
{
	QLineEdit::focusInEvent(e);
	emit gotFocus();
}

void QbDateTimeLineEdit::mouseDoubleClickEvent ( QMouseEvent * event )
{
	int pos=cursorPosition();
	QString t=text();
	QStringList l=t.split(QRegExp(":|-|\\s|\\."));
	int selected=0;
	if(l.size()<6) return;
	int boundaries[6];

	for(int i=0; i<6; ++i)
	{
		boundaries[i]=0;
		for(int j=0; j<=i; ++j)
		{
			boundaries[i]+=l[j].length()+1;
		}
	}
	
	if(pos<boundaries[0])
		selected=0;
	else if(pos>=boundaries[0] && pos<boundaries[1])
		selected=1;
	else if(pos>=boundaries[1] && pos<boundaries[2])
		selected=2;
	else if(pos>=boundaries[2] && pos<boundaries[3])
		selected=3;
	else if(pos>=boundaries[3] && pos<boundaries[4])
		selected=4;
	else if(pos>=boundaries[4] && pos<boundaries[5])
		selected=5;
	else selected=6;

	setSelection((selected==0?0:boundaries[selected-1]), l[selected].length());
}

////////////////// QbDateTimeWidget ////////////////////////

QbDateTimeWidget::QbDateTimeWidget(QWidget* parent):
	QAbstractSpinBox(parent)
{
	editor=new QbDateTimeLineEdit(parent);
	setLineEdit(editor);
	setAcceptDrops(true);
	
	connect(editor, SIGNAL(gotFocus()), this, SIGNAL(gotFocus()));
	connect(editor, SIGNAL(valueChanged()), this, SIGNAL(valueChanged()));
	connect(editor, SIGNAL(textEdited(const QString&)), this, SIGNAL(textEdited(const QString&)));
	
	QFontMetrics font_mt(qApp->font());
	int font_width = font_mt.width("n");
	setMinimumWidth(editor->minimumWidth()+2*font_width); // 2 char width added for the right up/down arrows
}

void QbDateTimeWidget::stepBy(int steps)
{
	editor->addToCurrent(steps);
}

QAbstractSpinBox::StepEnabled QbDateTimeWidget::stepEnabled () const
{
	return QAbstractSpinBox::StepUpEnabled|QAbstractSpinBox::StepDownEnabled;
}


void QbDateTimeWidget::dropEvent(QDropEvent * event)
{
	droppedObject  = getWLObject(event->mimeData()->text().toLatin1().data());

	if(droppedObject.is_nil()) return;
	
	
	if(droppedObject->is_time() && droppedObject->seqSize() == 1) 
	{
		editor->setText(droppedObject->time(0).getISOstring().c_str());
		return;
	}
	
	if(droppedObject->is_time() || droppedObject->is_event())
	{
		QMenu * selectPopup=new QMenu;
		selectPopup->addAction("Start", this, SLOT(droppedObjStart()));
		selectPopup->addAction("Centre", this, SLOT(droppedObjCentre()));
		selectPopup->addAction("End", this, SLOT(droppedObjEnd()));
		selectPopup->addAction("Cancel");

		selectPopup->exec(QCursor::pos());
		return;
	}
	
	std::cout << "Dropped type not handled" << std::endl;
	
	return;
}

void QbDateTimeWidget::dragEnterEvent(QDragEnterEvent * e)
{
	if(!editor->active) return;
	this->activateWindow();
	this->raise();
	if (e->mimeData()->hasText()) e->acceptProposedAction();
}

void QbDateTimeWidget::droppedObjStart()
{
	if(droppedObject->is_event())
	{
		editor->setText(droppedObject->event(0).start().getISOstring().c_str());
		return;
	}
	
	if(droppedObject->is_time())
	{
		editor->setText(droppedObject->time(0).getISOstring().c_str());
		return;
	}
}

void QbDateTimeWidget::droppedObjCentre()
{
	if(droppedObject->is_event())
	{
		DvTime t = droppedObject->event(0).centre();
		editor->setText(t.getISOstring().c_str());
		return;
	}
	
	if(droppedObject->is_time())
	{
		DvTime t=droppedObject->time(0)+droppedObject->time(droppedObject->seqSize()-1).getDiffSecs(droppedObject->time(0))/2.;
		editor->setText(t.getISOstring().c_str());
		return;
	}
}

void QbDateTimeWidget::droppedObjEnd()
{	
	if(droppedObject->is_event())
	{
		editor->setText(droppedObject->event(0).end().getISOstring().c_str());
		return;
	}

	if(droppedObject->is_time())
	{
		editor->setText(droppedObject->time(droppedObject->seqSize()-1).getISOstring().c_str());
		return;
	}
}
