#include "qwui.h"
#include "qops.h"


// ---------- WList class --------	 
  
WList::WList(QWidget* parent):
	QTreeWidget(parent)
{
	setDragEnabled(true);
	setAcceptDrops(true);
	setDropIndicatorShown(true);
	setUniformRowHeights(true);
	setSelectionMode(QAbstractItemView::ExtendedSelection);
	setDragDropMode(QAbstractItemView::DragDrop);

	
	// Set Right Button Menu
	RButtonMenu=new QMenu(this);

	setEditTriggers(QAbstractItemView::NoEditTriggers); // only access through rename menu item
	// double click opens data object
	connect(this, SIGNAL(itemDoubleClicked(QTreeWidgetItem *, int)), this, SLOT(openObjectEditor(QTreeWidgetItem *, int)));
	connect(this, SIGNAL(itemChanged(QTreeWidgetItem *, int)), this, SLOT(updateObjectName(QTreeWidgetItem *, int)));
		
	_data=new WLItem("Data", NULL, true);
	addTopLevelItem(_data);
	
	_time=new WLItem("Time Intervals", NULL, true);
	addTopLevelItem(_time);
	
	_constants=new WLItem("Constants", NULL, true);
	addTopLevelItem(_constants);
	
	_temp=new WLItem("Temp", NULL, true);
	addTopLevelItem(_temp);
	
	
	sortReverse = false;
	sortOn = "name";

	objTypeSelector=NULL;
	sortTypeSelector=NULL;
	
	setRootIsDecorated(true);
	setHeaderHidden(true);
}

//-------------

WList::~WList(){
	clear();
}

//-------------

void WList::updateObjectName(QTreeWidgetItem *item, int column)
{
	if(item == NULL) return;
	qApp->processEvents();

	if( item->parent() == NULL) return; // top level item, don't rename

	QString oldName = ((WLItem*)item)->getName();
	QString oldPath = ((WLItem*)item)->getPath();

	if(((WLItem*)item)->countName(item->text(column)) > 1){
		QuMsgBar("Sorry name Exists: ", item->text(column).toLatin1().data(), true);
		item->setText(column, oldName);
	}
		
	((WLItem*)item)->updateName();
    emitObjChanged(oldPath);

}

//-------------

void WList::openObjectEditor(QTreeWidgetItem *qitem, int column)
{
	if(qitem!=NULL)
	{
		WLItem *item = (WLItem *) qitem;
		DvString path = item->getPath().toLatin1().data();
        
		DvObject_var obj = getWLObjExact(path);
        
		DvObject_var nilObj;
		DataBrowseWindow * browser;
		
        browser = QbuLaunchBrowse(nilObj, this, path.c_str(), obj, path.c_str());
			
		if(browser) connect(browser, SIGNAL(objChanged(const QString&)), this, SIGNAL(objChanged(const QString&)));

	}
}

//-------------

void WList::dragEnterEvent(QDragEnterEvent *event)
{
	if(QuSlotDebug()) printf("WList::dragEnterEvent\n");

	QTreeView::dragEnterEvent(event);

	event->acceptProposedAction();
	activateWindow();
	emit raiseWindow();

}



//-------------

void WList::deleteItem(WLItem* item)
{	
	if(item==NULL || item==_data || item==_time || item==_constants) return;

	qApp->processEvents();
	item->parent()->removeChild(item);
	delete item;
	return;
	
}



// Drag and drop tools
//-------------

QStringList WList::mimeTypes() const
{
	QStringList types;
	types << "application/wlitems";
	types << "application/xrefitems";
	types << "text/plain";
	return types;
}
//-------------

QMimeData* WList::mimeData(const QList<QTreeWidgetItem *> items ) const
{
	QMimeData *mimeData = new QMimeData();
	
	QStringList idList;

	foreach (QTreeWidgetItem *item, items) {
		idList << ((WLItem*)item)->getPath();
	}
	
	QString data=idList.join("//");

	mimeData->setData("application/wlitems", data.toLatin1().data());
	mimeData->setText( ((WLItem*)items.first())->getPath());
	return mimeData;
}
	
//-------------

void WList::dropEvent(QDropEvent *event)
{
	WLItem* target = (WLItem*) itemAt(event->pos());

	const QMimeData *data = event->mimeData();

	Qt::DropAction action = event->dropAction();
	if (action == Qt::IgnoreAction)
		return;

	if(target == 0) return;
	
	// handle drag from attribute list in browsers
	if(data->hasFormat("application/xrefitems"))
	{
		QStringList attrList=QString(data->data("application/xrefitems")).split("//");
		
		foreach(QString attrString, attrList)
		{
			emit this->xrefDropDetected(attrString);
		}
		
		qApp->processEvents();
		return;
	}
	
	// handle drag from slot
	if(data->hasFormat("text/plain") && !data->hasFormat("application/wlitems") )
	{
		QString name(data->data("text/plain"));
		
		DvObject_var dobj = getObjectAndComp(name.toLatin1().data());
		
		// remove path
		int ptr = name.lastIndexOf("/");
		if(ptr >= 0) name.remove(0, ptr+1);
		
		// fix comp text
		bool comp=false;
		// metadata
		if((ptr = name.indexOf("[=>")) >= 0) 
		{
			comp = true;
			name.replace(ptr, 3, "::");
		}

		// component
		while((ptr = name.indexOf("[")) >= 0) 
		{
			comp = true;
			name.replace(ptr, 1, "_");
		}
				
		while((ptr = name.indexOf("]")) >= 0) 
		{
			name.remove(ptr, 1);
		}
		
		// record slice
		if((ptr = name.indexOf("<")) >= 0) 
		{
			comp = true;
			name.replace(ptr, 1, "{");
		}
		
		if((ptr = name.indexOf(">")) >= 0) 
		{
			name.replace(ptr, 1, "}");
		}
				
		// srip out white spaces
		int nSpace;
		while( (nSpace = name.indexOf(" ")) != -1) name.remove(nSpace, 1);
		
		if(comp) {
			QwuiPutOnWL(dobj, name);
		}
		else 
		{
			DvObject_var dobjCopy = new DvObject(dobj);
			name += "_copy";
			QwuiPutOnWL(dobjCopy, name);
		}
		return;
	}
	
	if (!data->hasFormat("application/wlitems"))
		return;

	// WList item

	// get dragged item
	QStringList itemStrList=QString(data->data("application/wlitems")).split("//");

	QList <QTreeWidgetItem*> itemList;
	foreach(QString itemStr, itemStrList){
		QTreeWidgetItem *item = itemFromPath(itemStr);
		if(item) itemList.push_back(item);
	}
						
	if(action==Qt::CopyAction || action==Qt::MoveAction ) 
	{
		this->clearSelection();
		foreach(QTreeWidgetItem *item, itemList)
		{
			// only move top of each tree dragged
			if( !(((WLItem*)item)->ancestorInList(itemList)) && item->parent()){
				// top level of selection
				if(target) {
					if(target == item) continue; // drop on self
					
					item->parent()->removeChild(item);

					if( target->isFolder()) {
						while(target->nameExists(item->text(0)) ) {
							item->setText(0, item->text(0)+"A");
						}
						target->addChild((WLItem*)item);
					}
					else target->insertAbove((WLItem*)item);
					
					expandItem(item);
					((WLItem*)item)->updateName();
				}
			}
		}
	}
	qApp->processEvents();
	
	setNoSelection();
	this->clearSelection();
	
	return;
}

//-------------
/*
bool WList::dropMimeData(QTreeWidgetItem *target, int index, const QMimeData *data, Qt::DropAction action)
{
	// Note: in qwui and qpui QT is giving us parent and index if item under mouse
	// in qcalc QT is giving us the item itself under the mouse

	if (action == Qt::IgnoreAction)
		return true;

	if(target == 0) return false;
	
	// handle drag from attribute list in browsers
	if(data->hasFormat("application/xrefitems"))
	{
		QStringList attrList=QString(data->data("application/xrefitems")).split("//");
		
		foreach(QString attrString, attrList)
		{
			emit this->xrefDropDetected(attrString);
		}
		
		qApp->processEvents();
		return true;
	}
	
	// handle drag from slot
	if(data->hasFormat("text/plain") && !data->hasFormat("application/wlitems") )
	{
		QString name(data->data("text/plain"));
		
		DvObject_var dobj = getObjectAndComp(name.toLatin1().data());
		
		// remove path
		int ptr = name.lastIndexOf("/");
		if(ptr >= 0) name.remove(0, ptr+1);
		
		// fix comp text
		bool comp=false;
		// metadata
		if((ptr = name.indexOf("[=>")) >= 0) 
		{
			comp = true;
			name.replace(ptr, 3, "::");
		}

		// component
		while((ptr = name.indexOf("[")) >= 0) 
		{
			comp = true;
			name.replace(ptr, 1, "_");
		}
				
		while((ptr = name.indexOf("]")) >= 0) 
		{
			name.remove(ptr, 1);
		}
		
		// record slice
		if((ptr = name.indexOf("<")) >= 0) 
		{
			comp = true;
			name.replace(ptr, 1, "{");
		}
		
		if((ptr = name.indexOf(">")) >= 0) 
		{
			name.replace(ptr, 1, "}");
		}
				
		// srip out white spaces
		int nSpace;
		while( (nSpace = name.indexOf(" ")) != -1) name.remove(nSpace, 1);
		
		if(comp) {
			QwuiPutOnWL(dobj, name.toLatin1().data());
		}
		else 
		{
			DvObject_var dobjCopy = new DvObject(dobj);
			name += "_copy";
			QwuiPutOnWL(dobjCopy, name.toLatin1().data());
		}
		return true;
	}
	
	if (!data->hasFormat("application/wlitems"))
		return false;

	// WList item

	// get dragged item
	QStringList itemStrList=QString(data->data("application/wlitems")).split("//");

	QList <QTreeWidgetItem*> itemList;
	foreach(QString itemStr, itemStrList){
		QTreeWidgetItem *item = itemFromPath(itemStr);
		if(item) itemList.push_back(item);
	}
						
	if(action==Qt::CopyAction || action==Qt::MoveAction ) 
	{
		this->clearSelection();
		foreach(QTreeWidgetItem *item, itemList)
		{
			// only move top of each tree dragged
			if( !(((WLItem*)item)->ancestorInList(itemList)) && item->parent()){
				// top level of selection
				if(target) {
					WLItem * dropOnItem = (WLItem *)target->child(index-1); // this seems to be a QT bug, needs -1 ???
					if(dropOnItem == item) continue; // drop on self
					
					item->parent()->removeChild(item);

					if( ((WLItem*)dropOnItem)->isFolder()) ((WLItem*)dropOnItem)->addChild((WLItem*)item);
					else ((WLItem*)target)->insertChild(target->indexOfChild(dropOnItem),  (WLItem*)item);
					
					expandItem(item);
					((WLItem*)item)->updateName();
				}
			}
		}
	}
	qApp->processEvents();
	
	setNoSelection();
	this->clearSelection();
	
	return true;
}

*/


// ---------------- 

void WList::moveUp()
{
	// find selected item
	WLItem * item = (WLItem*) this->currentItem(); 
	if(item == NULL) return;
	
	QTreeWidgetItem * itemParent = item->parent(); 
	if(itemParent == NULL) return; // top level folder (Data, Constants, Time Intervals)
	
	WLItem * above = (WLItem*)itemAbove(item);
    if(above == NULL) return; // nowhere to go
    
/*    if(above->isFolder() && (above == itemParent) ){
        // above is the folder containing this item, look above
        above = (WLItem*) itemAbove(above);
        if(above == NULL) {
            
        }
    }
*/
    if(above == NULL) return; // nowhere to go
	if(above->parent() == NULL) return; // top of top level folder
	
	bool sameFolder = (itemParent == above->parent());
					
	// folder or data
    
    
    if(above->isFolder() && (above != itemParent) ){

        // must be empty or parent to be item above and a folder
		if(itemParent == above && itemAbove(above) ){
			itemParent->removeChild(item);
			itemAbove(above)->insertChild(0, item);
		}
		else{
			itemParent->removeChild(item);
			above->insertChild(0, item);
		}
	}
	else{
		itemParent->removeChild(item);
		if(sameFolder || (above == itemParent)) above->insertAbove(item);
		else above->insertBelow(item);
	}
	
	clearSelection();
	this->setCurrentItem(item);
	item->expandChildren();

}

//-------------

void WList::moveDown()
{
	// find selected item
	WLItem * item = (WLItem*) this->currentItem(); 
	if(item == NULL) return;
	
	// close folder to ensure item below is at same level in tree
	collapseItem(item);
	qApp->processEvents();
	
	QTreeWidgetItem * itemParent = item->parent(); 
	if(itemParent == NULL) return; // top level folder (Data, Constants, Time Intervals)
	
	WLItem * below = (WLItem*) itemBelow(item);
	if(below == NULL) return; // nowhere to go
	if(below->parent() == NULL) return; // bottom of top level folder
		
	// folder or data
	if(below->isFolder() ){
		itemParent->removeChild(item);
		below->insertChild(0, item);
	}
	else{
		itemParent->removeChild(item);
		below->insertBelow(item);
	}
	
	clearSelection();
	this->setCurrentItem(item);
	item->expandChildren();
	
}

//-------------

void WList::sortWLSelection(QList <QTreeWidgetItem*> &itemList, string &sortOn, bool reverse)
{
	if(itemList.size() < 2) return;
	
	vector <WLItem*> newList;
	vector<WLItem*>::iterator it;
	
	WLItem *item = (WLItem*) itemList[0];
	newList.push_back(item);
	int row=item->childNumber();

	// find new order
	if(sortOn == string("name")){
		for(int i=1; i<itemList.size(); i++){
			item = (WLItem*) itemList[i];

			if(row > item->childNumber()) row=item->childNumber(); // find earliest entry

			bool inserted = false;
			it=newList.begin();
			while ( it<newList.end()){
				if( item->getName() < (*it)->getName() ){
					newList.insert(it, item);
					inserted = true;
					break; 
				}
				it++;
			}
			if( !inserted) newList.push_back(item);
		}
	}
	else if(sortOn == string("time")){
		for(int i=1; i<itemList.size(); i++){
			item = (WLItem*) itemList[i];

			if(row > item->childNumber()) row=item->childNumber(); // find earliest entry

			bool inserted = false;
			it=newList.begin();
			while ( it<newList.end()){
				if( item->getTime() < (*it)->getTime() ){
					newList.insert(it, item);
					inserted = true;
					break; 
				}
				it++;
			}
			if( !inserted) newList.push_back(item);
		}	
	}
	else if(sortOn == string("object")){
		for(int i=1; i<itemList.size(); i++){
			item = (WLItem*) itemList[i];

			if(row > item->childNumber()) row=item->childNumber(); // find earliest entry

			bool inserted = false;
			it=newList.begin();
			while ( it<newList.end()){
				if( item->getType() < (*it)->getType() ){
					newList.insert(it, item);
					inserted = true;
					break; 
				}
				it++;
			}
			if( !inserted) newList.push_back(item);
		}	
	}
	
	// place in new order on WL
	for(int i=0; i<newList.size(); i++){
		if(reverse) item = newList[newList.size() - i - 1];
		else		item = newList[i];
		
		WLItem *itemParent = (WLItem*)item->parent();
		if(item!=NULL && item!=_data && item!=_time && item!=_constants)
		{
			if(row!=item->childCount()-1)
			{
				itemParent->removeChild(item);
				itemParent->addChild(item); // adds to end of folder			
			}
			row++;
		}
	}
	
	return;
}

//-------------


void WList::setNoSelection()
{
	_data->select(false);
	_constants->select(false);
	_time->select(false);
    _temp->select(false);
}

//-------------

WLItem* WList::createItem(const QString& path)
{
	QStringList pathExpl=path.split("/", QString::SkipEmptyParts);
	
	if(pathExpl.isEmpty()) return NULL;
	
	QString currentPath=pathExpl.at(0);

	WLItem* currentParent=itemFromPath(currentPath);
	if(currentParent==NULL) // We don't add anything to the root
		return NULL;
	
	WLItem* currentItem=NULL;
	for(int i=1; i<pathExpl.size(); ++i)
	{
		currentPath.append("/").append(pathExpl.at(i));
		currentItem=itemFromPath(currentPath);		
		if( currentItem == NULL && i!=pathExpl.size()-1)
		{
			currentItem=new WLItem(pathExpl.at(i), currentParent, true); // true = folder			
		}
		else if(i==pathExpl.size()-1) // We always create something in the last step: if something with the same name exists, the name will get "A" appended
		{
			if(currentItem == NULL){
				// name is safe
				currentItem=new WLItem(pathExpl.at(i), currentParent, false); // false = object (may get changed by calling fn)
			}
			else{
				// name is already used
				currentItem=itemFromPath(currentPath);
				QString useName=pathExpl.at(i);
				while(currentItem) {
					currentPath.append("A");
					currentItem=itemFromPath(currentPath);
					useName.append("A");
				}		

				currentItem=new WLItem(useName, currentParent, false); // false = object (may get changed by calling fn)
			}
		}
		currentParent = currentItem;
	}
	
	if(currentItem) currentItem->setFolder(false);

	return currentItem;
}

//-------------

void WList::newFolder()
{
	WLItem* folder=NULL;
	
	QList <QTreeWidgetItem*> itemList = selectedItems();
	
	if(itemList.size() == 0) 
	{
		folder=createItem("Data/Group");
		folder->setFolder();
		expandItem(_data);
	}
	else
	{
		// Place the new folder at the same level as the first item in selection
		WLItem* firstItem = (WLItem*) itemList[0];
		
		if(firstItem==NULL || firstItem->parent() == NULL) {
			// inappropriate selection
			folder=createItem("Data/Group");
			folder->setFolder();
			expandItem(_data);
		}
		else if( firstItem->isFolder() ){
			QString folderName = firstItem->getPath()+"/Group";
			folder=createItem(folderName);
			folder->setFolder();
			expandItem(folder->parent());
		}
		else{
			QString folderName = ((WLItem*)firstItem->parent())->getPath()+"/Group";
			folder=createItem(folderName);
			folder->setFolder();
		
			for(int i=0; i<itemList.size(); i++)
			{
				WLItem* item = (WLItem*)itemList[i];
				if(item!=NULL && item!=_data && item!=_time && item!=_constants)
				{
					item->parent()->removeChild(item);
					folder->addChild(item);
				}
				item->updateName();
			}
		}
	}

	expandItem(folder);
    folder->setIcon(0, qApp->style()->standardIcon(QStyle::SP_DirOpenIcon));

	
    if(folder) {
        editItem(folder);
        folder->updateName();
    }
	return;
}

//-------------

Qt::DropActions WList::supportedDropActions() const
{
	return Qt::CopyAction|Qt::MoveAction;
}

//-------------

void WList::rename()
{
	WLItem * item = (WLItem*) this->currentItem(); 
	if(item == NULL) return;
	
	if( notEditable(item) ) return; // Just to avoid seeing the "editing failed" message if not editable

	editItem(item, 0);

}


//-------------

void WList::sortWL()
{
    this->setNoSelection();
	_data->setChildrenSelected();
	QList<QTreeWidgetItem*> itemList = selectedItems();
	sortWLSelection(itemList, sortOn, sortReverse);
	
}
//-------------

void WList::sortWLSelected()
{
	QList<QTreeWidgetItem*> itemList = selectedItems();
	sortWLSelection(itemList, sortOn, sortReverse);
	
}

//-------------

void WList::setSortWLname(bool on)
{
	sortOn = "name";
}
//-------------

void WList::setSortWLtime(bool on)
{
	sortOn = "time";
}

//-------------

void WList::setSortWLtype(bool on)
{
	sortOn = "object";
}
//-------------

void WList::setSortRev(bool on)
{
	sortReverse = on;
}

//-------------

void WList::displayMenu(const QPoint& p)
{

	WLItem* currItem = (WLItem*)currentItem();
	if(currItem==NULL) return;
	
	RButtonMenu->clear();
	
	if(currItem==_data || currItem==_time || currItem==_constants ) 
	{
		RButtonMenu->addAction("New Object", this, SLOT(newObject()));
		RButtonMenu->exec(p);
		return;
	}
	
	QList<QTreeWidgetItem*> itemList = selectedItems();
	bool oneObject = (itemList.size()==1);
	
	RButtonMenu->addAction("Move items to new folder", this, SLOT(newFolder()));
	if(oneObject){
		RButtonMenu->addAction("Re-Name", this, SLOT(rename()));
		RButtonMenu->addAction("Edit Object", this, SLOT(editObject()));
	}
	RButtonMenu->addAction("New Object", this, SLOT(newObject()));
	RButtonMenu->addSeparator();
	RButtonMenu->addAction("Duplicate", this, SLOT(duplicateObj()));
	if(oneObject)
	{
		RButtonMenu->addAction("Subsample/Extract", this, SLOT(sliceObj()));
	}
	if(oneObject && currItem->isFolder())
		RButtonMenu->addAction("Time Interval Folder to Event Table", this, SLOT(tif2ev()));
	if(oneObject /*and event table*/)
	{
		DvObject_var obj=getWLObject(currItem->getPath().toLatin1().data());
		if(obj.is_ok() && obj->is_event())
		{
			RButtonMenu->addAction("Put Events in Time Interval Folder", this, SLOT(ev2tif()));
			RButtonMenu->addAction("Export Event Table", this, SIGNAL(exportET()));
		}
	}
	RButtonMenu->addSeparator();
	RButtonMenu->addAction("Copy", this, SLOT(copy()));
	RButtonMenu->addAction("Move Up", this, SLOT(moveUp()), Qt::CTRL+Qt::Key_Up);
	RButtonMenu->addAction("Move Down", this, SLOT(moveDown()), Qt::CTRL+Qt::Key_Down);
	RButtonMenu->addSeparator();
	RButtonMenu->addAction(style()->standardIcon(QStyle::SP_TrashIcon, 0, this), "Delete", this, SLOT(deleteObject()));
	//if at least one object has time tags

	for(int i=0; i<itemList.size(); i++)
	{
		WLItem* item = (WLItem*)itemList[i];
		if(item!=NULL)
		{
			DvObject_var dobj = getWLObject(item->getPath().toLatin1().data());
			if(dobj.is_ok() && dobj->hasTimeTags()) {
				RButtonMenu->addAction("Remove Time Tags", this, SLOT(removeTimeTags()));
				break;
			}
		}
	}
	
	
	RButtonMenu->exec(p);
}

//------

void WList::mousePressEvent(QMouseEvent* event)
{
	if(event->button()==Qt::RightButton)
	{
		displayMenu(event->globalPos());
	}
	else QTreeWidget::mousePressEvent(event);

}

//------


void WList::editObject()
{
	WLItem* item = (WLItem*)currentItem();
	if(item==NULL) return;

	DataBrowseWindow * browser;
	
	if(item->isFolder()) return;
	
	DvObject_var obj = getWLObject(item->getPath().toLatin1().data());
	DvObject_var nilObj = new DvObject();
	
	browser = QbuLaunchBrowse(nilObj, this, item->getPath().toLatin1().data(), obj, item->getPath());
	if(browser) connect(browser, SIGNAL(objChanged(const QString&)), this, SIGNAL(objChanged(const QString&)));
}

//------

void WList::newObject()
{

	if(QuSlotDebug())
	{
		printf("WList::createObject \n");
		fflush(stdout);
	}
	
	static QPushButton *Sbtn;
	static QPushButton *Vbtn;
	static QPushButton *Abtn;
	static QPushButton *tbtn;
	static QPushButton *tibtn;
	static QPushButton *Txtbtn;
	static QPushButton *TxtArrbtn;
	static QWidget * typeBox;
	
	if(objTypeSelector==NULL)
	{
		objTypeSelector = new QDialog(0);
		typeBox = new QWidget(objTypeSelector);
		QVBoxLayout* typeBoxLayout=new QVBoxLayout;
		typeBoxLayout->setMargin(4);
		typeBoxLayout->setSpacing(6);
	
		Txtbtn = new QPushButton(" Text Constant  ", typeBox); typeBoxLayout->addWidget(Txtbtn);
		TxtArrbtn = new QPushButton(" Text Array   ", typeBox); typeBoxLayout->addWidget(TxtArrbtn);
		tbtn = new QPushButton(" Time value ", typeBox); typeBoxLayout->addWidget(tbtn);
		tibtn = new QPushButton(" Time Interval  ", typeBox); typeBoxLayout->addWidget(tibtn);
		Sbtn = new QPushButton(" Scalar Constant  ", typeBox); typeBoxLayout->addWidget(Sbtn);
		Vbtn = new QPushButton(" Vector Constant  ", typeBox); typeBoxLayout->addWidget(Vbtn);
		Abtn = new QPushButton(" Array Constant  ", typeBox); typeBoxLayout->addWidget(Abtn);
		connect(Txtbtn,SIGNAL(clicked()), this, SLOT(createText()));
		connect(TxtArrbtn,SIGNAL(clicked()), this, SLOT(createTextArr()));
		connect(tbtn,SIGNAL(clicked()), this, SLOT(createTime()));
		connect(tibtn,SIGNAL(clicked()), this, SLOT(createTInterval()));
		connect(Sbtn,SIGNAL(clicked()), this, SLOT(createConstant()));
		connect(Vbtn,SIGNAL(clicked()), this, SLOT(createVConstant()));
		connect(Abtn,SIGNAL(clicked()), this, SLOT(createArray()));
		typeBox->setLayout(typeBoxLayout);
		typeBox->adjustSize();
	}
	objTypeSelector->show();
}

//------

void WList::sortWLPrefs()
{
	if(sortTypeSelector==NULL)
	{
		
		sortTypeSelector = new QDialog(0);
		
		QWidget *typeBox = new QWidget(sortTypeSelector);
		QVBoxLayout* typeBoxLayout=new QVBoxLayout(typeBox);
		typeBoxLayout->setMargin(10);
		typeBoxLayout->setSpacing(6);
		
		QLabel *title = new QLabel("  Working List Sort Preferences  ");
		typeBoxLayout->addWidget(title);

		QLabel *titleSpace = new QLabel("  ");
		typeBoxLayout->addWidget(titleSpace);

		QCheckBox *revTickBox = new QCheckBox(" Sort in Reverse order", typeBox);
		typeBoxLayout->addWidget(revTickBox);
		revTickBox->setChecked(false);
		
		QLabel *spacer = new QLabel("  ");
		typeBoxLayout->addWidget(spacer);

		QRadioButton *nameBtn = new QRadioButton(" Sort on Name  ", typeBox); 
		typeBoxLayout->addWidget(nameBtn);
		nameBtn->setChecked(true);
		
		QRadioButton *timeBtn = new QRadioButton(" Sort on Start Time ", typeBox); 
		typeBoxLayout->addWidget(timeBtn);
		
		QRadioButton *typeBtn = new QRadioButton(" Sort on Object Type ", typeBox); 
		typeBoxLayout->addWidget(typeBtn);

		QLabel *spacer2 = new QLabel("  ");
		typeBoxLayout->addWidget(spacer2);

		
		connect(nameBtn,SIGNAL(clicked(bool)), this, SLOT(setSortWLname(bool)));
		connect(timeBtn,SIGNAL(clicked(bool)), this, SLOT(setSortWLtime(bool)));
		connect(typeBtn,SIGNAL(clicked(bool)), this, SLOT(setSortWLtype(bool)));
		connect(revTickBox,SIGNAL(clicked(bool)), this, SLOT(setSortRev(bool)));
		
		typeBox->setLayout(typeBoxLayout);
		typeBox->adjustSize();
	}
	sortTypeSelector->show();
}

//------

void WList::createText()
{
	if(QuSlotDebug())
	{
		printf("WList::createText \n");
		fflush(stdout);
	}
	WLItem* currItem = (WLItem*)currentItem();
	
	QString folderName;
	if(currItem!=NULL)
	{
		folderName=(currItem->isFolder() ? currItem->getPath(): ((WLItem*)currItem->parent())->getPath());
	}

	DvObject_var nil_obj;
	DataBrowseWindow *QBROWSER_widget;
	DvString txt("text");
	DvObject_var dobj = new DvObject(txt);

	objTypeSelector->close();
	QString path =  folderName+QString("/newText");

	QBROWSER_widget = QbuLaunchBrowse(nil_obj, this, "", dobj, path);
	QBROWSER_widget->name_field->setText("newText"); // set here so that it doesn't think object exists on WL

}

//------

void WList::createTextArr()
{
	if(QuSlotDebug())
	{
		printf("WList::createTextArr \n");
		fflush(stdout);
	}
	WLItem* currItem = (WLItem*)currentItem();
	QString folderName;
	if(currItem!=NULL)
	{
		folderName=(currItem->isFolder() ? currItem->getPath(): ((WLItem*)currItem->parent())->getPath());
	}
 
	objTypeSelector->close();
    
	arrTypeSelector = new arrDialog(this, "arrTypeSelector", 0, true);
	
	arrTypeSelector->arrType = QB_STRING;
	arrTypeSelector->folderPath = folderName;
	arrTypeSelector->show();
}

//------

void WList::createTime()
{
	if(QuSlotDebug())
	{
		printf("WList::createTime \n");
		fflush(stdout);
	}
	
	WLItem* currItem = (WLItem*)currentItem();
	QString folderName;
	if(currItem!=NULL)
	{
		folderName=(currItem->isFolder() ? currItem->getPath(): ((WLItem*)currItem->parent())->getPath());
	}
 
	DvTime t("2001-09-05 00:00:00.000");
	DvObject_var nil_obj;
	DataBrowseWindow *QBROWSER_widget;
  
	DvObject_var dobj = new DvObject(t);
	
	objTypeSelector->close();
  
	DvString SI_ptr("1>s");
	dobj->change_xref(SI_CONVERSION, SI_ptr);
	QString path =  folderName+QString("/new");
	QBROWSER_widget = QbuLaunchBrowse(nil_obj, this, "", dobj, path);
	QBROWSER_widget->name_field->setText("new"); // set here so that it doesn't think object exists on WL
}

//------

void WList::createTInterval()
{
	if(QuSlotDebug())
	{
		printf("WList::createTInterval \n");
		fflush(stdout);
	}
	
	WLItem* currItem = (WLItem*)currentItem();
	QString folderName;
	if(currItem!=NULL)
	{
		folderName=(currItem->isFolder() ? currItem->getPath(): ((WLItem*)currItem->parent())->getPath());
	}

	if(objTypeSelector) objTypeSelector->close();

	DvEvent e(DvTime(2001, 9, 4), DvTime(2001, 9, 5));
	DvObject_var dobj = new DvObject(e);

	DvObject_var nil_obj;
	DataBrowseWindow *QBROWSER_widget;
	QString path =  folderName+QString("/new");
	QBROWSER_widget = QbuLaunchBrowse(nil_obj, this, "", dobj, path);
	QBROWSER_widget->name_field->setText("new"); // set here so that it doesn't think object exists on WL
}


//------

void WList::createConstant()
{
	if(QuSlotDebug())
	{
		printf("WList::createConstant \n");
		fflush(stdout);
	}
	
	WLItem* currItem = (WLItem*)currentItem();
	QString folderName;
	if(currItem!=NULL)
	{
		folderName=(currItem->isFolder() ? currItem->getPath(): ((WLItem*)currItem->parent())->getPath());
	}

	DvObject_var dobj = new DvObject(0.0);
	DvObject_var nil_obj;
	DataBrowseWindow *QBROWSER_widget;

	objTypeSelector->close();

	DvString SI_ptr("1>(?)");
	dobj->change_xref(SI_CONVERSION, SI_ptr);
	QString path =  folderName+QString("/newScalar");
	QBROWSER_widget = QbuLaunchBrowse(nil_obj, this, "", dobj, path);
	QBROWSER_widget->name_field->setText("newScalar"); // set here so that it doesn't think object exists on WL
}

//------

void WList::createVConstant()
{
	if(QuSlotDebug())
	{
		printf("WList::createVConstant \n");
		fflush(stdout);
	}
	
	WLItem* currItem = (WLItem*)currentItem();
	QString folderName;
	if(currItem!=NULL)
	{
		folderName=(currItem->isFolder() ? currItem->getPath(): ((WLItem*)currItem->parent())->getPath());
	}
 
	DvObject_var nil_obj;
	DataBrowseWindow *QBROWSER_widget;
  
	vector <size_t> mDim(1);
	mDim[0] = 3;
	DvObject_var dobj = new DvObject(0.0, mDim);
  
	DvString Frame_ptr("vector>gse_xyz");
	dobj->change_xref(FRAME, Frame_ptr);
  
	DvString SI_ptr("1>(?)");
	dobj->change_xref(SI_CONVERSION, SI_ptr);
	QString path =  folderName+QString("/newVector");  
	QBROWSER_widget = QbuLaunchBrowse(nil_obj, this, "", dobj, path);
	QBROWSER_widget->name_field->setText("newVector"); // set here so that it doesn't think object exists on WL
	objTypeSelector->close();
}

//------

void WList::createArray()
{
	if(QuSlotDebug())
	{
		printf("WList::createArray \n");
		fflush(stdout);
	}
	
	WLItem* currItem = (WLItem*)currentItem();
	QString folderName;
	if(currItem!=NULL)
	{
		folderName=(currItem->isFolder() ? currItem->getPath(): ((WLItem*)currItem->parent())->getPath());
	}
  
	objTypeSelector->close();
  
	arrTypeSelector = new arrDialog(this, "arrayDimChooser", 0, true);
	
	arrTypeSelector->arrType = QB_DOUBLE;
	arrTypeSelector->folderPath = folderName;
	arrTypeSelector->move(qsasGUISize().width(),30);
	arrTypeSelector->show();
}

//------

void WList::sliceObj()
{
	if(QuSlotDebug())
	{
		printf("WList::launchSlice \n");
		fflush(stdout);
	}
	
	WLItem* item = (WLItem*)currentItem();

	QString path;
	if(item==NULL || item->isFolder()) path = "";
	else  path = item->getPath();
	
	// ask user for selection
	DataSliceWindow *SliceWindow = new DataSliceWindow(0);
	SliceWindow->setObject(path.toLatin1().data());
	SliceWindow->move(qsasGUISize().width(),30);
	SliceWindow->show();
}



//------

void WList::duplicateObj()
{
	WLItem* currItem = (WLItem*)currentItem();
	if(notEditable(currItem)) return;
	
	QString path = currItem->getPath();
	DvObject_var dobj=getWLObject(path.toLatin1().data());
	DvObject_var newObj = new DvObject(dobj);
	path = path + "_copy";

	QwuiPutOnWL(newObj, path.toLatin1().data());

}

//------

void WList::tif2ev()
{
	if(QuSlotDebug())
	{
		printf("WList::launchSlice \n");
		fflush(stdout);
	}
	
	WLItem* item = (WLItem*)currentItem();
	if(notEditable(item)) return;
	
	if(!item->isFolder()) return;
	
	DvObject_var itv;
		
	for(int i=0; i<item->childCount(); ++i)
	{
		WLItem* currItem = (WLItem*)item->child(i);
		if(!currItem->isFolder())
		{
			DvObject_var temp=getWLObject(currItem->getPath().toLatin1().data());
				
			if(temp.is_ok() && temp->is_event())
			{
				if(itv.is_nil()) itv = new DvObject(temp);
				else itv += temp;
			}
		}
	}
		
	QString name=QString(item->getName())+QString(" (evt table)");
		
	QwuiPutOnWL(itv, "Time Intervals/"+name);
}

//------

void WList::ev2tif()
{
	if(QuSlotDebug())
	{
		printf("WList::launchSlice \n");
		fflush(stdout);
	}
	
	WLItem* item = (WLItem*)currentItem();
	if(notEditable(item)) return;
	if(item->isFolder()) return;

	DvObject_var obj=getWLObject(item->getPath().toLatin1().data());

	if(obj.is_ok() && obj->is_event() )
	{
		QString folderPath=item->getPath();
		folderPath.append("_folder");
		WLItem* folder = createItem(folderPath);
		folder->setFolder();
		
		for(int i=obj->seqSize()-1; i>= 0; i-- )
		{
			DvObject_var temp = new DvObject(obj->event(i));
			QString name=folderPath+QString("/evt")+QString().setNum(i);
			QString context=QwuiPutOnWL(temp, name);
		}
	}
}

//------

void WList::deleteObject()
{
	QList<QTreeWidgetItem*> itemList = selectedItems();
	clearSelection();

	// select current to something never deleted
	setCurrentItem(_data );

	foreach(QTreeWidgetItem *item, itemList)
	{
		// only delete top of each tree dragged
		if( !(((WLItem*)item)->ancestorInList(itemList)) && item->parent()){
			QString path = ((WLItem*)item)->getPath();
			deleteItem((WLItem*)item);
			emitDeleted(path);

		}
	}
	
}

//------

void WList::removeTimeTags()
{
	QList<QTreeWidgetItem*> itemList = selectedItems();
	
	foreach(QTreeWidgetItem *item, itemList)
	{
		if(item!=NULL)
		{
			DvObject_var dobj = getWLObject(((WLItem*)item)->getPath().toLatin1().data());
			
			if (dobj.is_ok() && dobj->xref_exists(DEPEND_0)) dobj->delete_xref(DEPEND_0);  
		}
		QuMsgBar("Time Tags removed: ", ((WLItem*)item)->getPath().toLatin1().data(), false);

	}
}

//------

void WList::selectAllData()
{
	_data->setChildrenSelected();
}
		

//------

void WList::dragLeaveEvent(QDragLeaveEvent *event)
{

	if(QuSlotDebug()) printf("WList::dragLeaveEvent \n");
	QTreeView::dragLeaveEvent(event);
	this->setState(QAbstractItemView::NoState);

}

//------


void WList::copy()
{

	QMimeData* mimeData = new QMimeData();
	mimeData->setData("application/wlitems", ((WLItem*)currentItem())->getPath().toLatin1().data());
	mimeData->setText( ((WLItem*)currentItem())->getPath() );

	QApplication::clipboard()->setMimeData(mimeData);
}

void WList::collapseConst(){
	collapse(indexFromItem(_constants, 0));
}




