/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/

#include "qie_caa.h"
#include "cdf.h"

DvObject_var empty;
DvNode D0_node("", empty);
int D0_rec_start;
int D0_rec_end;


// Until Chury updated
#if !defined(CDF_TIME_TT2000)
#define CDF_TIME_TT2000		33L	
#define CDF_INT8		8L
#endif

#if !defined(TT2000_CDF_MAYNEEDUPDATE)
#define TT2000_CDF_MAYNEEDUPDATE        ((CDFstatus) 1010)
#endif

//#ifdef WINDOWS
//	typedef long long DvInt8;
//#else
//	typedef long int DvInt8;
//#endif

/**********************************************************************
************************ START OF CDF DATA IMPORT FUNCTIONS ***********
***********************************************************************/

/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
*
* MODULE:  QIE
*
* LANGUAGE: ANSI C
*
* FUNCTION: QieGetCSDSgenCDF
*
* PURPOSE: Get data for whole CDF file. Metadata are retrieved separately.
*
* DESCRIPTION: Opens named file and reads each variable in turn.
*              r-variables are coerced into z-variables on the fly.
*
* GLOBALS:
*
* RETURN:
*    TYPE:   int
*    VALUES: QMW_OK         Executed successfully.
*            CDF_OPEN_ERR   Unable to open named file.
*            CDF_N_VAR_ERR  Unable to get number of variables.
*
* FUNCTIONS_USED:
*
*
* ALGORITHM:
*
*
***********************************************************************/

int QieData::queryCDFfile(DvString fullName){
   
	CDFstatus status;
	CDFid id;
	long n_vars;
 
    D0_rec_start = 0;
    D0_rec_end = 0;
    D0_node.setName(" ");
    DvObject_var empty;
    D0_node.setObj(empty);
    
	inFile.name = fullName; // used only for setting Parents attribute for QBU browser

	// open named cdf file 
	status = CDFlib(OPEN_, CDF_, fullName.c_str(), &id, NULL_);
	if(status == CDF_WARN || status == TT2000_CDF_MAYNEEDUPDATE){
	   char message[CDF_STATUSTEXT_LEN+1];
		CDFerror (status, message);
		cout << "CDF warning "  << message << endl;fflush(stdout);	
	}
	else if (status != CDF_OK){
	   char message[CDF_STATUSTEXT_LEN+1];

		CDFerror (status, message);
     	QieMsg ("ERROR for   '", &(fullName[0]), "' Cannot open file\n");
		cout << "CDF error "  << message << endl;fflush(stdout);	
		
		return 0;
	}
     
   // set to z mode to convert r to z variables 
    CDFlib(SELECT_, CDF_zMODE_, zMODEon2, NULL_);

   status = CDFlib(GET_, CDF_NUMzVARS_, &n_vars,
                         zVARs_MAXREC_, &maxCDFnumRecs,
                   NULL_);
     if (status != CDF_OK) {
        QieMsg ("ERROR for '", fullName.c_str(), "' Cannot count variables or records\n");
        return 0;
     }
     
	 maxCDFnumRecs += 1;
	 
     // count vars and initialise variable objects
	    
     if( !getCDFvariables(n_vars) ){
        QieMsg ("ERROR for '", fullName.c_str(), "' Cannot get information on Variables\n");
        return 0;
     }
     
     // Get Global attributes
    
     getCDFglobals();
     
     // Read  Attribute Helper file if present
     getAttributeHelper(fullName.c_str());

     return maxCDFnumRecs;
     
}

// --------------------------------------------------------------------------- 


bool QieData::getCDFvariables(int n_vars){
	
	QieVariable *emptyVar=NULL;

	QieVars.assign(n_vars, emptyVar);
		
	for(int n=0; n<n_vars; n++){
		QieVars[n] = new QieVariable();
		getCDFVar(n); 
	}
   
	return true;
   
} 
// --------------------------------------------------------------------------- 

void QieData::getCDFVar(int nvar){
  
  	QieVariable *nextVar = QieVars[nvar];

	char name[CDF_VAR_NAME_LEN+1];
	long dim_sizes[CDF_MAX_DIMS];
	long dim_varys[CDF_MAX_DIMS];
	long data_type;
	long recVary;
	long nDims;
	long max_rec_num;
  

	// select the variable and its V attributes
	if ( CDFlib(SELECT_, zVAR_, nvar,
                       zENTRY_, nvar,
                       NULL_)!= CDF_OK ) {
		cout << "Variable number " << nvar << " not in cdf file" << endl;
		return;
	}
  

	if ( CDFlib(GET_, zVAR_NAME_, &name[0], NULL_) != CDF_OK) {
     cout << "Failed to get name of variable number " << nvar << endl;
		return;
	}
	nextVar->name = &name[0];

 
	if( CDFlib(GET_, zVAR_NUMDIMS_, &nDims, NULL_) != CDF_OK) {
     cout << "Failed to get number of dimensions of variable '" << nextVar->name.c_str() << "'" << endl;
		return;
	}

	if (nDims > 0 ){

		if( CDFlib(GET_, zVAR_DIMSIZES_, &dim_sizes[0], 
                     zVAR_DIMVARYS_, &dim_varys[0],
                     NULL_) != CDF_OK) {
			return;
		}
		nextVar->objDim.clear();
		for(int k=0; k<nDims; k++){
			if(dim_varys[k] == NOVARY) continue;
			int dimSize = dim_sizes[k];
			nextVar->objDim.push_back(dimSize);
		}

	}
	if(nextVar->objDim.size() == 0)nextVar->objDim.push_back(1);

	if( CDFlib(GET_, zVAR_MAXREC_, &max_rec_num, NULL_) != CDF_OK) {
     cout << "Failed to get number of records for variable '" << nextVar->name.c_str() << "'" << endl;
		return;
	}
	if (max_rec_num < 0)   nextVar->nRecs = 0;  // Virtual variable, no records written
	else                   nextVar->nRecs = max_rec_num+1;
	
	if( CDFlib(GET_, zVAR_RECVARY_, &recVary, NULL_)!= CDF_OK) {
    cout << " Failed to find record variance for variable '" << nextVar->name.c_str() << "'" << endl;
		return;
	}
	if (recVary ==  VARY) nextVar->recVary = true;
	else nextVar->recVary = false;

	if (CDFlib(GET_, zVAR_DATATYPE_, &data_type, NULL_) != CDF_OK){
     cout << "Failed to get data type of variable '" << nextVar->name.c_str() << "'" << endl;
		return;
	}
	if( data_type == CDF_EPOCH || data_type == CDF_EPOCH16|| data_type == CDF_TIME_TT2000){
		if(d0_var_num == -1) {
            // keep as default for when there is no DEPEND_0
            d0_var_num = nvar;
            d0_var_len = max_rec_num+1;
        }
		nextVar->datatype = DATA_TIME;
	}
  
      
  // add V attributes now that variable exists
	addVarAttrs(nvar); // nvar is cdf var num which may differ from index of QieVar vector

	return;
}
// ----------------------------------------------------------------------------------------

void QieData::addVarAttrs(int nvar){


long i;
long max_num_attrs;
long data_type;
long num_elems;
long attr_scope;
char attribute_name[ CDF_ATTR_NAME_LEN +1 ];

  	QieVariable *V = QieVars[nvar];

  // VARIABLE and zENTRY_ already selected
  
  if(CDFlib( GET_, CDF_NUMATTRS_, &max_num_attrs, NULL_ ) != CDF_OK ){
    cout << "Failed to get number of attributes for variable '" << V->name.c_str() << "'" << endl;
    return;
  }


  // count entries as we go, zENTRY_ already selected by calling routine 

  // read attribute entries: if we fail, try next in list 

  for( i=0 ; i < max_num_attrs ; i++ )
  {
    
     if( CDFlib( SELECT_, ATTR_, i, NULL_) != CDF_OK ) continue;

     if( CDFlib( GET_, ATTR_SCOPE_,  &attr_scope, NULL_) != CDF_OK ) continue;
    
     if(attr_scope != VARIABLE_SCOPE) continue; // skip globals here

     if( CDFlib( CONFIRM_, zENTRY_EXISTENCE_, (long) nvar, NULL_) != CDF_OK ) continue;
 
     if( CDFlib( GET_, ATTR_NAME_, &(attribute_name[0]), NULL_ ) != CDF_OK ) continue;

     if( CDFlib( GET_, zENTRY_DATATYPE_, &data_type,
                       zENTRY_NUMELEMS_, &num_elems,
                       NULL_ ) != CDF_OK ) continue;

    // allocate entry space

     void *data = GAttrEntryMalloc( num_elems, data_type );

     if(  CDFlib( GET_,  zENTRY_DATA_, data, NULL_ ) != CDF_OK ) continue;

     // Note NULL terminator for char is already handled by  GAttrEntryMalloc for char types

     // convert to DVOS object
     DvObject_var attrObj = CDFattrBufferToObj(data, data_type, num_elems);
	 free(data);

     V->Vxref.append(attribute_name, attrObj); 

  } // end for loop over attributes 

    // ensure has Dep0
  DvNode *Xnode = V->Vxref.find(DEPEND_0);
  if( !Xnode && V->datatype != DATA_TIME && V->recVary && d0_var_num >= 0 && d0_var_num != nvar && V->Vobj->seqSize() == d0_var_len)
  {
	 DvString dep0 = QieVars[d0_var_num]->name;
	 DvObject_var xref = new DvObject(dep0);
     V->Vxref.append(DEPEND_0, xref); 
  }
  
  
  return;

}

// -------------------------------------------------------------

void *GAttrEntryMalloc( long num_elems, long data_type )
{
void *p;
size_t size;

if( num_elems < 1 ) num_elems = 1;   /* some defensive programming ... */

switch (data_type) {
  case CDF_BYTE: case CDF_INT1: case CDF_UINT1:
  {
    size = num_elems;
    break;
  }
  case CDF_INT2: case CDF_UINT2:
  {
    size = 2 * num_elems;
    break;
  }
  case CDF_INT4: case CDF_UINT4: case CDF_REAL4: case CDF_FLOAT:
  {
    size = 4 * num_elems;
    break;
  }

  case CDF_REAL8:  case CDF_EPOCH: case CDF_DOUBLE: case CDF_TIME_TT2000: case CDF_INT8:
  {
    size = 8 * num_elems;
    break;
  }

  case CDF_EPOCH16: 
  {
    size = 16 * num_elems;
    break;
  }

  case CDF_CHAR: case CDF_UCHAR:
  {
    size = num_elems + 1;
    break;
  }
  default: size = 8 * num_elems;
}

p = (void *)malloc( size );

if( data_type == CDF_CHAR || data_type == CDF_UCHAR )
  ((char *)p)[num_elems] = '\0' ;

return p;

}   

// --------------------------------------------------------------------

void QieData::getCDFglobals(){

  long num_attrs;
  long attr_scope;

  /* loop through all attributes and get G attrs */

  if( CDFlib( GET_, CDF_NUMATTRS_, &num_attrs, NULL_) != CDF_OK ){
    cout << "Failed to get number of attributes " << endl;
    return;
  }
  
  for(int i=0 ; i < (int) num_attrs ; i++ )
  {
    if( CDFlib( SELECT_, ATTR_, i, NULL_) != CDF_OK ) continue;

    if( CDFlib( GET_, ATTR_SCOPE_,  &attr_scope, NULL_) != CDF_OK ) continue;

    if( attr_scope != GLOBAL_SCOPE ) continue;  // skip V attrs

    getGlobalAttr();    

  } // end loop over num_attrs



  return;

}

// --------------------------------------------------

void QieData::getGlobalAttr()
{  
	char attribute_name[ CDF_ATTR_NAME_LEN +1 ];
	long num_entries;
	long data_type;
	long num_elems;
  
	// selection of next G attr by number already done
  
	if(  CDFlib( GET_,
             ATTR_NAME_, &(attribute_name[0]),
             ATTR_MAXgENTRY_, &num_entries,
             NULL_ )!= CDF_OK ){
		cout << "Failed to get name for Global attribute " << endl;
		return;
	}
  
	// actually what comes back is the maximum gEntry number, but we want
	// to use the number of gEntries _including_ any gEntries which
	// have no data (ie ones which "don't exist")
	// This is why we do not use ATTR_NUMgENTRIES_
	// therefore increment ..

	num_entries++;
	if(num_entries <= 0) return;
  
	long data_type_0 = 0;
	long num_elems_0 = 0;
	if( CDFlib( CONFIRM_, gENTRY_EXISTENCE_, 0, NULL_) != CDF_OK ) 
	{
		cout << "Failed to confirm first entry for Global attribute " << endl;
		return;
    }
	
    if( CDFlib( SELECT_, gENTRY_, 0,
                GET_, gENTRY_DATATYPE_, &data_type_0,
                      gENTRY_NUMELEMS_, &num_elems_0,
                      NULL_ ) != CDF_OK ){
		cout << "Failed to get data type/number of elements for Global attribute " << endl;
		return;
    }

	vector <size_t> dim(1);
	if(data_type_0 == CDF_CHAR || data_type_0 == CDF_UCHAR)	dim[0] = 1; // CDF does not have str array attributes
	else dim[0] = (size_t) num_elems_0;
	
	DvString gattr_datatype;
	DvObject_var Gobj = createObj(data_type_0, num_entries, dim, gattr_datatype);

	for(int n = 0; n < num_entries; n++ ){
	
		// fetch one entry at a time to catch CDF non-homogeneous G attributes
					
		if( CDFlib( CONFIRM_, gENTRY_EXISTENCE_, n, NULL_) != CDF_OK ) continue;
    
		if( CDFlib( SELECT_, gENTRY_, n,
			GET_, gENTRY_DATATYPE_, &data_type,
			gENTRY_NUMELEMS_, &num_elems,
			NULL_ ) != CDF_OK ){
			cout << "Failed to get data type/number of elements for Global attribute " << endl;
			continue;
		}

		if( data_type != data_type_0  ) 
		{
			cout << "Skipping inhomogeneous data type entry in Global attribute " << endl;
			continue; 
		}
		
		if(  num_elems != num_elems_0 && data_type != CDF_CHAR && data_type != CDF_UCHAR) 
		{
			cout << "Skipping inhomogeneous entry length in Global attribute " << endl;
			continue; 
		}
		
		void *data = GAttrEntryMalloc( num_elems, data_type );

    	if( CDFlib( GET_, gENTRY_DATA_, data, NULL_ ) != CDF_OK) {
			cout << "Failed to get data for Global attribute " << endl;
			free(data);
			return;
    	}
     
	 	int len = (int) num_elems;

    	// turn void * buffer into object
		switch( data_type )
		{
        	case CDF_BYTE:
        	case CDF_INT1:
			{
          		for(int i=0; i<len; i++)
					Gobj->itg(n*len + i) = *( static_cast<char*>(data) + i );
				break;
			}
        	case CDF_UINT1:
			{
          		for(int i=0; i<len; i++)
					Gobj->itg(n*len + i) = *( static_cast<unsigned char*>(data) + i );
          		break;
			}
        	case CDF_INT2:
			{
          		for(int i=0; i<len; i++)
					Gobj->itg(n*len + i) = *( static_cast<short*>(data) + i );
          		break;
			}
        	case CDF_UINT2:
			{
         		for(int i=0; i<len; i++)
					Gobj->itg(n*len + i) = *( static_cast<unsigned short*>(data) + i );
         		 break;
			}
        	case CDF_INT4:
			{
          		for(int i=0; i<len; i++)
					Gobj->itg(n*len + i) = *( static_cast<int*>(data) + i );
          		break;
			}
        	case CDF_INT8:
			{
          		for(int i=0; i<len; i++)
					Gobj->itg(n*len + i) = *( static_cast<long long*>(data) + i );
          		break;
			}
        	case CDF_UINT4:
			{
          		for(int i=0; i<len; i++)
					Gobj->itg(n*len + i) = *( static_cast<unsigned int*>(data) + i );
          		break;
			}
        	case CDF_REAL4:
        	case CDF_FLOAT:
			{
          		for(int i=0; i<len; i++)
					Gobj->dbl(n*len + i) = *( static_cast<float*>(data) + i );
          		break;
			}
        	case CDF_REAL8:
        	case CDF_DOUBLE:
			{
          		for(int i=0; i<len; i++)
					Gobj->dbl(n*len + i) = *( static_cast<double*>(data) + i );
          		break;
			}
        	case CDF_CHAR:
			case CDF_UCHAR:
        	{
				char *sData = (char *) data;
          		for(int i=0; i<len; i++)
                    // if avoids stray NULLs written into file by accident
    				if( *(sData + i) ) Gobj->str(n) += *(sData + i);
          		break;
			}
			case CDF_EPOCH: 
        	{
				if(num_elems == 1){
					double *dp = (double *)data;
						Gobj->time(n).setFromCDFepoch( *( static_cast<double*>(dp)) );
				}
				else if(num_elems == 2){
					double *dp = (double *)data;
					DvTime st;
					DvTime ed;
					st.setFromCDFepoch( *( static_cast<double*>(dp)));
					ed.setFromCDFepoch( *( static_cast<double*>(dp + 1)));
   
					Gobj->event(n).set( st, ed);
				}
				else{
					cout << "Global attribute time arrays not handled" << endl;
					free(data);
					return;
				}

          		break;
			}
			case CDF_EPOCH16: 
			{		
				if(num_elems == 1){
					double *dp = (double *)data;
                    Gobj->secRes = 9;
                    Gobj->time(n).setFromCDFepoch16(  static_cast<double*>(dp) );
                }
				else if(num_elems == 2){
					double *dp = (double *)data;
					DvTime st;
					DvTime ed;
					st.setFromCDFepoch16( static_cast<double*>(dp));
					ed.setFromCDFepoch16( static_cast<double*>(dp + 1 ));
   
					Gobj->secRes = 9;
					Gobj->event(n).set(st, ed);
				}
				else{
					cout << "Global attribute time arrays not handled" << endl;
					free(data);
					return;
				}         		
				break;

			}
			case CDF_TIME_TT2000: 
        	{
				if(num_elems == 1){
					DvInt8 *ip = (DvInt8 *)data;
					Gobj->secRes = 9;
					Gobj->time(n).setFromTT2000( *ip );
				}
				else if(num_elems == 2){
					DvInt8 *ip = (DvInt8 *)data;
					DvTime st;
					DvTime ed;
					st.setFromCDFepoch(*ip);
					ed.setFromCDFepoch(*(ip+1));
   
					Gobj->secRes = 9;
					Gobj->event(n).set( st, ed);
				}
				else{
					cout << "Global attribute time arrays not handled" << endl;
					free(data);
					return;
				}

          		break;
			}

		}
    	free(data);
	}

	if (Gobj.is_nil() ) return;

	DvString typeXref(gattr_datatype.c_str());
	Gobj->change_xref("value_type", typeXref);
	
    GlobalList.append(attribute_name, Gobj);
	
	return;
}

// ------------------------------------------------------------

DvObject *createObj(int data_type, size_t seqLen, vector <size_t> &dim, DvString &typeStr){

	int num_elems = 1;
	for(size_t i=0; i<dim.size(); i++) num_elems *= dim[i];
	
	switch( data_type )
	{
		case CDF_BYTE:
		case CDF_INT1:
		case CDF_UINT1:
		case CDF_INT2:
		case CDF_UINT2:
		case CDF_INT4:
		case CDF_UINT4:
		case CDF_INT8:
		{
			typeStr = "INT";
			return new DvObject((int) 0, dim, seqLen);
          		
		}
		case CDF_REAL4:
		case CDF_FLOAT:
		case CDF_REAL8:
		case CDF_DOUBLE:
		{
			typeStr = "DOUBLE";
			return new DvObject((double) 0.0, dim, seqLen);
		}
		case CDF_CHAR:
		case CDF_UCHAR:
		{
			typeStr = "CHAR";
			return new DvObject(DvString(), dim, seqLen);
		}
		case CDF_EPOCH: 
		case CDF_EPOCH16:
		case CDF_TIME_TT2000:
		case ISO_TIME:
		{
			if(num_elems == 1){
				typeStr = "ISO_TIME";
				return new DvObject(DvTime(), seqLen);
			}
			else if(num_elems == 2){
				typeStr = "ISO_TIME_RANGE";
				return new DvObject(DvEvent(), seqLen);
			}
			else{
				typeStr = "ISO_TIME";
				return new DvObject(DvTime(), dim, seqLen);
			}
		}
		case ISO_TIME_RANGE:
		{
			typeStr = "ISO_TIME_RANGE";
			return new DvObject(DvEvent(), dim, seqLen);
		}
	}
	
	return new DvObject();
}


// ------------------------------------------------------------

DvObject *CDFattrBufferToObj(void *data, int data_type, int num_elems)
{
    // For V Attrs  

	if(num_elems == 0) return new DvObject();

	vector <size_t> dim(1);
	if(data_type == CDF_CHAR || data_type == CDF_UCHAR)	dim[0] = 1; // CDF does not have str array attributes
	else dim[0] = (size_t) num_elems;
	
	DvString typeStr;
	    
	DvObject *a_obj = createObj(data_type, 1, dim, typeStr);
  
    // switch on data type
 
	switch( data_type )
	{
	    case CDF_CHAR:
		case CDF_UCHAR:
      	{
       		// cdf uses num_elems for string length but we also NULL terminate
        
       		char *p = ( char*) data;
         
       		int strLen = (int)strlen(p);
       		if(strLen > num_elems) strLen = num_elems; // for safety only
       		while ( strLen > 0 && p[strLen-1] == ' ' ) strLen--; // remove trailing spaces
       
       		a_obj->str(0) = DvString(p, strLen);
	   		break;
		}
    	case CDF_EPOCH:
		{
      
      		if( num_elems == 1){
         		// normal case of single time
         		double *dp = (double *)data;
	 
         		a_obj->time(0).setFromCDFepoch( *dp);  
      		}
      		else if( num_elems == 2) {
         		// special case if time interval

				double *dp = (double *)data;
				DvTime st;
				DvTime ed;
				st.setFromCDFepoch( *( static_cast<double*>(dp)  ));
				ed.setFromCDFepoch( *( static_cast<double*>(dp) + 1 ));
		
				a_obj->event(0).set(st, ed);
      		}
      		else {
         		// normal case of single time
				double *dp = (double *)data;
	 			for(int i=0; i<num_elems; i++) a_obj->time(i).setFromCDFepoch( *(dp + i));  
      		}
			break;
    	}
    	case CDF_EPOCH16:
		{
      
       		if( num_elems == 1){
        		 // normal case of single time
         		double *dp = (double *)data;
	 
				a_obj->secRes = 9;
				a_obj->time(0).setFromCDFepoch16(dp);  
      		}
      		else if( num_elems == 2) {
         		// special case if time interval

				double *dp = (double *)data;
				DvTime st;
				DvTime ed;
				st.setFromCDFepoch16( static_cast<double*>(dp) );
				ed.setFromCDFepoch16( static_cast<double*>(dp) + 2 );
		
				a_obj->secRes = 9;
				a_obj->event(0).set(st, ed); 
      		}
      		else {
        	 	// normal case of single time
				double *dp = (double *)data;
	 			for(int i=0; i<num_elems; i++) a_obj->time(i).setFromCDFepoch16( dp + 2*i);  
				 
				a_obj->secRes = 9;
      		}
			break;
    	}
		case CDF_TIME_TT2000:
		{
      
      		if( num_elems == 1){
         		// normal case of single time
         		DvInt8 *ip = (DvInt8 *)data;
	 
				a_obj->secRes = 9;
         		a_obj->time(0).setFromTT2000( *ip);   // convert to sec
      		}
      		else if( num_elems == 2) {
         		// special case if time interval

				DvInt8 *ip = (DvInt8 *)data;
				DvTime st;
				DvTime ed;
				st.setFromTT2000(*ip);
				ed.setFromTT2000(*(ip+1));
		
				a_obj->secRes = 9;
				a_obj->event(0).set(st, ed);
      		}
      		else {
         		// normal case of single time
				a_obj->secRes = 9;
				DvInt8 *ip = (DvInt8 *)data;
	 			for(int i=0; i<num_elems; i++) a_obj->time(i).setFromTT2000(*(ip + i));
      		}
			break;
    	}

		case CDF_BYTE:
		case CDF_INT1:
		{
			for(int i=0; i<num_elems; i++)
				a_obj->itg(i) = *( static_cast<char*>(data) + i );
			break;
		}
		case CDF_UINT1:
			for(int i=0; i<num_elems; i++)
				a_obj->itg(i) = *( static_cast<unsigned char*>(data) + i );
			break;
		case CDF_INT2:
			for(int i=0; i<num_elems; i++)
				a_obj->itg(i) = *( static_cast<short*>(data) + i );
			break;
		case CDF_UINT2:
			for(int i=0; i<num_elems; i++)
				a_obj->itg(i) = *( static_cast<unsigned short*>(data) + i );
			break;
		case CDF_INT4:
			for(int i=0; i<num_elems; i++)
				a_obj->itg(i) = *( static_cast<int*>(data) + i );
			break;
		case CDF_INT8:
			for(int i=0; i<num_elems; i++)
                a_obj->itg(i) = *( static_cast<long long*>(data) + i );
			break;
		case CDF_UINT4:
			for(int i=0; i<num_elems; i++)
				a_obj->itg(i) = *( static_cast<unsigned int*>(data) + i );
			break;
		case CDF_REAL4:
		case CDF_FLOAT:
			for(int i=0; i<num_elems; i++)
				a_obj->dbl(i) = *( static_cast<float*>(data) + i );
			break;
		case CDF_REAL8:
		case CDF_DOUBLE:
		{
			for(int i=0; i<num_elems; i++)
				a_obj->dbl(i) = *( static_cast<double*>(data) + i );
			break;
		}
	}

	return a_obj;
}

// -------------------------------------------------------------------------

void QieData::getCDFvarDataList(vector <size_t> getList, int recStart, int recEnd){

    for (size_t i=0; i<getList.size(); i++){

       getCDFvarData(getList[i], recStart, recEnd);

    }

    // now all dependent vars and attrs have been read we can fix Themis epoch

    fixThemisEpoch(getList);
	
    // The variables needed have now been read into their Vobj pointers for the records requested

 }
 
// --------------------------------------------------------------------------- 

void QieData::getCDFvarData(size_t nvar, int recStart, int recEnd){

  long dim_indices[CDF_MAX_DIMS];
  long dim_counts[CDF_MAX_DIMS];
  long dim_varys[CDF_MAX_DIMS];
  long data_type;
  long num_elems=1;
  long varMaxRec=0;
  int strLength=1;
  size_t size_of_datatype =1; 
  int nRecsToGet;
  long majority;
    
  if(nvar >= QieVars.size()) return;
  
  QieVariable * V = QieVars[nvar];
    
  if(recEnd == -1) recEnd = V->nRecs-1; // cope with request for xref to virtual object
  
  // select the variable 
  if ( CDFlib(SELECT_, zVAR_, (long) nvar,
                       NULL_)!= CDF_OK ) {
	cout << "Variable number " << nvar << " for variable named '" << V->name.c_str() << "' not in CDF file" << endl;
	return;
  }

  if (CDFlib(GET_, zVAR_DATATYPE_, &data_type,
                  zVAR_DIMVARYS_, &(dim_varys[0]),
				  zVAR_MAXREC_, &varMaxRec,
                  NULL_) != CDF_OK){
     cout << "Failed to get data type or dimension varys of variable '" << V->name.c_str() << "'" << endl;
     return;
  }

	if(recEnd > varMaxRec) recEnd = varMaxRec;
	if(recStart > varMaxRec) recStart = 0;
	
    if( CDFlib(GET_, zVAR_NUMELEMS_, &num_elems, NULL_) != CDF_OK){
    cout << " Failed to find number of elements for variable '" << V->name.c_str() << "'" << endl;
    return;
  }

  // convert to CAA text used on QSAS side
  switch(data_type){
    case CDF_CHAR:
    case CDF_UCHAR:
    {
      V->datatype = DATA_STR;
      size_of_datatype = sizeof(char);
      strLength = num_elems;
      break;
    }
    case CDF_UINT1:
    case CDF_INT1:
    case CDF_BYTE:
    {
      V->datatype = DATA_STR;
      size_of_datatype = sizeof(char);
      break;
    }
    case CDF_INT2:
    case CDF_UINT2:
    {
      V->datatype = DATA_INT;
      size_of_datatype = 2;
      break;
    }
    case CDF_INT4:
    case CDF_UINT4:
    {
      V->datatype = DATA_INT;
      size_of_datatype = 4;
      break;
    }
    case CDF_INT8:
    {
      V->datatype = DATA_INT;
      size_of_datatype = 8;
      break;
    }
    case CDF_REAL4:
    case CDF_FLOAT:
    {
      V->datatype = DATA_FLT;
      size_of_datatype = 4;
      break;
    }
    case CDF_REAL8:
    case CDF_DOUBLE:
    {
      V->datatype = DATA_DBL;
      size_of_datatype = 8;
      break;
    }
    case CDF_EPOCH:
    {
      V->datatype = DATA_TIME;
      size_of_datatype = 8;
      break;
    }
    case CDF_TIME_TT2000:
    {
      V->datatype = DATA_TIME;
      size_of_datatype = 8;
      break;
    }
    case CDF_EPOCH16:
    {
      V->datatype = DATA_TIME;
      size_of_datatype = 16;
      break;
    }
    default:
    {
      cout << " Unrecognised data type for variable '" << V->name.c_str() << "'" << endl;
      return;
    }
  }
  

  // allocate memory and dimension counts
  size_t length;
  if(V->recVary) length = (size_t) num_elems * (recEnd - recStart + 1) ;
  else  length = (size_t) num_elems;
  
  for (size_t n=0; n < V->objDim.size(); n++){

    dim_indices[n] = 0;

    if(dim_varys[n] == VARY){
      length = length * V->objDim[n];
      dim_counts[n] = V->objDim[n];
    }
    else{
      dim_counts[n] = 1;
    }
  }

  if (length <= 0) return;
  
  if (V->recVary) {
      nRecsToGet = recEnd-recStart+1;
      if( CDFlib (SELECT_, zVAR_RECNUMBER_, (long)recStart, 
                       zVAR_RECCOUNT_,  (long)nRecsToGet, 
		       zVAR_DIMCOUNTS_, dim_counts,
                       zVAR_DIMINDICES_, dim_indices,
		       NULL_) != CDF_OK) {
        cout << " Cannot set start record and number of records to get" << endl;
        return;
      }
  }
  else{
      nRecsToGet = 1;
      if( CDFlib (SELECT_, zVAR_RECNUMBER_, (long) 0, 
                       zVAR_RECCOUNT_,  (long) 1, 
		       zVAR_DIMCOUNTS_, dim_counts,
                       zVAR_DIMINDICES_, dim_indices,
		       NULL_) != CDF_OK) {
        cout << " Cannot set start record and number of records to get" << endl;
        return;
      }
   
  }

    // use timeline if already known
    if(D0_node.name() == V->name && recStart == D0_rec_start && recEnd == D0_rec_end){
        // tt are cached
        V->Vobj = D0_node.obj();
    }
    else{
        void *data = (void *) calloc (length, size_of_datatype);

        // get data array in same ordering as cdf file 
        CDFstatus status = CDFlib (GET_, zVAR_HYPERDATA_, data, NULL_);
        if( status < CDF_WARN  )  {
            cout << "failed to get CDF data for variable " << V->name.c_str() << "'" << endl;
            free (data);
            return;
        }
    
        bool isD0 = convertCDFbufferToObj(nvar, nRecsToGet, data_type, strLength, data);
        if(isD0){
            D0_rec_start = recStart;
            D0_rec_end = recEnd; 
        }
        free (data);
    }
    
    CDFlib (GET_, CDF_MAJORITY_, &majority, NULL_);

  // having filled void pointer unpack into dvos object

    if(majority == COLUMN_MAJOR && V->objDim.size() > 1) {
        // if Column major and more than 1D must re-order data 
        convertColToRow(nvar);
    }

  return;
}

  // ----------------------------------------------------------------------------
 
void QieData::convertColToRow(size_t vnum){
    
    QieVariable *V = QieVars[vnum];

    if( V->Vobj.is_nil() ) return;
    
    V->Vobj = V->Vobj->correctColToRow();
}

    
  // ----------------------------------------------------------------------------

bool QieData::convertCDFbufferToObj(size_t vnum, size_t seq_size, long data_type, size_t strLength, void *data){

	if(seq_size == 0) return false;
  
	if(vnum >= QieVars.size()) return false;
  
	QieVariable *V = QieVars[vnum];
	// test for Data Object is TimeTags object

	DvString typeStr;
	
    V->Vobj = createObj(data_type, seq_size, V->objDim, typeStr);

	size_t len = 1;
	for(size_t j=0; j<V->objDim.size(); j++) len *= V->objDim[j];
	
	size_t totalLen = len * seq_size;

	if( data_type == CDF_EPOCH ) {
    
		if(V->objDim.size() == 0 || (V->objDim.size() == 1 && V->objDim[0] ==1) ) { 

			// Time sequence
			double *ep = (double *)data;
			for (size_t r=0; r<seq_size; r++){
				V->Vobj->time(r).setFromCDFepoch(ep[r]);
			}
		}
		else if(V->objDim.size() == 1 && V->objDim[0] == 2){
			double *ep = (double *)data;

			for (size_t i=0; i<seq_size; i++){
				DvTime iv_start;
				iv_start.setFromCDFepoch(ep[2*i+0]);
				DvTime iv_end;
				iv_end.setFromCDFepoch(ep[2*i+1]);
				V->Vobj->event(i).set(iv_start, iv_end );
			}
		}
		else{ 
			double *ep = (double *)data;
			for (size_t r=0; r<totalLen; r++){
				V->Vobj->time(r).setFromCDFepoch(ep[r]);
			}
		}
	}

	else if( data_type == CDF_EPOCH16 ) { 

		// epoch 16 uses seconds in field 1 and picoseconds in field 2
		if(V->objDim.size() == 0 || (V->objDim.size() == 1 && V->objDim[0] ==1) ) { 


			// Time sequence
			double *ep = (double *)data;
			V->Vobj->secRes = 9;
			for (size_t r=0; r<seq_size; r++){
				V->Vobj->time(r).setFromCDFepoch16(&ep[2*r]);
			}
		}
		else if(V->objDim.size() == 1 && V->objDim[0] == 2){
			double *ep = (double *)data;

			V->Vobj->secRes = 9;
			for (size_t r=0; r<seq_size; r++){
				DvTime iv_start;
				iv_start.setFromCDFepoch16(&ep[4*r+0]);
				DvTime iv_end;
				iv_end.setFromCDFepoch16(&ep[4*r+2]);
				V->Vobj->event(r).set(iv_start, iv_end );
			}
		}
		else{ 
			double *ep = (double *)data;
			V->Vobj->secRes = 9;
			for (size_t r=0; r<totalLen; r++){
				V->Vobj->time(r).setFromCDFepoch16(&ep[2*r]);
			}
		}
	}
    
	else if( data_type == CDF_TIME_TT2000 ) {
    
		if(V->objDim.size() == 0 || (V->objDim.size() == 1 && V->objDim[0] ==1) ) { 

			// Time sequence
			DvInt8 *ep = (DvInt8 *)data;
            V->Vobj->secRes = 9;
			for (size_t r=0; r<seq_size; r++){
				V->Vobj->time(r).setFromTT2000(ep[r]);
			}
		}
		else if(V->objDim.size() == 1 && V->objDim[0] == 2){
			DvInt8 *ep = (DvInt8 *)data;

			V->Vobj->secRes = 9;
			for (size_t i=0; i<seq_size; i++){
				DvTime iv_start;
				iv_start.setFromTT2000(ep[2*i+0]);
				DvTime iv_end;
				iv_end.setFromTT2000(ep[2*i+1]);
				V->Vobj->event(i).set(iv_start, iv_end );
			}
		}
		else{ 
			DvInt8 *ep = (DvInt8 *)data;
			V->Vobj->secRes = 9;
			for (size_t r=0; r<totalLen; r++){
				V->Vobj->time(r).setFromTT2000(ep[r]);
			}
		}
	}
	
	else if(data_type == CDF_CHAR || data_type == CDF_UCHAR) {
	
        const char* p = (const char*)data;
		// Note CDF string sequences must have same length of string and get space padded
		
		if(V->objDim.size() == 0 || (V->objDim.size() == 1 && V->objDim[0] ==1) )
		{
			// sequence of strings
			for (size_t r=0; r<seq_size; r++){	
				const char* next = p + r*strLength;
          
				// count forward till we hit NULLS to get real length
				int actualLen=1;
				for (size_t kk=1; kk<strLength; kk++) {
					if( *(next+kk) != '\0') actualLen++;
				}
				// strip trailing spaces (this way as there may be spaces inside string)
				while (*(next+actualLen-1) == ' ')  actualLen--;
	      
				V->Vobj->str(r) = DvString(next, actualLen);
			}
		}
		else
		{
			// sequence of string arrays
			
			for(size_t r=0; r<seq_size; r++ )
			{
				const char* next = p + r*strLength*len;
				for(size_t i=0; i<len; i++)
				{ 
					// count forward till we hit NULLS to get real length
					int actualLen=1;
					for (size_t kk=1; kk<strLength; kk++) {
						if( *(next+i*strLength+kk) != '\0') actualLen++;
					}
					// strip trailing spaces 
					while (*(next+i*strLength+actualLen-1) == ' ')  actualLen--;
	      
					V->Vobj->str(r*len + i) = DvString(next+i*strLength, actualLen);
				}
			}			
		}   
    
	}
    
	else{
			// for numeric types use element by element static casting
			//   and let = (double) do the conversion to double etc

			switch( data_type )
			{
				case CDF_BYTE:
				case CDF_INT1:
					for(size_t i=0; i<totalLen; i++){
						V->Vobj->itg(i) = *( (char *)data+i);
					}
					break;
				case CDF_UINT1:
					for(size_t i=0; i<totalLen; i++){
						V->Vobj->itg(i) = *((unsigned char *)data+i);
					}
					break;
				case CDF_INT2:
					for(size_t i=0; i<totalLen; i++){
						V->Vobj->itg(i) = *((short *)data+i);
					}
					break;
				case CDF_UINT2:
					for(size_t i=0; i<totalLen; i++){
						V->Vobj->itg(i) = *((unsigned short *)data+i);
					}
					break;
				case CDF_INT4:
					for(size_t i=0; i<totalLen; i++){
						V->Vobj->itg(i) = *((int *)data+i);
					}
					break;
				case CDF_INT8:
					for(size_t i=0; i<totalLen; i++){
						V->Vobj->itg(i) = *((long long *)data+i);
					}
					break;
				case CDF_UINT4:
					for(size_t i=0; i<totalLen; i++){
						V->Vobj->itg(i) = *((unsigned int *)data+i);
					}
					break;
				case CDF_REAL4:
				case CDF_FLOAT:
					for(size_t i=0; i<totalLen; i++){
						V->Vobj->dbl(i) = *((float *)data+i);
					}
					break;
				case CDF_REAL8:
				case CDF_DOUBLE:
					for(size_t i=0; i<totalLen; i++){
						V->Vobj->dbl(i) =  *((double *)data+i);
					}
					break;
			}
         
	} // end else on data type

    if(vnum == d0_var_num){
        D0_node.setName(V->name);
        D0_node.setObj(V->Vobj); 
        return true;
    }
    return false;
}



// ----------------------------------------------------------------------------

void QieData::fixThemisEpoch(vector <size_t> getList){

	DvString virt, comp0, comp1;
	DvTime baseT;
	DvObject_var offset;

	for(size_t j=0; j<getList.size(); j++){
		int i = getList[j];

		if(QieVars[i]->datatype != DATA_TIME) continue;
		
		QieVariable *epochV = QieVars[i];
		// look for Themis virtual flag 

		DvNode *Xnode = epochV->Vxref.find("VIRTUAL");
		if(Xnode) 
		{
			virt = Xnode->obj()->asStr();
		}
		Xnode = epochV->Vxref.find("COMPONENT_0");
		if(Xnode) 
		{
			comp0 = Xnode->obj()->asStr();
		}
		
		Xnode = epochV->Vxref.find("COMPONENT_1");
		if(Xnode) 
		{
			comp1 = Xnode->obj()->asStr();
		}
		if( !comp0.empty() ){
			// get variables for the two parts 
			int comp0_num = findVarNumByName( comp0.c_str());
			if(comp0_num >= 0 && QieVars[comp0_num]->Vobj.is_ok() && QieVars[comp0_num]->Vobj->is_time()) 
			{
				baseT = QieVars[comp0_num]->Vobj->time(0);
			}
		}
		
		if( !comp1.empty() ) {
			int comp1_num = findVarNumByName(comp1.c_str());
			if(comp1_num >= 0) offset = QieVars[comp1_num]->Vobj;
		}

		if( !virt.empty() && offset.is_ok() && offset->is_dbl() ){

			// determine Themis time tags */
			if (virt == "TRUE" ){
				DvObject_var real_tt = new DvObject(baseT, offset->seqSize() );
				for (size_t recN=0; recN< offset->seqSize(); recN++){
					real_tt->time(recN) = baseT + offset->dbl(recN); 
				}
	 
				epochV->Vobj = real_tt;
				epochV->Vobj->secRes = 9;

			}
		}
	}
}

/**********************************************************************
************************ START OF CDF DATA EXPORT FUNCTIONS ***********
***********************************************************************/

bool QieData::writeCDF( bool replace, bool compress)
{
	bool OK;
	
	if( !ensureFileName(replace) ) {
		cerr << "Write Error: unable to open file for write (file exists or no write permission)\n   " << outFile.name << endl;
		return false;
	} 


	// convert attribute string arrays to pointers to NRV vars 
	removeStrArrays();

	// create .cdf file 
	OK = createCDF(outFile.CDFbackward, replace);
	if( !OK ) {
		cout << "Failed to create CDF for write" << endl;
		return false;
	}

	// compression
	if(compress){
		long comp[CDF_MAX_PARMS];
		comp[0] = 6L;
		CDFlib(PUT_, CDF_COMPRESSION_, GZIP_COMPRESSION, comp, NULL_);
	}
	
	/* write structure to file */

	OK = writeContentsCDF();
    if( !OK ){
       /* delete badly formed CDF file and return warning */
       CDFlib(DELETE_, CDF_, NULL_);
       return false;
    }

	/* Close CDF and exit cleanly */

	CDFlib(CLOSE_, CDF_, NULL_);


	return true;

}



// -------------------------------------------

void  QieData::removeStrArrays(){

	/* for each variable, test if attribute is string array */

	for (size_t n=0; n < QieVars.size(); n++){

		QieVariable* V = QieVars[n];
	
		DvNode *Xnode = V->Vobj->first_xref();

		while(Xnode){
			if(Xnode->obj().is_nil()) continue;
			
			// if ptr to var get var
			if( Xnode->obj()->is_str() && Xnode->obj()->arraySize() > 1 ){
				
				// name
				DvString attrName = V->name;
				attrName += "_";
				attrName += Xnode->name();
				
				// create variable 
			 	QieVariable *Vattr = new QieVariable(attrName.c_str());
				Vattr->Vobj = Xnode->obj();
				Vattr->recVary = NOVARY;
				
				DvObject_var type = new DvObject("Support_Data");
				Vattr->Vobj->change_xref("PARAMETER_TYPE", type);
				
				QieVars.push_back(Vattr);
				
				// change xref to point to new var
				DvObject_var ptrObj = new DvObject(attrName);
				Xnode->setObj(ptrObj);
				
			}
			Xnode = Xnode->next;
		}
	}
}

// -------------------------------------------

bool QieData::writeContentsCDF()
{

	// write Global Attributes 
	writeGlobalsCDF();
	
	// truncate variable names to cdf limit CDF_VAR_NAME_LEN 
   
	for(size_t n=0; n< QieVars.size(); n++){
   	
		if(QieVars[n]->name.size() > CDF_VAR_NAME_LEN){
			cout << "Truncating name  " << QieVars[n]->name;
			DvString oldName = QieVars[n]->name;
			QieVars[n]->name = oldName.substr(0, 10);
			QieVars[n]->name += "...";
			QieVars[n]->name += oldName.substr(oldName.size() - CDF_VAR_NAME_LEN + 13, CDF_VAR_NAME_LEN-13);
			cout << " to " << QieVars[n]->name << endl;
			
			// correct attributes that point to it 
			for (size_t nn=0; nn< QieVars.size(); nn++){
				if( QieVars[nn]->Vobj.is_nil() ) continue;
				DvNode *node = QieVars[nn]->Vobj->first_xref();
				while(node){
					if( node->obj()->is_str() )
					{
						if( node->obj()->str() == oldName )
						{
							node->obj()->str() = QieVars[n]->name;
						}
					}
					node = node->next;
				}
			}
			
			// add extra attribute with full name 
			
			QieVars[n]->Vobj->change_xref("PARAMETER_FULL_NAME", oldName);
			

		} // end if on name too long 
	}

	// Write Variables and their metadata 

	for(size_t n=0; n< QieVars.size(); n++){
		if(QieVars[n]->skip) continue; // removes partial variables
		if( !writeVariableCDF(n) ) cout << "Error: Variable not written " << endl;
	}


	return true;
}

// -------------------------------------------

bool QieData::writeGlobalsCDF(){

	CDFstatus status;
	long attr_num;
	long n_entry;
	
	// loop through entries in Global attributes in structure 

	DvNode *node = GlobalList.first();
	while(node){

    	// select or create attribute 
		
		
    	status = CDFlib(SELECT_, ATTR_NAME_, node->name().c_str(), NULL_);
		if(status != CDF_OK) status = CDFlib(CREATE_, ATTR_, node->name().c_str(), GLOBAL_SCOPE, &attr_num, NULL_);
		
		if(status != CDF_OK || node->obj().is_nil() ){
			cout << "Failed to create Global attribute '" << node->name() << "'" << endl;
			node = node->next;
			continue;    
      	}

		// write each entry in turn 

		if(node->obj()->is_dbl()) {

			for (n_entry=0; n_entry < (long) node->obj()->seqSize(); n_entry++){

      			status = CDFlib(SELECT_, gENTRY_, n_entry, NULL_);
        		if(status != CDF_OK){
					cout << "Failed to select Global entry " << n_entry << endl;
					node = node->next;
          			continue;
        		}
				
				// fill data buffer
				void *data = malloc( node->obj()->arraySize() * sizeof(double) );
				double *ddata = (double *) data;
				for(size_t i=0; i<node->obj()->arraySize(); i++) ddata[i] = node->obj()->dbl(n_entry*node->obj()->arraySize() + i);
			
      			status = CDFlib(PUT_, gENTRY_DATA_, CDF_DOUBLE,
                      	(long) node->obj()->arraySize(), data, NULL_);
      			if (status != CDF_OK) {
					cout << "Failed to write Global DBL entry " << n_entry << endl;
				}
				free(data);
			}
    	} 
		else if(node->obj()->is_int()) {
				
			for (n_entry=0; n_entry < (long) node->obj()->seqSize(); n_entry++){

      			status = CDFlib(SELECT_, gENTRY_, n_entry, NULL_);
        		if(status != CDF_OK){
					cout << "Failed to select Global entry " << n_entry << endl;
					node = node->next;
          			continue;
        		}
				
				// fill data buffer
				void *data = malloc( node->obj()->arraySize() * sizeof(int) );
				int *idata = (int *) data;
				for(size_t i=0; i<node->obj()->arraySize(); i++) idata[i] = node->obj()->itg(n_entry*node->obj()->arraySize() + i);
			
      			status = CDFlib(PUT_, gENTRY_DATA_, CDF_INT4,
                      	(long) node->obj()->arraySize(), data, NULL_);
      			if (status != CDF_OK) {
					cout << "Failed to write Global INT entry " << n_entry << endl;
				}
				free(data);
			}
		} 
		else if(node->obj()->is_str()) {
			// find longest str
			size_t len = node->obj()->str(0).size();
			for (size_t n=0; n < node->obj()->seqSize(); n++) if(len < node->obj()->str(n).size()) len = node->obj()->str(n).size();
			
			for (n_entry=0; n_entry < (long) node->obj()->seqSize(); n_entry++){

      			status = CDFlib(SELECT_, gENTRY_, n_entry, NULL_);
        		if(status != CDF_OK){
					cout << "Failed to select Global entry " << n_entry << endl;
					node = node->next;
          			continue;
        		}
				
				// fill data buffer
				void *data = malloc( sizeof(char) * (len+1) );
				char *sdata = (char *) data;
				strcpy( &(sdata[0]), node->obj()->str(n_entry).c_str() );
				size_t pad = node->obj()->str(n_entry).size();
				while (pad < len) {
					sdata[pad] = ' ';	
					pad++;
				}
				sdata[pad] = '\0';
      			status = CDFlib(PUT_, gENTRY_DATA_, CDF_CHAR, len, data, NULL_);
      			if (status != CDF_OK) {
					cout << "Failed to write Global STR entry " << n_entry << " of " << node->name() << endl;
				}
				free(data);
			}
		}
		else if(node->obj()->is_time()) {
			long cdfType = getCDFDataType(node->obj());
			void *data;
				
			for (n_entry=0; n_entry < (long) node->obj()->seqSize(); n_entry++){

      			status = CDFlib(SELECT_, gENTRY_, n_entry, NULL_);
        		if(status != CDF_OK){
					cout << "Failed to select Global entry " << n_entry << endl;
					node = node->next;
          			continue;
        		}
				
				// fill data buffer
				if(cdfType == CDF_EPOCH16){
					data = malloc( node->obj()->arraySize() * sizeof(double) * 2 );
					double *tdata = (double *)data;
					for(size_t i=0; i<node->obj()->arraySize(); i++) {
						node->obj()->time(n_entry*node->obj()->arraySize() + i).getCDFepoch16(&(tdata[2*i]));
					}
				}
				else{
					data = malloc( node->obj()->arraySize() * sizeof(double) );
					double *tdata = (double *)data;
					for(size_t i=0; i<node->obj()->arraySize(); i++) {
						tdata[i] = node->obj()->time(n_entry*node->obj()->arraySize() + i).getCDFepoch();
					}
				}
				
      			status = CDFlib(PUT_, gENTRY_DATA_, cdfType,
                      	(long) node->obj()->arraySize(), data, NULL_);
      			if (status != CDF_OK) {
					cout << "Failed to write Global TIME entry " << n_entry << endl;
				}
				free(data);
			}
		}
		else if(node->obj()->is_event()) {
			long cdfType = getCDFDataType(node->obj());
			void *data;
			
			for (n_entry=0; n_entry < (long) node->obj()->seqSize(); n_entry++){

      			status = CDFlib(SELECT_, gENTRY_, n_entry, NULL_);
        		if(status != CDF_OK){
					cout << "Failed to select Global entry " << n_entry << endl;
					node = node->next;
					continue;
        		}
				
				// fill data buffer
				if(cdfType == CDF_EPOCH16){
					data = malloc( node->obj()->arraySize() * sizeof(double) *4);
					double *edata = (double *)data;
					for(size_t i=0; i<node->obj()->arraySize(); i++) {
						node->obj()->event(n_entry*node->obj()->arraySize() + i).start().getCDFepoch16(&(edata[4*i]));
						node->obj()->event(n_entry*node->obj()->arraySize() + i).end().getCDFepoch16(&(edata[4*i+2]));
					}
				}
				else{
					data = malloc( node->obj()->arraySize() * sizeof(double) *2);
					double *edata = (double *)data;
					for(size_t i=0; i<node->obj()->arraySize(); i++) {
						edata[2*i] = node->obj()->event(n_entry*node->obj()->arraySize() + i).start().getCDFepoch();
						edata[2*i+1] = node->obj()->event(n_entry*node->obj()->arraySize() + i).end().getCDFepoch();
					}
				}
      			status = CDFlib(PUT_, gENTRY_DATA_, cdfType,
                      	(long) node->obj()->arraySize()*2, data, NULL_);
      			if (status != CDF_OK) {
					cout << "Failed to write Global TIME (event) entry " << n_entry << endl;
				}
				free(data);
			}
		}
		node = node->next;
	} 

    /* Append the software version of this module */

    status = CDFlib(SELECT_, ATTR_NAME_, "Software_version", NULL_);
    if(status != CDF_OK){
      status = CDFlib(CREATE_, ATTR_, "Software_version",
                                      GLOBAL_SCOPE, &attr_num, NULL_);
    }

    status = CDFlib(GET_, ATTR_MAXgENTRY_, &n_entry, NULL_);
    if (status != CDF_OK) return false;
    
    n_entry++;
    status = CDFlib(SELECT_, gENTRY_, n_entry, NULL_);

    status = CDFlib(PUT_, gENTRY_DATA_, CDF_CHAR,
                                        strlen(QIE_VERSION_NUM),
                                        QIE_VERSION_NUM, NULL_);
    if(status != CDF_OK) return false;

	return true;
}

// ------------------------------------------------

size_t QieData::sizeOfCDF( DvObject_var &obj, size_t &len)
{	
	len = 1;

	if(obj->is_dbl()) return sizeof(double);
	if(obj->is_int()) return sizeof(int);
	if(obj->is_str()) {
		for(size_t i=0; i<obj->totalSize(); i++) if(obj->str(i).size() > len) len = obj->str(i).size();
		return len*sizeof(char);
	}
	if(obj->is_time()) return sizeof(double);
	if(obj->is_event()) return sizeof(double)*2;

    return sizeof(double);
} 

// ------------------------------------------------

long QieData::getCDFDataType(DvObject_var &obj)
{
	if(obj->is_dbl() ) return CDF_DOUBLE;
	else if(obj->is_int()) return CDF_INT4;
	else if(obj->is_time()) {
		if(dataSecRes >= 0){
			// force resolution
			if(dataSecRes > 3) return CDF_EPOCH16;
			else return CDF_EPOCH;
			
		}
		else if(obj->secRes > 3) return CDF_EPOCH16;
		else return CDF_EPOCH;
	}
	else if( obj->is_event()) {
		if(dataSecRes >= 0){
			// force resolution
			if(dataSecRes > 3) return CDF_EPOCH16;
			else return CDF_EPOCH;
			
		}
		else if(obj->secRes > 3) return CDF_EPOCH16;
		else return CDF_EPOCH;
	}
	else return CDF_CHAR;
	
}

// -------------------------------------------

bool QieData::writeVariableCDF(int n){

	CDFstatus status;
	long var_num;
	long num_elems = 1;
    long numDims = 0;
	long recVary = NOVARY;
	long dim_sizes[CDF_MAX_DIMS];
	long dim_varys[CDF_MAX_DIMS];
	long dim_indices[CDF_MAX_DIMS];
	long dim_intervals[CDF_MAX_DIMS];

	// ensure at least one record exists (create 1 rec of fills), and ensure Vobj not nil
	ensureVariable(n);

	// set up for this variale
	QieVariable *V = QieVars[n];
	
    if(V->Vobj->arraySize() > 1 ){
        for(size_t i=0; i<V->Vobj->nDims(); i++){
            if(V->Vobj->is_event()) dim_sizes[i] = V->Vobj->Dims()[i] * 2; // convert event to time
            else dim_sizes[i] = V->Vobj->Dims()[i];
            dim_varys[i] = VARY;
            dim_indices[i] = 0;
            dim_intervals[i] = 1;
        }
        numDims = (long)V->Vobj->nDims();
    }
    else if(V->Vobj->is_event()){
        // special case of convert event to time
        dim_sizes[0] = 2;
        dim_varys[0] = VARY;
        dim_indices[0] = 0;
        dim_intervals[0] = 1;
        numDims = 1;
    }
    
	
	if(V->Vobj->is_str()) num_elems = V->Vobj->maxStrLen()+1;
	if(V->recVary) recVary = VARY;
	
	long cdfType = getCDFDataType(V->Vobj);
	if(V->datatype == DATA_FLT) cdfType = CDF_FLOAT;
	
    /* create variable in CDF file */
   
    status = CDFlib(CREATE_, zVAR_, V->name.c_str(),
                                    cdfType,
                                    num_elems,
                                    numDims,
                                    dim_sizes,
                                    recVary,
                                    dim_varys,
                                    &var_num,
                    NULL_);

	if (status != CDF_OK) {
		cout << "Failed to create variable in CDF" << endl;
  		return false;
	}


	/* allocate space in cdf */

	status = CDFlib(PUT_, zVAR_EXTENDRECS_, (long)V->Vobj->seqSize(), NULL_);

	/* select parameters for hyper write */

	status = CDFlib (SELECT_, zVAR_RECNUMBER_, 0,
                            zVAR_RECCOUNT_, (long)V->Vobj->seqSize(),
                            zVAR_RECINTERVAL_, 1,
                            zVAR_DIMCOUNTS_, dim_sizes,
                            zVAR_DIMINDICES_, dim_indices,
                            zVAR_DIMINTERVALS_, dim_intervals,
                   NULL_);
	if (status != CDF_OK) {
		cout << "Failed to select variable in CDF" << endl;
  		return false;
	}

	void *data;
	
  	// fill data array
	if(V->Vobj->is_dbl()) {
		if(cdfType == CDF_FLOAT){
			// special case not using double
			size_t len = V->Vobj->totalSize();
			data = malloc( len * sizeof(float) );
			float *fdata = (float *) data;

			for (size_t j=0; j<len; j++){
				fdata[j] = (float) V->Vobj->dbl(j);
    		} 
		}
		else{
			size_t len = V->Vobj->totalSize();
			data = malloc( len * sizeof(double) );
			double *ddata = (double *) data;

			for (size_t j=0; j<len; j++){
				ddata[j] = V->Vobj->dbl(j);
    		} 
		}
	}
	else if(V->Vobj->is_int()) {
		
		size_t len = V->Vobj->totalSize();
		data = malloc( len * sizeof(int) );
		int *idata = (int *) data;

		for (size_t j=0; j<len; j++){
			idata[j] = V->Vobj->itg(j);
    	} 
	}
	else if(V->Vobj->is_str()) {
		
		size_t len = V->Vobj->totalSize();
		data = malloc( len * sizeof(char) * num_elems);
		char *sdata = (char *) data;

		for (size_t j=0; j<len; j++){
			strcpy(&(sdata[j*num_elems]), V->Vobj->str(j).c_str());
    	} 
	}
	else if(V->Vobj->is_time()) {
		if(cdfType == CDF_EPOCH16){
			size_t len = V->Vobj->totalSize();
			data = malloc( len * sizeof(double) * 2 );
			double *tdata = (double *) data;

			for (size_t j=0; j<len; j++){
				 V->Vobj->time(j).getCDFepoch16(&(tdata[2*j]));
    		} 
		}
		else{
			size_t len = V->Vobj->totalSize();
			data = malloc( len * sizeof(double) );
			double *tdata = (double *) data;

			for (size_t j=0; j<len; j++){
				tdata[j] = V->Vobj->time(j).getCDFepoch();
    		} 
		}
	}
	else if(V->Vobj->is_event()) {
		if(cdfType == CDF_EPOCH16){
			size_t len = V->Vobj->totalSize();
			data = malloc( len * sizeof(double) * 4);
			double *edata = (double *) data;

			for (size_t j=0; j<len; j++){
				V->Vobj->event(j).start().getCDFepoch16(&(edata[4*j]));
				V->Vobj->event(j).end().getCDFepoch16(&(edata[4*j+2]));
    		} 
		}
		else{
			size_t len = V->Vobj->totalSize();
			data = malloc( len * sizeof(double) * 2);
			double *edata = (double *) data;

			for (size_t j=0; j<len; j++){
				edata[2*j] = V->Vobj->event(j).start().getCDFepoch();
				edata[2*j+1] = V->Vobj->event(j).end().getCDFepoch();
    		} 
		}
	}
	else return false;
	
	// write out data 

	status = CDFlib(PUT_, zVAR_HYPERDATA_, data, NULL_);
	
	free(data);

	if (status != CDF_OK) {
		cout << "Failed to write variable in CDF" << endl;
  		return false;
	}

	/* write variable attributes */

	writeVarAttrs(n);

	return true;
}


// -------------------------------------------

bool QieData::createCDF( bool backCompat, bool replace)
{
	DvString command;
	DvString fullName;
	
	CDFid id;               /* ID of CDF for subsequent operations */
	CDFstatus status;
	long relNum;
	long verNum;

	// test if file exists 
	fullName = outFile.name + outFile.extn;
	
	status = CDFlib(OPEN_, CDF_, fullName.c_str(), &id, NULL_);
	
	if( status == CDF_OK) {
		if(replace){
			status = CDFlib(DELETE_, CDF_, NULL_);
			if( status != CDF_OK) {
				cout << "Write Error: Cannot delete file: " << fullName << endl;
				return false;
			}
		}
		else{
			CDFlib(CLOSE_, CDF_, NULL_);
			cout << "Write Error: File already exists: " << fullName << endl;
			return false;
		}
	}
	
	CDFlib(GET_, LIB_RELEASE_, &relNum, NULL_);
	       
	CDFlib(GET_, LIB_VERSION_, &verNum, NULL_);
	printf(" cdf V%ld", verNum);
	printf(".%ld\n", relNum);
   
	if(backCompat ){
		CDFsetFileBackward(BACKWARDFILEon);
		printf("file created backwardly compatible with cdf V2.7\n");
	}
	else {
		CDFsetFileBackward(BACKWARDFILEoff);
		printf("file created with cdf V%ld.%ld format\n", verNum, relNum );
	}
   
    long numDims = 0;
	static long dimSizes[1] = {0};
	status = CDFlib(CREATE_, CDF_, fullName.c_str(), numDims, dimSizes, &id,
                   PUT_, CDF_ENCODING_, NETWORK_ENCODING,
				   CDF_MAJORITY_, ROW_MAJOR, 
				   CDF_FORMAT_, SINGLE_FILE, NULL_);

	if (status != CDF_OK) return false;

/*
	CDFlib(CLOSE_, CDF_, NULL_);

	status = CDFlib(OPEN_, CDF_, fullName.c_str(), &id, NULL_);
	if (status != CDF_OK) {
		cout << "Failed to open CDF for write" << endl;
		return false;
	}

*/
	return true;

 }


// -------------------------------------------

void QieData::ensureVariable(int n){

	QieVariable *V = QieVars[n];

	if(V->Vobj.is_ok()) return; // Data Exists

	/* ensure at least one record exists (create 1 rec of fills) */

	V->nRecs = 1;
	V->recVary = NOVARY;
	
	switch (V->datatype){
	
		case DATA_DBL:
		{
			if(V->objDim.size() > 0) V->Vobj = new DvObject((double) 0.0, V->objDim, (size_t) 1 );
			else V->Vobj = new DvObject((double) 0.0, (size_t) 1 );
			break;
		}
		case DATA_FLT:
		{
			if(V->objDim.size() > 0) V->Vobj = new DvObject((double) 0.0, V->objDim, (size_t) 1 );
			else V->Vobj = new DvObject((double) 0.0, (size_t) 1 );
			break;
		}
		case DATA_INT:
		{
			if(V->objDim.size() > 0) V->Vobj = new DvObject((int) 0, V->objDim, (size_t) 1 );
			else V->Vobj = new DvObject((int) 0, (size_t) 1 );
			break;
		}
		case DATA_STR:
		{
			if(V->objDim.size() > 0) V->Vobj = new DvObject(DvString(), V->objDim, (size_t) 1 );
			else V->Vobj = new DvObject(DvString(), (size_t) 1 );
			break;
		}
		case DATA_TIME:
		{
			if(V->objDim.size() > 0) V->Vobj = new DvObject(DvTime(), V->objDim, (size_t) 1 );
			else V->Vobj = new DvObject(DvTime(), (size_t) 1 );
			break;
		}
		case DATA_EVENT:
		{
			if(V->objDim.size() > 0) V->Vobj = new DvObject(DvEvent(), V->objDim, (size_t) 1 );
			else V->Vobj = new DvObject(DvEvent(), (size_t) 1 );
			break;
		}
	}
               
}

void QieData::writeVarAttrs(int n){

	long attr_num;
	CDFstatus status;
  
	// select entry appropriate for variable 
	QieVariable *V = QieVars[n];

    status = CDFlib(SELECT_, zENTRY_NAME_, V->name.c_str(), NULL_);
	if(status != CDF_OK) return;              

	// loop through entries in var attributes in structure 
	
	DvNode *node = V->Vxref.first(); 
 	while(node){

   		if(node->obj()->is_ok()){

     		// select or create attribute 

     		status = CDFlib(SELECT_, ATTR_NAME_, node->name().c_str(), NULL_);
     		if(status != CDF_OK ){
       			status = CDFlib(CREATE_, ATTR_, node->name().c_str(), VARIABLE_SCOPE, &attr_num, NULL_);
       			if(status != CDF_OK){
					node = node->next;
					continue;   // skip this as we can't select or create it 
       			}
     		}
			
			// check G Attr not already using this name
            long len = (long) node->obj()->totalSize();
			long attrScope;
			status = CDFlib (GET_, ATTR_SCOPE_, &attrScope, NULL_);
			if(status != CDF_OK){
				char text[CDF_STATUSTEXT_LEN+1];
				CDFgetStatusText (status, text);
				cout << text << endl;
				node = node->next;
				continue;
			}
			if(attrScope != VARIABLE_SCOPE){
				DvString vAttrName = node->name();
				vAttrName += "_V";
				status = CDFlib(SELECT_, ATTR_NAME_, vAttrName.c_str(), NULL_);
     			if(status != CDF_OK ){
					status = CDFlib(CREATE_, ATTR_, vAttrName.c_str(), VARIABLE_SCOPE, &attr_num, NULL_);
       				if(status != CDF_OK){
						node = node->next;
						continue;   // skip this as we can't select or create it 
       				}
				}
			}
			
     		// write  entry 

			if(node->obj()->is_str() ){
					
				// num_elems for a Vattr is the number of strings in attr array : Changed for CEF2
				DvString concatAttr="";
				for (size_t i=0; i<node->obj()->totalSize(); i++){
					if( !concatAttr.empty()) concatAttr += ", ";
					concatAttr += node->obj()->str(i);
				}
				char *concatAttrChar;
				
				if(concatAttr.empty()){
					concatAttrChar = (char *) malloc (sizeof(char)*2);
					strcpy(concatAttrChar, " "); // ensure something is written or CDFlib fails
				}
				else{
					// allow space for each string + 2 spaces per string for "' " This also allows for terminator 
					concatAttrChar = (char *) malloc (sizeof(char) * (concatAttr.size()+1) );
					strcpy(concatAttrChar, concatAttr.c_str() );
				}
				status = CDFlib(PUT_, zENTRY_DATA_,
                        CDF_CHAR,
                        (long)strlen(concatAttrChar),
                        (void *)concatAttrChar,
                        NULL_);

				if (status != CDF_OK)  {
					printf( "CDF error writing char V attribute %s of variable %s, GAttr of same name?\n",
						node->name().c_str(), V->name.c_str() );
				}
			}
			else{
				long cdfType = getCDFDataType(node->obj());
				void * data;
				if(node->obj()->is_dbl() ){
					data = (double *) malloc(sizeof(double) * node->obj()->totalSize());
					double *ddata = (double *) data;
					for(size_t i=0; i<node->obj()->totalSize(); i++){
						ddata[i] = node->obj()->dbl(i);
					}
				}
				else if(node->obj()->is_int() ){
					data = (int *) malloc(sizeof(int) * node->obj()->totalSize());
					int *idata = (int *) data;
					for(size_t i=0; i<node->obj()->totalSize(); i++){
						idata[i] = node->obj()->itg(i);
					}
				}
				else if(node->obj()->is_time() ){
					if(cdfType == CDF_EPOCH16){
						data = (double *) malloc(sizeof(double) * node->obj()->totalSize() * 2);
						double *ddata = (double *) data;
						for(size_t i=0; i<node->obj()->totalSize(); i++){
							node->obj()->time(i).getCDFepoch16(&(ddata[2*i]));
						}
					}
					else{
						data = (double *) malloc(sizeof(double) * node->obj()->totalSize());
						double *ddata = (double *) data;
						for(size_t i=0; i<node->obj()->totalSize(); i++){
							ddata[i] = node->obj()->time(i).getCDFepoch();
						}
					}
				}
				else if(node->obj()->is_event() ){
                    len = (long) node->obj()->totalSize() * 2; // converting event to 2 times
					if(cdfType == CDF_EPOCH16){
						data = (double *) malloc(sizeof(double) * 4 * node->obj()->totalSize());
						double *ddata = (double *) data;
						for(size_t i=0; i<node->obj()->totalSize(); i++){
							node->obj()->event(i).start().getCDFepoch16(&(ddata[4*i]));
							node->obj()->event(i).end().getCDFepoch16(&(ddata[4*i+2]));
						}

					}
					else{
						data = (double *) malloc(sizeof(double) * 2 * node->obj()->totalSize());
						double *ddata = (double *) data;
						for(size_t i=0; i<node->obj()->totalSize(); i++){
							ddata[2*i] = node->obj()->event(i).start().getCDFepoch();
							ddata[2*i+1] = node->obj()->event(i).end().getCDFepoch();
						}
					}
				}
				else status = CDF_WARN;
				
				if (status == CDF_OK) {
                    status = CDFlib(PUT_, zENTRY_DATA_, cdfType, len, data, NULL_);
				}
				
				if (status != CDF_OK) {
					printf( "CDF error writing  V attribute %s of variable %s\n",
					node->name().c_str(), V->name.c_str() );
				}

			}

   		} // end if on attr not NULL */
		node = node->next;
		
	} // end while over V attributes 


}


