
#include "qie_caa.h"
#include <dirent.h>

// Globals
// ------------------------------------------

FILE * fp_display;

// ------------------------------------------

void setDebug(bool d_on, const char * fpath){
	
	if(d_on) 
	{
		if(fpath == NULL) fp_display = stdout;
		else 
		{
			if(fp_display == stdout || fp_display == NULL) fp_display = fopen(fpath, "w");
			else fp_display = freopen(fpath, "w", fp_display);
		}
	}
	else fp_display = NULL;
	
}

// ------------------------------------------

void  QieMsg(const char * status, const char *line1, const char *line2, const char *line3  )  
{
   
   std::string msg_line;

					msg_line =  std::string(" ");
  if(line1 != NULL) msg_line += std::string(line1);
					msg_line += " ";
  if(line2 != NULL) msg_line += line2;
					msg_line += " ";
  if(line3 != NULL) msg_line += line3;
  
   if (fp_display == stdout){   // trap NULL file pointer when debug is off
      printf( "%s\n", msg_line.c_str());
   }
   else if(fp_display != NULL)
   {
         fprintf(fp_display, "%s\n", msg_line.c_str());
   }
   

   return;

} 

// ------------------------------------------------

vector <DvString> QieFindHelperFiles(const char * dataFile, const char * dotExtn){

	vector <DvString> helperList;
 
	struct dirent *dir_ent;
	DIR *dir_ptr;
	int ptr;
	
	char pathDelim = '/';
	DvString desktop("~/Desktop/");
	
#if defined _WIN32 || defined WIN32

#ifndef WINDOWS
	// QSAS defines WINDOWS, Qtran does not. 
	// QSAS uses msys so '/' used in QSAS 
	pathDelim = '\\'; // Windows uses backslash
	desktop = "~\\Desktop\\";
#endif

#endif

	DvString hdPath(dataFile);
 
	hdPath = hdPath.beforeR(pathDelim);
	if(hdPath.empty()) hdPath = ".";
	
	/*  search directory for a header */

	if( (dir_ptr=opendir(hdPath.c_str())) == NULL) {
		QieMsg("ERROR", "Unable to open directory " , hdPath.c_str());
		return helperList;
	}

	/* list directory entries */

	while( (dir_ent=readdir(dir_ptr)) != NULL ){
		std::string hdName =  dir_ent->d_name;
		ptr = hdName.find(dotExtn);

		if( ptr != -1 && ptr == (int) hdName.length()-4) {
			DvString hdrFound = hdPath.c_str();
			hdrFound += pathDelim;
			hdrFound += hdName;
			helperList.push_back(hdrFound);
		}
	}
	closedir(dir_ptr);
	
	return helperList;

}

// ------------------------------------------------

bool   QieStrIsNULL(const char *txt)
{
	if(txt == NULL) return true;
	if( strcmp(txt, "") == 0)  return true;
    return false;

} 



 // ---------------------------------------------------------

char QieGetDelim (std::string value)
{
	size_t ptr = 0;
	// may be quoted 
	if(value[ptr] == '\'') ptr++;
	if(value[ptr] == '\"') ptr++;
	
	if (value.empty()) return ' ';
	if (value[ptr] == ' ') return ' ';

	if( value[ptr] == '\n' || value[ptr] == '\0') return ' ';
	
	// escaped chars
	if( value[ptr] != '\\' || value.length() == ptr+1) return value[ptr]; // this is non-escaped or \ itself
	ptr++;
	if(value[ptr] == 't') return '\t';
	if(value[ptr] == '\'') return '\'';
	if(value[ptr] == '\"') return '\"';
	if(value[ptr] == 'n') return '\n';

  return ' ';

} 



// ---------------------------------------------------------

bool QieIsISOtime(std::string TestStrISO){

	size_t ptr=0;
	size_t str_len = 0;
	std::string subString;

	/* trap null string */

	if(TestStrISO.empty() || TestStrISO == "") return false;

	// ISO is " 1995-01-23 02:33:17.235 " or " 1995-01-23T02:33:17.235Z " 

	// remove any leading space 

	while(ptr < TestStrISO.size() && TestStrISO[ptr] == ' ')	ptr++;
		
	//  check str is at least long enough to give seconds 
	if( TestStrISO.size() - ptr < 19 ) return false;
	
	/* parse off date - no field can be zero */

	str_len = TestStrISO.find( "-"); 
	if(str_len == string::npos) return false;
	if( !QieIsNumber( TestStrISO.substr(ptr, str_len - ptr) ) ) return false;
	
	subString = TestStrISO.substr(str_len+1, TestStrISO.size() - str_len -1); // next token

	str_len = subString.find( "-"); 
	if(str_len == string::npos) return false;
	if(!QieIsNumber(subString.substr(0, str_len ))) return false;
	subString = subString.substr(str_len+1, subString.size() - str_len -1); // next token

	str_len = subString.find( "T"); 
	if(str_len == string::npos) str_len = subString.find( " "); 
	if(str_len == string::npos) return false;
	if(!QieIsNumber(subString.substr(0, str_len ))) return false;
	subString = subString.substr(str_len+1, subString.size() - str_len -1); // next token

	// parse off time portion - time fields can be zero, just test for existence 

	str_len = subString.find( ":"); 
	if(str_len == string::npos) return false;  // hours
	if (!QieIsNumber(subString.substr(0, str_len )) ) return false;
	subString = subString.substr(str_len+1, subString.size() - str_len -1); // next token

	str_len = subString.find( ":"); if(str_len == string::npos) return false;  // minutes
	if (!QieIsNumber(subString.substr(0, str_len )) ) return false;
	subString = subString.substr(str_len+1, subString.size() - str_len -1); // next token

	// just test seconds, decimal and fractional seconds are optional

	if (!QieIsNumber(subString.substr(0, 2 )) ) return false;

	//if we get this far it can be read as a time string (might still be wrong though, e.g. 90 instead of 1990 

	return true;

}
bool QieIsISOtimeRange(std::string TestStrISO){

	size_t str_len = 0;

	/* trap null string */

	if(TestStrISO.empty() || TestStrISO == "") return false;

	// ISO is " 1995-01-23 02:33:17.235/1995-01-23 02:33:17.235 " or " 1995-01-23T02:33:17.235Z/1995-01-23T02:33:17.235Z " 

	str_len = TestStrISO.find( "/");

	if(str_len == string::npos) return false;

	if( !QieIsISOtime(TestStrISO.substr(0, str_len)) ) return false;
	return QieIsISOtime(TestStrISO.substr(str_len+1));

}

// ---------------------------------------------------------

bool QieIsNumber(std::string ptr){

  for(size_t i=0; i<ptr.size(); i++){
    if(!isdigit(ptr[i]) && ptr[i] != '.') return false;
  }
  return true;
}



// utilities to track if cef 1 or 2 is default

bool cefRead2 (int set){
  static bool read_cef_2=true;
  if(set == -1) return read_cef_2;
  if(set == 0) {
    read_cef_2 = false;
    return read_cef_2;
  }
  if(set == 1) {
    read_cef_2 = true;
    return read_cef_2;
  }
  return read_cef_2;

}

bool cefWrite2 (int set){
  static bool write_cef_2=true;

    if(set == 0) {
    write_cef_2 = false;
    return write_cef_2;
  }
  if(set == 1) {
    write_cef_2 = true;
    return write_cef_2;
  }
  return write_cef_2;

}

