//  qpolar.h
//
//  Tony Allen
//  Sept 2010

#ifndef QSAS_POLAR_UI_H_
#define QSAS_POLAR_UI_H_

#include <qsplitter.h>
#include <qstatusbar.h>
#include <qspinbox.h>
#include <qtabwidget.h>
#include <qtabbar.h>
#include <qlayout.h>

#include <QApplication>
#include <QMimeData>
#include <QStyle>
#include <QHeaderView>
#include <QPainter>
#include <QScrollArea>

#include "qt_version.h"
#include "qslot.h"
#include "QbTimeEditors.h"
#include "IntervalDialog.h"

#include "floatspinbox.h"
#include "IntervalWidgets.h"
#include "ColourTable.h"

#include "qpui.h"
#include "PlotWindow.h"

#define QPOLAR_VERSION "1.0"

// class  QpWindow
// -----------------------

class QpolarWindow : public QMainWindow
{
	Q_OBJECT
 
 	// GUI widgets
	QTabWidget *tab_widget;
	QStatusBar* _status_bar;
	QTextEdit *textArea;
	
	// Data tab widgets
 	QguWLinput * dataSlot;
	RecordSlider *atSlider;
	QCheckBox * swapR;
	QCheckBox * autoPlot;
	QSpinBox *delay;
	
	// plot tab widgets
	QLineEdit *titleSlot;
	QCheckBox * autoTitle;
	QLineEdit *RLabSlot;
	QLineEdit *ThLabSlot;
	QComboBox *plotType;
	QComboBox *axisType;
	FloatSpinBox *rScale;
	QComboBox *dataPtOpts;
	QSpinBox *fractData;
	QLineEdit *dataOrigin;
	QCheckBox *showCont;
	QLineEdit * contSpace;
	QComboBox *contLabelType;
	QCheckBox *radiusLog;
	QCheckBox *radAuto;
	QLineEdit *RadMin;
	QLineEdit *RadMax;
	QCheckBox *colourLog;
	QCheckBox *colourAuto;
	QLineEdit *colMin;
	QLineEdit *colMax;
	ColourTableButton *colourTable;
	QLineEdit *colLabSlot;
	FloatSpinBox *title_size_fld;
	FloatSpinBox *text_size_fld;
    FloatSpinBox *line_width_fld;
    FloatSpinBox *axis_width_fld;
	QCheckBox *useLocalTime;

	// page tab widgets
	QComboBox *page_size;
	QComboBox *pageUnits;
	int oldUnits;
	QComboBox *dpiPage;
	QLineEdit *pageWidth;
	QLineEdit *pageHeight;
	QComboBox *pageOrientation;
	PlotID *plotWin_id;
	QCheckBox *newWindow;
	
	// flags and pointers
	bool userTitle;
	bool userRLab;
	bool userThLab;
	bool userColLab;
	bool orderIsRTh;
	bool full; //indicates whether bins span over 360 degree
	bool dThetaSet;
	bool dRSet;
	bool zLogEmpty;
	bool isArcfill;
	DvObject_var dobj;
	DvObject_var rDep;
	DvObject_var thetaDep;
	PLFLT **z;
	PLFLT **zLog;
	PLFLT *theta_plus;
	PLFLT *theta_minus;
	PLFLT *r_plus;
	PLFLT *r_minus;
	PLFLT *r;
	PLFLT *theta;
	PLFLT width;
	PLFLT height;
	int plot_id;
	QList<PlotWindow*> PwindowList;
	PlotWindow* win;
	int DPI;
	double conv;
	string timeText;

	// from ppp
	int nr; //length of r array
	int ntheta; //length of theta array
	int n; //total no. of data points = nr*ntheta
	int nr_start; // starting radial index (> userRmin)
	int nr_end; // end radial index (< userRmax)
	double contStep;
	double ang2rad;
	double rad2deg;
	double rfillVal;
	double thetafillVal;
	double zfillVal;
	PLFLT zmin;
	PLFLT zmax;
	PLFLT rmax;
	PLFLT rmin;
	PLFLT rminNonZero;
	PLFLT userRmax;
	PLFLT userRmin;
	PLFLT rOffset;
	double fract;
	PLFLT thetamin;
	PLFLT thetamax;
	PLFLT rroom;
	PLcGrid2 cgrid;
	// --------
	
	public:
		QpolarWindow( QWidget *parent=0);
		~QpolarWindow();
        void saveAsXML(QString path);
        void restoreXML(QString path, bool replace=true);
        void setwindow(bool animation=false);
		void setData();
		bool initData();
		void sort_r_theta();
		void quick_sort_2d(PLFLT xarray[], PLFLT z[], int start, int end, int length, PLFLT x_plus[], PLFLT x_minus[]);
		inline PLFLT min(PLFLT x, PLFLT y) { return (x < y ? x : y); }
		inline PLFLT max(PLFLT x, PLFLT y) { return (x > y ? x : y); }
		inline PLFLT SaferLog10(PLFLT x) { return (x < 1.0e-35) ? -35.0 : log10(x); }
		PLFLT near_nice_number(PLFLT x);
		PLFLT nice_min(PLFLT zmin, PLFLT stepsize);
		void shade_cont() ;
		void arcfill();
		void plot_data_pts();
		int colscale();
		void drawcolscale();
		void drawaxis(const char *rlabel) ;
		int cont_lvl(PLFLT clevel[]);
		void select_vpor(int n);
		bool getDimensions(int &nx, int &ny);
		void QpolarAppendTextDisplay(const char * txt);
		void clearTextArea();
		void setColLab();
		void nonZeroMinMax();
		void getMinMax();
		bool setDeltaTheta(PLFLT *result, const char *delta);
		bool setDeltaRad(PLFLT *result, const char *delta);
		QDomElement toXMLElement(QDomDocument&) const;
		void fromXMLElement(QDomElement&);
        QString _filename;        ///< Filename used to save the current session or working list content
        QString SAVE_EXTENSION;   ///< File extension used to save the current session or working list content

	public slots:
		void status(const QString& str) {_status_bar->showMessage(str,2000);};
		void doPlot(bool animation=false);
		void doRun();
        void saveAsXML();
        void restoreXML();
		void setDataObj();
		void setRecord();
		void launchManual();
		void launchPolarPlotHelp();
		void launchVersion();
		void autosetCol();
		void autosetRad();
		void setUserTitle();
		void setUserRLab();
		void setUserThLab();
		void setColUser();
		void setSwapRTheta();
		void setLogRad();
		void setLogCol();
		void setAutoRad();
		void changeRadius();
		void setUserColLab();
//		void setPlotId(const QString& txt);
		void setPage();
		void setTitle();
        void saveSession(QString path){saveAsXML(path+"/saveSet.qx_polar_plot");}
        void restoreSession(QString path){restoreXML(path+"/saveSet.qx_polar_plot");}
		void setLocalTime();
		void setAxisLabels();
		
	signals:
		void reValidateData();

	protected:
		void keyPressEvent(QKeyEvent * e );
		QPushButton *plotBtn;
		QPushButton *runBtn;
		QPushButton *dismissBtn;
};



void QPolarMakeWindow(QWidget*,const char*);
void QpolarShow();







# endif // QSAS_POLAR_UI_H_
